@@ -1,5 +1,4 @@
-
-// Copyright (C) 2021 - 2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2021 - 2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -19,7 +18,8 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 // THE SOFTWARE.
 
-#include "accuracy_test.h"
+#include "../../shared/accuracy_test.h"
+#include "../../shared/params_gen.h"
 
 std::vector<std::vector<size_t>> adhoc_sizes = {
     // sizes that exercise L1D_TRTRT subplan of 2D_RTRT or 3D_TRTRTR
@@ -67,7 +67,8 @@ static std::vector<std::vector<size_t>> ooffset_range = {{0, 0}, {1, 1}};
 
 INSTANTIATE_TEST_SUITE_P(adhoc,
                          accuracy_test,
-                         ::testing::ValuesIn(param_generator(adhoc_sizes,
+                         ::testing::ValuesIn(param_generator(test_prob,
+                                                             adhoc_sizes,
                                                              precision_range_sp_dp,
                                                              batch_range,
                                                              stride_range,
@@ -80,7 +81,8 @@ INSTANTIATE_TEST_SUITE_P(adhoc,
 
 INSTANTIATE_TEST_SUITE_P(DISABLED_offset_adhoc,
                          accuracy_test,
-                         ::testing::ValuesIn(param_generator(adhoc_sizes,
+                         ::testing::ValuesIn(param_generator(test_prob,
+                                                             adhoc_sizes,
                                                              precision_range_full,
                                                              batch_range,
                                                              stride_range,
@@ -120,7 +122,28 @@ inline auto param_permissive_iodist()
                     param.placement      = std::get<1>(types);
                     param.itype          = std::get<2>(types);
                     param.otype          = std::get<3>(types);
-                    params.push_back(param);
+
+                    param.validate();
+
+                    const double roll = hash_prob(random_seed, param.token());
+                    const double run_prob
+                        = test_prob * (param.is_planar() ? complex_planar_prob_factor : 1.0)
+                          * (param.is_interleaved() ? complex_interleaved_prob_factor : 1.0)
+                          * (param.is_real() ? real_prob_factor : 1.0);
+
+                    if(roll > run_prob)
+                    {
+                        if(verbose > 4)
+                        {
+                            std::cout << "Test skipped (probability " << run_prob << " > " << roll
+                                      << ")\n";
+                        }
+                        continue;
+                    }
+                    if(param.valid(0))
+                    {
+                        params.push_back(param);
+                    }
                 }
             }
         }
@@ -137,7 +160,8 @@ INSTANTIATE_TEST_SUITE_P(adhoc_dist,
 inline auto param_adhoc_colmajor()
 {
     // generate basic FFTs of adhoc sizes
-    auto params = param_generator(adhoc_sizes,
+    auto params = param_generator(test_prob,
+                                  adhoc_sizes,
                                   {fft_precision_single},
                                   {2},
                                   stride_range,
@@ -218,6 +242,7 @@ inline auto param_adhoc_stride()
         for(const auto& types :
             generate_types(fft_transform_type_real_forward, {fft_placement_notinplace}, true))
         {
+
             fft_params param;
             param.length         = {4, 4, 4};
             param.precision      = precision;
@@ -230,7 +255,33 @@ inline auto param_adhoc_stride()
             param.otype          = std::get<3>(types);
             param.istride        = {16, 4, 1};
             param.ostride        = {16, 4, 1};
-            params.push_back(param);
+
+            param.validate();
+
+            {
+                const double roll = hash_prob(random_seed, param.token());
+                const double run_prob
+                    = test_prob * (param.is_planar() ? complex_planar_prob_factor : 1.0)
+                      * (param.is_interleaved() ? complex_interleaved_prob_factor : 1.0)
+                      * (param.is_real() ? real_prob_factor : 1.0);
+
+                if(roll > run_prob)
+                {
+                    if(verbose > 4)
+                    {
+                        std::cout << "Test skipped (probability " << run_prob << " > " << roll
+                                  << ")\n";
+                    }
+                    continue;
+                }
+                else
+                {
+                    if(param.valid(0))
+                    {
+                        params.push_back(param);
+                    }
+                }
+            }
 
             param.length         = {2, 2, 2};
             param.precision      = precision;
@@ -243,7 +294,33 @@ inline auto param_adhoc_stride()
             param.otype          = std::get<3>(types);
             param.istride        = {20, 6, 1};
             param.ostride        = {20, 6, 1};
-            params.push_back(param);
+
+            param.validate();
+
+            {
+                const double roll = hash_prob(random_seed, param.token());
+                const double run_prob
+                    = test_prob * (param.is_planar() ? complex_planar_prob_factor : 1.0)
+                      * (param.is_interleaved() ? complex_interleaved_prob_factor : 1.0)
+                      * (param.is_real() ? real_prob_factor : 1.0);
+
+                if(roll > run_prob)
+                {
+                    if(verbose > 4)
+                    {
+                        std::cout << "Test skipped (probability " << run_prob << " > " << roll
+                                  << ")\n";
+                    }
+                    continue;
+                }
+                else
+                {
+                    if(param.valid(0))
+                    {
+                        params.push_back(param);
+                    }
+                }
+            }
         }
     }
 
@@ -255,16 +332,27 @@ INSTANTIATE_TEST_SUITE_P(adhoc_stride,
                          ::testing::ValuesIn(param_adhoc_stride()),
                          accuracy_test::TestName);
 
-auto adhoc_tokens = {
-    "complex_forward_len_512_64_single_ip_batch_3_istride_192_3_CI_ostride_192_3_CI_idist_1_odist_"
-    "1_ioffset_0_0_ooffset_0_0",
-    "real_forward_len_1024_1024_1024_single_op_batch_1_istride_1048576_1024_1_R_ostride_525312_513_"
-    "1_HI_idist_1073741824_odist_537919488_ioffset_0_0_ooffset_0_0",
-    "complex_forward_len_6144_single_ip_batch_34_istride_35_CI_ostride_35_CI_idist_1_odist_1_"
-    "ioffset_0_0_ooffset_0_0",
+const auto adhoc_tokens = {
+    // clang-format off
+    "complex_forward_len_4_4_4_single_op_batch_2_istride_16_4_1_CI_ostride_4_16_1_CI_idist_64_odist_64_ioffset_0_0_ooffset_0_0",
+    "complex_forward_len_512_64_single_ip_batch_3_istride_192_3_CI_ostride_192_3_CI_idist_1_odist_1_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_1024_1024_1024_single_op_batch_1_istride_1048576_1024_1_R_ostride_525312_513_1_HI_idist_1073741824_odist_537919488_ioffset_0_0_ooffset_0_0",
+    "complex_forward_len_6144_single_ip_batch_34_istride_35_CI_ostride_35_CI_idist_1_odist_1_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_8192_single_ip_batch_65537_istride_1_R_ostride_1_HI_idist_8194_odist_4097_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_520_single_op_batch_270400_istride_1_R_ostride_1_HI_idist_520_odist_261_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_630_single_op_batch_396900_istride_1_R_ostride_1_HI_idist_630_odist_316_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_660_single_op_batch_435600_istride_1_R_ostride_1_HI_idist_660_odist_331_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_700_single_op_batch_490000_istride_1_R_ostride_1_HI_idist_700_odist_351_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_728_single_op_batch_529984_istride_1_R_ostride_1_HI_idist_728_odist_365_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_968_single_op_batch_937024_istride_1_R_ostride_1_HI_idist_968_odist_485_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_1020_single_op_batch_1040400_istride_1_R_ostride_1_HI_idist_1020_odist_511_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_378_42_single_ip_batch_66000_istride_44_1_R_ostride_22_1_HI_idist_16632_odist_8316_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_527_25_single_ip_batch_67500_istride_26_1_R_ostride_13_1_HI_idist_13702_odist_6851_ioffset_0_0_ooffset_0_0",
+    "real_forward_len_630_38_single_ip_batch_65540_istride_40_1_R_ostride_20_1_HI_idist_25200_odist_12600_ioffset_0_0_ooffset_0_0",
+    // clang-format on
 };
 
 INSTANTIATE_TEST_SUITE_P(adhoc_token,
                          accuracy_test,
-                         ::testing::ValuesIn(param_generator_token(adhoc_tokens)),
+                         ::testing::ValuesIn(param_generator_token(test_prob, adhoc_tokens)),
                          accuracy_test::TestName);
