@@ -18,10 +18,18 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 // THE SOFTWARE.
 
-#include "accuracy_test.h"
+#include "../../shared/accuracy_test.h"
+#include "../../shared/params_gen.h"
+
+extern bool fftw_compare;
 
 inline auto param_checkstride()
 {
+    // checkstride requires us to copy data back to the host for
+    // checking, which we only do when comparing against FFTW.
+    if(!fftw_compare)
+        return std::vector<fft_params>{};
+
     // tuples of length,stride,nbatch,dist to test.  strides are arranged so
     // there's space either between elements on the fastest dim, or
     // between dims, or both.
@@ -87,7 +95,28 @@ inline auto param_checkstride()
                         param.otype                = std::get<3>(types);
                         param.run_callbacks        = callback;
                         param.check_output_strides = true;
-                        params.push_back(param);
+
+                        param.validate();
+
+                        const double roll = hash_prob(random_seed, param.token());
+                        const double run_prob
+                            = test_prob * (param.is_planar() ? complex_planar_prob_factor : 1.0)
+                              * (param.is_interleaved() ? complex_interleaved_prob_factor : 1.0)
+                              * (param.is_real() ? real_prob_factor : 1.0);
+
+                        if(roll > run_prob)
+                        {
+                            if(verbose > 4)
+                            {
+                                std::cout << "Test skipped (probability " << run_prob << " > "
+                                          << roll << ")\n";
+                            }
+                            continue;
+                        }
+                        if(param.valid(0))
+                        {
+                            params.push_back(param);
+                        }
                     }
                 }
             }
