@@ -26,9 +26,10 @@
 
 #include <hip/hip_runtime.h>
 
+#include "../../shared/fftw_transform.h"
+#include "../../shared/hip_object_wrapper.h"
 #include "../../shared/rocfft_params.h"
-#include "fftw_transform.h"
-#include "rocfft.h"
+#include "rocfft/rocfft.h"
 
 // -------------------------------------
 // default load callback definitions
@@ -82,6 +83,15 @@ struct is_hip_complex<rocfft_complex<double>>
     static const bool value = true;
 };
 
+// ASAN introduces some problems with mixing library and client
+// callbacks, so skip these tests if it's enabled
+#ifdef ADDRESS_SANITIZER
+#define TEST_CALLBACK_CHECK_ASAN \
+    GTEST_SKIP() << "mixed library/client callbacks not supported for ASAN";
+#else
+#define TEST_CALLBACK_CHECK_ASAN
+#endif
+
 // -------------------------------------
 // test callbacks struct
 // -------------------------------------
@@ -241,8 +251,8 @@ struct Test_Callback
             ASSERT_EQ(hipMalloc(&work_buffer, work_buffer_size), hipSuccess);
         }
 
-        hipStream_t stream = nullptr;
-        ASSERT_EQ(hipStreamCreate(&stream), hipSuccess);
+        hipStream_wrapper_t stream;
+        stream.alloc();
 
         rocfft_execution_info info;
         ASSERT_EQ(rocfft_execution_info_create(&info), rocfft_status_success);
@@ -401,6 +411,7 @@ struct Test_Callback
 
 TEST(rocfft_UnitTest, default_load_callback_complex_single)
 {
+    TEST_CALLBACK_CHECK_ASAN;
     Test_Callback test(256,
                        1,
                        rocfft_transform_type_complex_forward,
@@ -411,6 +422,7 @@ TEST(rocfft_UnitTest, default_load_callback_complex_single)
 
 TEST(rocfft_UnitTest, default_load_callback_complex_double)
 {
+    TEST_CALLBACK_CHECK_ASAN;
     Test_Callback test(512,
                        1,
                        rocfft_transform_type_complex_forward,
@@ -421,6 +433,7 @@ TEST(rocfft_UnitTest, default_load_callback_complex_double)
 
 TEST(rocfft_UnitTest, default_load_callback_real_single)
 {
+    TEST_CALLBACK_CHECK_ASAN;
     Test_Callback test(1024,
                        1,
                        rocfft_transform_type_real_forward,
@@ -431,6 +444,7 @@ TEST(rocfft_UnitTest, default_load_callback_real_single)
 
 TEST(rocfft_UnitTest, default_load_callback_real_double)
 {
+    TEST_CALLBACK_CHECK_ASAN;
     Test_Callback test(2048,
                        1,
                        rocfft_transform_type_real_forward,
@@ -446,6 +460,7 @@ TEST(rocfft_UnitTest, default_load_callback_real_double)
 
 TEST(rocfft_UnitTest, default_store_callback_complex_single)
 {
+    TEST_CALLBACK_CHECK_ASAN;
     Test_Callback test(256,
                        1,
                        rocfft_transform_type_complex_forward,
@@ -456,6 +471,7 @@ TEST(rocfft_UnitTest, default_store_callback_complex_single)
 
 TEST(rocfft_UnitTest, default_store_callback_complex_double)
 {
+    TEST_CALLBACK_CHECK_ASAN;
     Test_Callback test(512,
                        1,
                        rocfft_transform_type_complex_forward,
@@ -466,6 +482,7 @@ TEST(rocfft_UnitTest, default_store_callback_complex_double)
 
 TEST(rocfft_UnitTest, default_store_callback_real_single)
 {
+    TEST_CALLBACK_CHECK_ASAN;
     Test_Callback test(1024,
                        1,
                        rocfft_transform_type_real_forward,
@@ -476,6 +493,7 @@ TEST(rocfft_UnitTest, default_store_callback_real_single)
 
 TEST(rocfft_UnitTest, default_store_callback_real_double)
 {
+    TEST_CALLBACK_CHECK_ASAN;
     Test_Callback test(2048,
                        1,
                        rocfft_transform_type_real_forward,
