@@ -18,12 +18,13 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 // THE SOFTWARE.
 
+#include "../../shared/accuracy_test.h"
 #include "../../shared/arithmetic.h"
 #include "../../shared/gpubuf.h"
+#include "../../shared/hip_object_wrapper.h"
+#include "../../shared/rocfft_against_fftw.h"
 #include "../../shared/rocfft_params.h"
-#include "accuracy_test.h"
-#include "rocfft.h"
-#include "rocfft_against_fftw.h"
+#include "rocfft/rocfft.h"
 #include <gtest/gtest.h>
 #include <hip/hip_runtime.h>
 #include <memory>
@@ -268,11 +269,8 @@ static void compare_data(const std::vector<rocfft_complex<float>>& original_host
     EXPECT_LT(diff.l_inf / input_norm.l_inf, log2(original_host_data.size()) * MAX_TRANSFORM_ERROR);
 }
 
-TEST(rocfft_UnitTest, DISABLED_hipGraph_execution)
+TEST(rocfft_UnitTest, hipGraph_execution)
 {
-    hipStream_t stream       = nullptr;
-    hipStream_t other_stream = nullptr;
-
     hipGraph_t     graph      = nullptr;
     hipGraphExec_t graph_exec = nullptr;
 
@@ -314,8 +312,10 @@ TEST(rocfft_UnitTest, DISABLED_hipGraph_execution)
 
     EXPECT_EQ(hipDeviceSynchronize(), hipSuccess);
 
-    ASSERT_EQ(hipStreamCreate(&stream), hipSuccess);
-    ASSERT_EQ(hipStreamCreate(&other_stream), hipSuccess);
+    hipStream_wrapper_t stream;
+    hipStream_wrapper_t other_stream;
+    stream.alloc();
+    other_stream.alloc();
 
     ASSERT_EQ(hipStreamBeginCapture(stream, hipStreamCaptureModeGlobal), hipSuccess);
 
