@@ -18,11 +18,12 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 // THE SOFTWARE.
 
+#include "../../shared/accuracy_test.h"
 #include "../../shared/gpubuf.h"
+#include "../../shared/hip_object_wrapper.h"
+#include "../../shared/rocfft_against_fftw.h"
 #include "../../shared/rocfft_params.h"
-#include "accuracy_test.h"
-#include "rocfft.h"
-#include "rocfft_against_fftw.h"
+#include "rocfft/rocfft.h"
 #include <gtest/gtest.h>
 #include <hip/hip_runtime.h>
 #include <memory>
@@ -86,12 +87,11 @@ struct Test_Transform
     Test_Transform(const Test_Transform&) = delete;
     void operator=(const Test_Transform&) = delete;
     Test_Transform(Test_Transform&& other)
-        : stream(other.stream)
+        : stream(std::move(other.stream))
         , work_buffer(other.work_buffer)
         , device_mem_in(std::move(other.device_mem_in))
         , device_mem_out(std::move(other.device_mem_out))
     {
-        other.stream      = nullptr;
         other.work_buffer = nullptr;
         host_mem_in.swap(other.host_mem_in);
         host_mem_out.swap(other.host_mem_out);
@@ -130,7 +130,7 @@ struct Test_Transform
             ASSERT_EQ(hipMalloc(&work_buffer, work_buffer_size), hipSuccess);
         }
 
-        ASSERT_EQ(hipStreamCreate(&stream), hipSuccess);
+        stream.alloc();
         rocfft_execution_info info;
         ASSERT_EQ(rocfft_execution_info_create(&info), rocfft_status_success);
         ASSERT_EQ(rocfft_execution_info_set_stream(info, stream), rocfft_status_success);
@@ -172,8 +172,7 @@ struct Test_Transform
         if(stream)
         {
             ASSERT_EQ(hipStreamSynchronize(stream), hipSuccess);
-            ASSERT_EQ(hipStreamDestroy(stream), hipSuccess);
-            stream = nullptr;
+            stream.free();
         }
 
         ASSERT_EQ(hipFree(work_buffer), hipSuccess);
@@ -235,10 +234,10 @@ struct Test_Transform
     {
         do_cleanup();
     }
-    size_t                             N                = 0;
-    size_t                             dim              = 0;
-    uint32_t                           seed             = 0;
-    hipStream_t                        stream           = nullptr;
+    size_t                             N    = 0;
+    size_t                             dim  = 0;
+    uint32_t                           seed = 0;
+    hipStream_wrapper_t                stream;
     rocfft_plan                        plan             = nullptr;
     rocfft_plan                        plan_inv         = nullptr;
     size_t                             work_buffer_size = 0;
@@ -321,17 +320,17 @@ static void multistream_transform(size_t N, size_t dim, size_t num_streams)
 
 // pick arbitrary sizes here to get some parallelism while still
 // fitting into e.g. 8 GB of GPU memory
-TEST(rocfft_UnitTest, simple_multithread_1D)
+TEST(DISABLED_rocfft_UnitTest, simple_multithread_1D)
 {
     multithread_transform(1048576, 1, 64);
 }
 
-TEST(rocfft_UnitTest, simple_multithread_2D)
+TEST(DISABLED_rocfft_UnitTest, simple_multithread_2D)
 {
     multithread_transform(1024, 2, 64);
 }
 
-TEST(rocfft_UnitTest, simple_multithread_3D)
+TEST(DISABLED_rocfft_UnitTest, simple_multithread_3D)
 {
     multithread_transform(128, 3, 40);
 }
