@@ -18,14 +18,16 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 // THE SOFTWARE.
 
-#include "rocfft.h"
+#include "rocfft/rocfft.h"
 
+#include "../../shared/concurrency.h"
 #include "../../shared/environment.h"
 #include "../../shared/gpubuf.h"
 #include "../../shared/rocfft_complex.h"
 #include "hip/hip_runtime_api.h"
 #include <boost/scope_exit.hpp>
 #include <condition_variable>
+#include <cstdlib>
 #include <fstream>
 #include <gtest/gtest.h>
 #include <mutex>
@@ -33,6 +35,10 @@
 #include <thread>
 #include <vector>
 
+#ifdef _OPENMP
+#include <omp.h>
+#endif
+
 #if __has_include(<filesystem>)
 #include <filesystem>
 #else
@@ -95,6 +101,96 @@ TEST(rocfft_UnitTest, plan_description)
     ASSERT_TRUE(rocfft_status_success == rocfft_plan_destroy(plan));
 }
 
+TEST(rocfft_UnitTest, plan_description_reuse)
+{
+    // check that a plan description can be reused between different
+    // plans, with different layout parameters for each.
+
+    // allocate plan description once
+    rocfft_plan_description desc = nullptr;
+    ASSERT_EQ(rocfft_plan_description_create(&desc), rocfft_status_success);
+
+    std::vector<rocfft_complex<float>> output;
+
+    // do length-8 FFTs with different strides.  first one is
+    // stride-1 and we use that as our baseline to know what output
+    // to expect for the rest
+    const size_t length = 8;
+    for(const size_t stride : {1, 2, 4})
+    {
+        // set layout for this stride
+        ASSERT_EQ(rocfft_plan_description_set_data_layout(desc,
+                                                          rocfft_array_type_complex_interleaved,
+                                                          rocfft_array_type_complex_interleaved,
+                                                          nullptr,
+                                                          nullptr,
+                                                          1,
+                                                          &stride,
+                                                          length * stride,
+                                                          1,
+                                                          &stride,
+                                                          length * stride),
+                  rocfft_status_success);
+
+        static const rocfft_complex<float> input[8]{{-0.100, 0.380},
+                                                    {0.0166, 0.439},
+                                                    {-0.475, 0.212},
+                                                    {0.440, -0.432},
+                                                    {0.445, 0.0589},
+                                                    {0.296, 0.164},
+                                                    {-0.084, 0.077},
+                                                    {0.320, 0.087}};
+
+        // allocate host buffer.  initialize the whole thing to zero
+        // but set a known input along the strides we want
+        std::vector<rocfft_complex<float>> data_host(length * stride, {0.0, 0.0});
+        for(size_t i = 0; i < length; ++i)
+        {
+            data_host[i * stride] = input[i];
+        }
+
+        // copy to device
+        const size_t data_bytes = data_host.size() * sizeof(rocfft_complex<float>);
+        gpubuf_t<rocfft_complex<float>> data_dev;
+        ASSERT_EQ(data_dev.alloc(data_bytes), hipSuccess);
+        void* data_dev_ptr = data_dev.data();
+        ASSERT_EQ(hipMemcpy(data_dev_ptr, data_host.data(), data_bytes, hipMemcpyHostToDevice),
+                  hipSuccess);
+
+        // do the transform
+        rocfft_plan plan = nullptr;
+        ASSERT_EQ(rocfft_plan_create(&plan,
+                                     rocfft_placement_inplace,
+                                     rocfft_transform_type_complex_forward,
+                                     rocfft_precision_single,
+                                     1,
+                                     &length,
+                                     1,
+                                     desc),
+                  rocfft_status_success);
+        ASSERT_EQ(rocfft_execute(plan, &data_dev_ptr, nullptr, nullptr), rocfft_status_success);
+        ASSERT_EQ(hipMemcpy(data_host.data(), data_dev_ptr, data_bytes, hipMemcpyDeviceToHost),
+                  hipSuccess);
+        ASSERT_EQ(hipDeviceSynchronize(), hipSuccess);
+
+        // save output for reference on first run
+        if(output.empty())
+        {
+            output = data_host;
+        }
+        else
+        {
+            // check that the output matches output from the first
+            // (stride-1) run.
+            for(size_t i = 0; i < length; ++i)
+                ASSERT_EQ(data_host[i * stride], output[i]);
+        }
+        ASSERT_EQ(rocfft_plan_destroy(plan), rocfft_status_success);
+    }
+
+    ASSERT_EQ(rocfft_plan_description_destroy(desc), rocfft_status_success);
+}
+
 // Check whether logs can be emitted from multiple threads properly
 TEST(rocfft_UnitTest, log_multithreading)
 {
@@ -248,7 +344,6 @@ TEST(rocfft_UnitTest, workmem_null)
     workmem_test([](size_t requested) { return requested; }, rocfft_status_success, true);
 }
 
-#ifdef ROCFFT_RUNTIME_COMPILE
 static const size_t RTC_PROBLEM_SIZE = 2304;
 // runtime compilation cache tests
 TEST(rocfft_UnitTest, rtc_cache)
@@ -420,8 +515,8 @@ TEST(rocfft_UnitTest, rtc_helper_crash)
     fs::path test_exe    = filename;
     fs::path crasher_exe = test_exe.replace_filename("rtc_helper_crash.exe");
 #else
-    fs::path test_exe    = program_invocation_name;
-    fs::path crasher_exe = test_exe.replace_filename("rtc_helper_crash");
+    fs::path           test_exe     = program_invocation_name;
+    fs::path           crasher_exe  = test_exe.replace_filename("rtc_helper_crash");
 #endif
 
     // use the crashing helper
@@ -477,4 +572,138 @@ TEST(rocfft_UnitTest, rtc_helper_crash)
     plan = nullptr;
 }
 
+TEST(rocfft_UnitTest, rtc_test_harness)
+{
+    // check that hipcc is available since this test requires it
+    //
+    // NOTE: using system() for launching subprocesses for simplicity
+    // and portability
+#ifdef WIN32
+    static const char* test_command = "amdclang++ --version > NUL";
+#else
+    static const char* test_command = "amdclang++ --version > /dev/null";
 #endif
+    if(std::system(test_command) != 0)
+        GTEST_SKIP();
+
+    rocfft_cleanup();
+
+    BOOST_SCOPE_EXIT_ALL()
+    {
+        // reinit rocFFT so caching goes back to normal
+        rocfft_cleanup();
+        rocfft_setup();
+    };
+
+    // extra scope to control lifetime of env vars
+    {
+        // rtc test harness writes to system's temp directory
+        auto tmp_path = fs::temp_directory_path();
+
+        // activate writing of rtc test harnesses
+        EnvironmentSetTemp env_harness("ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS", "1");
+
+        // set path for writing rtc test harnesses source files
+        EnvironmentSetTemp env_harness_path("ROCFFT_DEBUG_KERNEL_HARNESS_PATH",
+                                            tmp_path.string().c_str());
+
+        // ensure every kernel gets compiled once
+        EnvironmentSetTemp env_cache("ROCFFT_RTC_CACHE_PATH", ":memory:");
+        EnvironmentSetTemp env_sys_cache("ROCFFT_RTC_SYS_CACHE_PATH", ":memory:");
+
+        rocfft_setup();
+
+        // ensure stale files from previous runs of this test won't cause
+        // problems - clean up any rocfft_kernel_harness_*.cpp files that
+        // might be left behind
+        for(const auto& entry : std::filesystem::directory_iterator{tmp_path})
+        {
+            auto filename = entry.path().filename();
+            if(filename.string().compare(0, 22, "rocfft_kernel_harness_") == 0
+               && filename.extension().string() == ".cpp")
+                fs::remove(entry);
+        }
+
+        // construct a few different types of plans to try to get all
+        // different kernels compiled
+
+        auto create_destroy_plan
+            = [](rocfft_transform_type type, const size_t dim, const size_t* lengths) -> void {
+            rocfft_plan plan = nullptr;
+            ASSERT_EQ(rocfft_plan_create(&plan,
+                                         rocfft_placement_inplace,
+                                         type,
+                                         rocfft_precision_single,
+                                         dim,
+                                         lengths,
+                                         1,
+                                         nullptr),
+                      rocfft_status_success);
+            ASSERT_EQ(rocfft_plan_destroy(plan), rocfft_status_success);
+            plan = nullptr;
+        };
+        // large 1D R2C + C2R
+        const size_t L1D_PROBLEM_SIZE[1] = {16384};
+        create_destroy_plan(rocfft_transform_type_real_forward, 1, L1D_PROBLEM_SIZE);
+        create_destroy_plan(rocfft_transform_type_real_inverse, 1, L1D_PROBLEM_SIZE);
+
+        // small bluestein R2C + C2R (also covers odd length)
+        const size_t SMALL_BLUESTEIN_PROBLEM_SIZE[1] = {37};
+        create_destroy_plan(rocfft_transform_type_real_forward, 1, SMALL_BLUESTEIN_PROBLEM_SIZE);
+        create_destroy_plan(rocfft_transform_type_real_inverse, 1, SMALL_BLUESTEIN_PROBLEM_SIZE);
+
+        // large bluestein C2C
+        const size_t LARGE_BLUESTEIN_PROBLEM_SIZE[1] = {8191};
+        create_destroy_plan(rocfft_transform_type_complex_forward, 1, LARGE_BLUESTEIN_PROBLEM_SIZE);
+
+        // L1D_TRTRT
+        const size_t L1D_TRTRT_PROBLEM_SIZE[1] = {680};
+        create_destroy_plan(rocfft_transform_type_complex_forward, 1, L1D_TRTRT_PROBLEM_SIZE);
+
+        // small 3D (exercises 2D_SINGLE)
+        const size_t SMALL_3D_PROBLEM_SIZE[3] = {25, 25, 25};
+        create_destroy_plan(rocfft_transform_type_complex_forward, 3, SMALL_3D_PROBLEM_SIZE);
+
+        // larger 3D
+        const size_t LARGE_3D_PROBLEM_SIZE[3] = {200, 200, 200};
+        create_destroy_plan(rocfft_transform_type_complex_forward, 3, LARGE_3D_PROBLEM_SIZE);
+
+        // now try to compile each file - they'd need hand-editing to test
+        // something useful, but we can at least ensure they build.
+
+        // enumerate all the files
+        std::vector<std::pair<std::string, int>> files;
+        size_t                                   i = 0;
+        for(;; ++i)
+        {
+            // construct name of main file
+            fs::path main_file = tmp_path / ("rocfft_kernel_harness_" + std::to_string(i) + ".cpp");
+
+            if(!fs::exists(main_file))
+                break;
+
+            files.emplace_back(main_file.string(), -1);
+        }
+
+        // we should have generated at least a few kernels
+        ASSERT_FALSE(files.empty());
+
+#ifdef _OPENMP
+#pragma omp parallel for num_threads(rocfft_concurrency())
+#endif
+        for(i = 0; i < files.size(); ++i)
+        {
+#ifdef WIN32
+            const std::string command = "amdclang++ -x hip -c -std=c++17 -o NUL " + files[i].first;
+#else
+            const std::string command
+                = "amdclang++ -x hip -c -std=c++17 -o /dev/null " + files[i].first;
+#endif
+            files[i].second = std::system(command.c_str());
+        }
+
+        // check that all compiles succeeded
+        for(const auto& file : files)
+            ASSERT_EQ(file.second, 0);
+    }
+}
