@@ -190,10 +190,10 @@ install_packages( )
     local library_dependencies_fedora=( "make" "cmake" "gcc-c++" "libcxx-devel" "rpm-build" )
     local library_dependencies_sles=( "make" "cmake" "gcc-c++" "gcc-fortran" "libcxxtools9" "rpm-build" )
 
-    local client_dependencies_ubuntu=( "libfftw3-dev" "libboost-program-options-dev" )
+    local client_dependencies_ubuntu=( "libfftw3-dev")
     local client_dependencies_centos=( "fftw-devel" "boost-devel" )
     local client_dependencies_fedora=( "fftw-devel" "boost-devel" )
-    local client_dependencies_sles=( "fftw3-devel" "libboost_program_options1_66_0-devel" "pkg-config" "dpkg")
+    local client_dependencies_sles=( "fftw3-devel" "pkg-config" "dpkg")
 
     case "${ID}" in
         ubuntu)
@@ -279,7 +279,7 @@ group_num=false
 manual_small_arg=false
 manual_large_arg=false
 build_address_sanitizer=false
-build_freorg_bkwdcomp=true
+build_freorg_bkwdcomp=false
 solmap_data_folder=false
 
 # #################################################
@@ -432,7 +432,7 @@ pushd .
 # #################################################
 # configure & build
 # #################################################
-cmake_common_options=""
+cmake_common_options="--toolchain=toolchain-linux.cmake"
 cmake_client_options=""
 
 if [[ "${build_address_sanitizer}" == true ]]; then
@@ -484,23 +484,21 @@ fi
 
 # clients
 if [[ "${build_clients}" == true ]]; then
-    cmake_client_options="${cmake_client_options} -DBUILD_CLIENTS_SAMPLES=ON -DBUILD_CLIENTS_TESTS=ON -DBUILD_CLIENTS_RIDER=ON"
+    cmake_client_options="${cmake_client_options} -DBUILD_CLIENTS_SAMPLES=ON -DBUILD_CLIENTS_TESTS=ON -DBUILD_CLIENTS_BENCH=ON"
 fi
 
-compiler="hipcc"
-
 # Build library with AMD toolchain because of existense of device kernels
 if [[ "${build_clients}" == false ]]; then
     cmake_client_options=" "
 fi
 if [[ "${build_relocatable}" == true ]]; then
-    CXX=${compiler} CC=${compiler} ${cmake_executable} ${cmake_common_options} ${cmake_client_options} -DCPACK_SET_DESTDIR=OFF -DCMAKE_INSTALL_PREFIX="${install_prefix}" -DCPACK_PACKAGING_INSTALL_PREFIX="${rocm_path}" \
-       -DCMAKE_PREFIX_PATH="${rocm_path} ${rocm_path}/hipcc ${rocm_path}/hip" \
+    ${cmake_executable} ${cmake_common_options} ${cmake_client_options} -DCPACK_SET_DESTDIR=OFF -DCMAKE_INSTALL_PREFIX="${install_prefix}" -DCPACK_PACKAGING_INSTALL_PREFIX="${rocm_path}" \
+       -DCMAKE_PREFIX_PATH="${rocm_path} ${rocm_path}/amdclang++ ${rocm_path}/hip" \
        -DCMAKE_SHARED_LINKER_FLAGS="${rocm_rpath}" \
        -DROCM_DISABLE_LDCONFIG=ON \
        ../..
 else
-    CXX=${compiler} CC=${compiler} ${cmake_executable} ${cmake_common_options} ${cmake_client_options} -DCPACK_SET_DESTDIR=OFF -DCMAKE_INSTALL_PREFIX=${install_prefix} -DCPACK_PACKAGING_INSTALL_PREFIX=/opt/rocm ../..
+    ${cmake_executable} ${cmake_common_options} ${cmake_client_options} -DCPACK_SET_DESTDIR=OFF -DCMAKE_INSTALL_PREFIX=${install_prefix} -DCPACK_PACKAGING_INSTALL_PREFIX=/opt/rocm ../..
 fi
 check_exit_code
 make -j$(nproc)
