@@ -111,9 +111,25 @@ void PlacementTrace::Backtracking(ExecPlan& execPlan, int execSeqID)
     node->inArrayType  = this->iType;
     node->outArrayType = this->oType;
 
-    // correct array type of callback, since it could be marked as CI during the process
-    if(node->scheme == CS_KERNEL_APPLY_CALLBACK)
-        node->outArrayType = node->inArrayType = rocfft_array_type_real;
+    // Even-length internal nodes have real data for input or output but
+    // child nodes treat it as complex interleaved
+    auto realEvenAncestor = node->GetRealEvenAncestor();
+    if(realEvenAncestor)
+    {
+        // forward transform, first node (if it's a leaf) needs to treat real input as complex
+        if(node->direction == -1 && node == realEvenAncestor->GetFirstLeaf())
+            node->inArrayType = rocfft_array_type_complex_interleaved;
+        // inverse transform, last node (if it's a leaf) needs to treat real output as complex
+        if(node->direction == 1 && node == realEvenAncestor->GetLastLeaf())
+            node->outArrayType = rocfft_array_type_complex_interleaved;
+    }
+
+    // Ensure that all nodes that take real input are declared as
+    // such.  This is particularly important for leaf nodes, since
+    // kernelio debugging depends on knowing the correct type of the
+    // array to print.
+    if(node->scheme == CS_KERNEL_COPY_R_TO_CMPLX)
+        node->inArrayType = rocfft_array_type_real;
 
     // for nodes that uses bluestein buffer
     auto setBluesteinOffset = [node](size_t& offset) {
@@ -133,8 +149,12 @@ void PlacementTrace::Backtracking(ExecPlan& execPlan, int execSeqID)
     };
     if(node->obIn == OB_TEMP_BLUESTEIN)
         setBluesteinOffset(node->iOffset);
+    else
+        node->iOffset = 0;
     if(node->obOut == OB_TEMP_BLUESTEIN)
         setBluesteinOffset(node->oOffset);
+    else
+        node->oOffset = 0;
 
     // keep going backward to next node, skipping over chirp setup nodes
     int nextExecSeqID = execSeqID;
@@ -148,9 +168,11 @@ void PlacementTrace::Backtracking(ExecPlan& execPlan, int execSeqID)
             return;
         }
     }
-    // if we're here, then 'node' must have been the first node, or
-    // preceded only by chirp setup nodes.  use root input type.
-    node->inArrayType = execPlan.rootPlan->inArrayType;
+
+    // if we're here, then 'node' must have already been the first
+    // node (and had its input set properly), or it's preceded only
+    // by chirp setup nodes.  input type we set earlier is already
+    // right.
 }
 
 // test if rootArrayType == testArrayType,
@@ -171,31 +193,6 @@ bool AssignmentPolicy::EquivalentArrayType(rocfft_array_type rootAryType,
     return (testAryType == rootAryType);
 }
 
-bool AssignmentPolicy::BufferIsUnitStride(const ExecPlan& execPlan, OperatingBuffer buf)
-{
-    // temp buffers are unit stride
-    if(buf != OB_USER_IN && buf != OB_USER_OUT)
-        return true;
-    auto stride = (buf == OB_USER_IN) ? execPlan.rootPlan->inStride : execPlan.rootPlan->outStride;
-    auto length = (buf == OB_USER_IN) ? execPlan.iLength : execPlan.oLength;
-    auto dist   = (buf == OB_USER_IN) ? execPlan.rootPlan->iDist : execPlan.rootPlan->oDist;
-    size_t curStride = 1;
-    do
-    {
-        if(stride.front() != curStride)
-            return false;
-        curStride *= length.front();
-        stride.erase(stride.begin());
-        length.erase(length.begin());
-    } while(!stride.empty());
-
-    // NB: users may input incorrect i/o-dist value for inplace transform
-    //     however, when the batch-size is 1, we can simply make it permissive
-    //     since the dist is not used in single batch. But note that we still need
-    //     to pass the above do-while to ensure all the previous strides are valid.
-    return (execPlan.rootPlan->batch == 1) || (curStride == dist);
-}
-
 // return true if OB_TEMP_BLUESTEIN is a valid output buffer for the node
 static bool ValidOutBufferBluestein(TreeNode& node)
 {
@@ -281,14 +278,7 @@ bool AssignmentPolicy::ValidOutBuffer(ExecPlan&           execPlan,
         if(BufferIsUnitStride(execPlan, buffer))
         {
             // just check if there's enough space
-            return std::accumulate(nodeLen.begin(),
-                                   nodeLen.end(),
-                                   static_cast<size_t>(1),
-                                   std::multiplies<size_t>())
-                   <= std::accumulate(bufLen.begin(),
-                                      bufLen.end(),
-                                      static_cast<size_t>(1),
-                                      std::multiplies<size_t>());
+            return product(nodeLen.begin(), nodeLen.end()) <= product(bufLen.begin(), bufLen.end());
         }
 
         // ensure that the node's dimensions fit exactly into the
@@ -324,11 +314,6 @@ bool AssignmentPolicy::ValidOutBuffer(ExecPlan&           execPlan,
     {
         test_result = false;
     }
-    // the only restriction of APPLY_CALLBACK is that it needs to be inplace
-    else if(node.scheme == CS_KERNEL_APPLY_CALLBACK)
-    {
-        test_result = true;
-    }
     // bluestein nodes must write to temp bluestein buffer
     else if(buffer == OB_TEMP_BLUESTEIN)
     {
@@ -353,20 +338,18 @@ bool AssignmentPolicy::ValidOutBuffer(ExecPlan&           execPlan,
     {
         auto fitArrayType = buffer == OB_USER_IN ? execPlan.rootPlan->inArrayType
                                                  : execPlan.rootPlan->outArrayType;
-        if(dataFits(node, buffer) && EquivalentArrayType(fitArrayType, arrayType))
+        test_result       = dataFits(node, buffer) && EquivalentArrayType(fitArrayType, arrayType);
+        if(!test_result)
         {
-            test_result = true;
+            // if that didn't fit, and we're writing to OB_USER_OUT, and
+            // this is an in-place R2C/C2R transform, then we could also
+            // try fitting into the shape of the input.
+            if(buffer == OB_USER_OUT && execPlan.rootPlan->placement == rocfft_placement_inplace
+               && (execPlan.rootPlan->inArrayType == rocfft_array_type_real
+                   || execPlan.rootPlan->outArrayType == rocfft_array_type_real))
+                test_result = dataFits(node, OB_USER_IN)
+                              && EquivalentArrayType(execPlan.rootPlan->inArrayType, arrayType);
         }
-        // if that didn't fit, and we're writing to OB_USER_OUT, and
-        // this is an in-place R2C/C2R transform, then we could also
-        // try fitting into the shape of the input.
-        else if(buffer == OB_USER_OUT && execPlan.rootPlan->placement == rocfft_placement_inplace
-                && (execPlan.rootPlan->inArrayType == rocfft_array_type_real
-                    || execPlan.rootPlan->outArrayType == rocfft_array_type_real))
-            test_result = dataFits(node, OB_USER_IN)
-                          && EquivalentArrayType(execPlan.rootPlan->inArrayType, arrayType);
-        else
-            test_result = false;
     }
 
     node_buf_test_cache[cacheMapKey] = test_result;
@@ -605,14 +588,14 @@ void AssignmentPolicy::AssignBuffers_internal(ExecPlan& execPlan)
     availableArrayTypes.clear();
     node_buf_test_cache.clear();
 
-    // Start from a minimal requirement; // in, out buffer
-    availableBuffers.insert(execPlan.rootPlan->obIn);
+    // Start from a minimal requirement; // out buffer
     availableBuffers.insert(execPlan.rootPlan->obOut);
-    if(execPlan.rootPlan->IsRootPlanC2CTransform())
+
+    // real-inverse transforms are allowed to modify input
+    if(execPlan.rootPlan->direction == 1
+       && execPlan.rootPlan->outArrayType == rocfft_array_type_real)
     {
-        // For real-transform, USER_IN is always allowed to be modified.
-        // For c2c-transform, we should keep USER_IN read-only. So remove it if exists.
-        availableBuffers.erase(OB_USER_IN);
+        availableBuffers.insert(execPlan.rootPlan->obIn);
     }
 
     // Insert the valid ArrayTypes to use internally
@@ -1248,6 +1231,10 @@ void AssignmentPolicy::PadPlan(ExecPlan& execPlan)
         return;
 
     RecursiveTraverse(execPlan.rootPlan.get(), [&execPlan](TreeNode* n) {
+        // Skip nodes that have partial passes
+        if(n->applyPartialPass)
+            return;
+
         // Look for nodes that begin writing to a new temp buffer
         // (i.e. obOut is a paddable temp buffer, and obIn was a
         // different buffer)
@@ -1334,12 +1321,11 @@ void AssignmentPolicy::PadPlan(ExecPlan& execPlan)
                     return;
             }
 
-            // R2C/C2R changes length, which confuses padding logic
-            // if it's in-place.  So just ignore those cases.
+            // standalone real-complex kernels can't handle arbitrary
+            // strides at the moment, so exclude them from padding
             if(std::any_of(users.begin(), users.end(), [](TempBufOp& op) {
                    return (op.node.scheme == CS_KERNEL_R_TO_CMPLX
-                           || op.node.scheme == CS_KERNEL_CMPLX_TO_R)
-                          && op.node.obIn == op.node.obOut;
+                           || op.node.scheme == CS_KERNEL_CMPLX_TO_R);
                }))
                 return;
 
