@@ -20,11 +20,13 @@
 * THE SOFTWARE.
 *******************************************************************************/
 
+#include "../../shared/device_properties.h"
 #include "../../shared/environment.h"
 #include "../../shared/rocfft_hip.h"
 #include "logging.h"
 #include "repo.h"
-#include "rocfft.h"
+#include "rocfft/rocfft.h"
+#include "rocfft_exception.h"
 #include "rocfft_ostream.hpp"
 #include "rtc_cache.h"
 #include "solution_map.h"
@@ -42,6 +44,7 @@ int log_plan_fd     = -1;
 int log_kernelio_fd = -1;
 int log_rtc_fd      = -1;
 int log_tuning_fd   = -1;
+int log_graph_fd    = -1;
 
 /**
  *  @brief Logging function
@@ -79,12 +82,10 @@ static void open_log_stream(const char* environment_variable_name, int& log_fd)
 
 // library setup function, called once in program at the start of library use
 rocfft_status rocfft_setup()
+try
 {
     rocfft_ostream::setup();
-
-#ifdef ROCFFT_RUNTIME_COMPILE
     RTCCache::single = std::make_unique<RTCCache>();
-#endif
 
     // set layer_mode from value of environment variable ROCFFT_LAYER
     auto str_layer_mode = rocfft_getenv("ROCFFT_LAYER");
@@ -122,27 +123,47 @@ rocfft_status rocfft_setup()
         // open log_tuning file
         if(layer_mode & rocfft_layer_mode_log_tuning)
             open_log_stream("ROCFFT_LOG_TUNING_PATH", log_tuning_fd);
+
+        // open log_graph file
+        if(layer_mode & rocfft_layer_mode_log_graph)
+            open_log_stream("ROCFFT_LOG_GRAPH_PATH", log_graph_fd);
     }
 
     // setup solution map once in program at the start of library use
-    solution_map::get_solution_map().setup();
+    int devcount = 0;
+    if(hipGetDeviceCount(&devcount) != hipSuccess)
+    {
+        return rocfft_status_failure;
+    }
+    if(devcount == 0)
+    {
+        return rocfft_status_failure;
+    }
+    auto arch_name = get_arch_name(get_curr_device_prop());
+    solution_map::get_solution_map().setup(arch_name);
     TuningBenchmarker::GetSingleton().Setup();
 
     log_trace(__func__);
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 // library cleanup function, called once in program after end of library use
 rocfft_status rocfft_cleanup()
+try
 {
-    log_trace(__func__);
+    // Logging is potentially unsafe if we're in the middle of static
+    // deinitialization, as log structures might have already been
+    // cleaned up.
+    // log_trace(__func__);
 
     // close the RTC cache and clear the repo, so that subsequent
     // rocfft_setup() + plan creation will start from scratch
     Repo::Clear();
-#ifdef ROCFFT_RUNTIME_COMPILE
     RTCCache::single.reset();
-#endif
 
     TuningBenchmarker::GetSingleton().Clean();
 
@@ -183,18 +204,32 @@ rocfft_status rocfft_cleanup()
         CLOSE(log_tuning_fd);
         log_tuning_fd = -1;
     }
+    if(log_graph_fd != -1)
+    {
+        CLOSE(log_graph_fd);
+        log_graph_fd = -1;
+    }
 
     // stop all log worker threads
     rocfft_ostream::cleanup();
 
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 #ifdef ROCFFT_BUILD_OFFLINE_TUNER
 rocfft_status rocfft_get_offline_tuner_handle(void** offline_tuner)
+try
 {
     TuningBenchmarker::GetSingleton().SetBindingSolutionMap(&solution_map::get_solution_map());
     *offline_tuner = &(TuningBenchmarker::GetSingleton());
     return rocfft_status_success;
 }
-#endif
\ No newline at end of file
+catch(...)
+{
+    return rocfft_handle_exception();
+}
+#endif
