@@ -23,9 +23,9 @@
 #include "function_pool.h"
 #include "node_factory.h"
 
-size_t TransformsPerThreadblock(const size_t len, rocfft_precision precision)
+size_t TransformsPerThreadblock(const FMKey& kernel_key)
 {
-    return function_pool::get_kernel(fpkey(len, precision)).transforms_per_block;
+    return function_pool::get_kernel(kernel_key).transforms_per_block;
 }
 
 bool canOptimizeWithStride(TreeNode* stockham)
@@ -34,7 +34,7 @@ bool canOptimizeWithStride(TreeNode* stockham)
     // diagonal transpose
     if(IsPo2(stockham->length[0]) && stockham->length.size() >= 3)
         return false;
-    size_t numTrans = TransformsPerThreadblock(stockham->length[0], stockham->precision);
+    size_t numTrans = TransformsPerThreadblock(stockham->GetKernelKey());
 
     // ensure we are doing enough rows to coalesce properly. 4
     // seems to be enough for double-precision, whereas some
@@ -152,7 +152,8 @@ std::unique_ptr<TreeNode> TRFuseShim::FuseKernels()
     auto fused = NodeFactory::CreateNodeFromScheme(stockham->scheme, stockham->parent);
     fused->CopyNodeData(*stockham);
     // actually no need to check kernel exists, we already have the kernel with the same length/scheme
-    if(!fused->KernelCheck())
+    std::vector<FMKey> kernelkey = {stockham->GetKernelKey()};
+    if(!fused->KernelCheck(kernelkey))
         return nullptr;
 
     fused->placement   = rocfft_placement_notinplace;
@@ -234,7 +235,8 @@ std::unique_ptr<TreeNode> RTFuseShim::FuseKernels()
     auto fused = NodeFactory::CreateNodeFromScheme(stockham->scheme, stockham->parent);
     fused->CopyNodeData(*stockham);
     // actually no need to check kernel exists, we already have the kernel with the same length/scheme
-    if(!fused->KernelCheck())
+    std::vector<FMKey> kernelkey = {stockham->GetKernelKey()};
+    if(!fused->KernelCheck(kernelkey))
         return nullptr;
 
     fused->placement    = rocfft_placement_notinplace;
@@ -507,27 +509,26 @@ bool TransC2R_FuseShim::CheckSchemeFusable()
     // since it's possible we'll change its input buffer
     nodes.resize(3);
     nodes[1] = realEven->childNodes.front().get();
-    nodes[2] = realEven->childNodes[1]->GetFirstLeaf();
+    nodes[2] = std::find_if(realEven->childNodes.begin() + 1,
+                            realEven->childNodes.end(),
+                            [](const auto& n) { return !n->IsBluesteinChirpSetup(); })
+                   ->get();
 
     firstFusedNode = 0;
     lastFusedNode  = 1;
 
-    // if the nextLeafNode is stockham, SBCC or PAD-MUL,
+    // if the nextLeafNode is stockham, SBCC or PAD-MUL
     //   we allow the EffectivePlacement of (trans-in, c2r-out) to be inplace,
     //   then we force it to be OP, but change the nextLeafNode's input..
     // So if the nextLeafNode isn't one of these (ex, a transpose for TRTRT)
     //   then we couldn't change tranpose's input buffer
     ComputeScheme nextFFTScheme = nodes[2]->scheme;
     if(nextFFTScheme == CS_KERNEL_STOCKHAM || nextFFTScheme == CS_KERNEL_STOCKHAM_BLOCK_CC
-       || nextFFTScheme == CS_KERNEL_CHIRP)
+       || nextFFTScheme == CS_KERNEL_PAD_MUL)
         allowInplace = true;
     else
         allowInplace = false;
 
-    // if the nextLeaf is chirp, what we actually want is next sibling PAD-MUL
-    if(nextFFTScheme == CS_KERNEL_CHIRP)
-        nodes[2] = nodes[2]->parent->childNodes[1].get();
-
     return true;
 }
 
