@@ -21,7 +21,7 @@
 #ifndef ROCFFT_RTC_H
 #define ROCFFT_RTC_H
 
-#include "rocfft.h"
+#include "rocfft/rocfft.h"
 #include <hip/hip_runtime_api.h>
 
 #include <algorithm>
@@ -31,10 +31,12 @@
 #include <vector>
 
 #include "../../../shared/rocfft_complex.h"
+#include "../device/kernels/callback.h"
 #include "rtc_generator.h"
 
 struct DeviceCallIn;
 class TreeNode;
+class LeafNode;
 struct GridParam;
 
 // Helper class that handles alignment of kernel arguments
@@ -66,9 +68,9 @@ public:
     {
         append(&f, sizeof(float));
     }
-    void append_half(_Float16 f)
+    void append_half(rocfft_fp16 f)
     {
-        append(&f, sizeof(_Float16));
+        append(&f, sizeof(rocfft_fp16));
     }
     template <typename T>
     void append_struct(const T& data)
@@ -112,11 +114,18 @@ struct RTCKernel
     // supported scheme + problem size.  throws runtime_error on
     // error.
     static std::shared_future<std::unique_ptr<RTCKernel>>
-        runtime_compile(const TreeNode&    node,
+        runtime_compile(const LeafNode&    node,
                         const std::string& gpu_arch,
                         std::string&       kernel_name,
                         bool               enable_callbacks = false);
 
+    // take already-compiled code object and prepare to launch the
+    // named kernel
+    RTCKernel(const std::string&       kernel_name,
+              const std::vector<char>& code,
+              dim3                     gridDim  = {},
+              dim3                     blockDim = {});
+
     virtual ~RTCKernel()
     {
         kernel = nullptr;
@@ -131,20 +140,23 @@ struct RTCKernel
     void operator=(const RTCKernel&) = delete;
 
     // normal launch from within rocFFT execution plan
-    void launch(DeviceCallIn& data);
+    void launch(DeviceCallIn& data, const hipDeviceProp_t& deviceProp);
     // direct launch with kernel args
-    void launch(RTCKernelArgs& kargs,
-                dim3           gridDim,
-                dim3           blockDim,
-                unsigned int   lds_bytes,
-                hipStream_t    stream = nullptr);
+    void launch(RTCKernelArgs&         kargs,
+                dim3                   gridDim,
+                dim3                   blockDim,
+                unsigned int           lds_bytes,
+                const hipDeviceProp_t& deviceProp,
+                hipStream_t            stream = nullptr);
 
     // normal launch from within rocFFT execution plan
     bool get_occupancy(dim3 blockDim, unsigned int lds_bytes, int& occupancy);
 
+#ifndef ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS
     // Subclasses implement this - each kernel type has different
     // parameters
     virtual RTCKernelArgs get_launch_args(DeviceCallIn& data) = 0;
+#endif
 
     // function to construct the correct RTCKernel object, given a kernel name and its compiled code
     using rtckernel_construct_t = std::function<std::unique_ptr<RTCKernel>(
@@ -157,13 +169,10 @@ struct RTCKernel
     dim3 gridDim;
     dim3 blockDim;
 
-protected:
-    // protected ctor, use "runtime_compile" to build kernel for a node
-    RTCKernel(const std::string&       kernel_name,
-              const std::vector<char>& code,
-              dim3                     gridDim  = {},
-              dim3                     blockDim = {});
+    std::string kernel_name;
 
+protected:
+#ifndef ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS
     struct RTCGenerator
     {
         kernel_name_gen_t     generate_name;
@@ -185,13 +194,14 @@ protected:
         dim3 gridDim;
         dim3 blockDim;
     };
+#endif
 
     hipModule_t   module = nullptr;
     hipFunction_t kernel = nullptr;
-
-private:
 };
 
+#ifndef ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS
+
 // helper functions to construct pieces of RTC kernel names
 static const char* rtc_array_type_name(rocfft_array_type type)
 {
@@ -225,25 +235,53 @@ static const char* rtc_precision_name(rocfft_precision precision)
     }
 }
 
-static const char* rtc_precision_type_decl(rocfft_precision precision)
+static const char* rtc_precision_type_decl(rocfft_precision precision, bool is_complex = true)
 {
     switch(precision)
     {
     case rocfft_precision_single:
-        return "typedef rocfft_complex<float> scalar_type;\n";
+        return is_complex ? "typedef rocfft_complex<float> scalar_type;\n"
+                          : "typedef float scalar_type;\n";
     case rocfft_precision_double:
-        return "typedef rocfft_complex<double> scalar_type;\n";
+        return is_complex ? "typedef rocfft_complex<double> scalar_type;\n"
+                          : "typedef double scalar_type;\n";
     case rocfft_precision_half:
-        return "typedef rocfft_complex<_Float16> scalar_type;\n";
+        return is_complex ? "typedef rocfft_complex<rocfft_fp16> scalar_type;\n"
+                          : "typedef rocfft_fp16 scalar_type;\n";
     }
 }
 
-static const std::string rtc_const_cbtype_decl(bool enable_callbacks)
+static const char* rtc_cbtype_name(CallbackType cbtype)
 {
-    if(enable_callbacks)
-        return "static const CallbackType cbtype = CallbackType::USER_LOAD_STORE;\n";
-    else
-        return "static const CallbackType cbtype = CallbackType::NONE;\n";
+    switch(cbtype)
+    {
+    case CallbackType::NONE:
+        return "";
+    case CallbackType::USER_LOAD_STORE:
+        return "_CB";
+    case CallbackType::USER_LOAD_STORE_R2C:
+        return "_CBr2c";
+    case CallbackType::USER_LOAD_STORE_C2R:
+        return "_CBc2r";
+    }
 }
 
+// realDataAsComplex is true if we're treating real data as complex
+// (in an even-length real-complex FFT)
+static const std::string rtc_const_cbtype_decl(CallbackType cbtype)
+{
+    switch(cbtype)
+    {
+    case CallbackType::NONE:
+        return "static const CallbackType cbtype = CallbackType::NONE;\n";
+    case CallbackType::USER_LOAD_STORE:
+        return "static const CallbackType cbtype = CallbackType::USER_LOAD_STORE;\n";
+    case CallbackType::USER_LOAD_STORE_R2C:
+        return "static const CallbackType cbtype = CallbackType::USER_LOAD_STORE_R2C;\n";
+    case CallbackType::USER_LOAD_STORE_C2R:
+        return "static const CallbackType cbtype = CallbackType::USER_LOAD_STORE_C2R;\n";
+    }
+}
+#endif
+
 #endif
