@@ -25,7 +25,8 @@
 
 #include "../device/generator/stockham_gen.h"
 #include "compute_scheme.h"
-#include "rocfft.h"
+#include "load_store_ops.h"
+#include "rocfft/rocfft.h"
 #include "rtc_kernel.h"
 
 #include "../device/kernels/common.h"
@@ -47,9 +48,10 @@ std::string stockham_rtc_kernel_name(const StockhamGeneratorSpecs& specs,
                                      DirectRegType                 dir2regMode,
                                      IntrinsicAccessType           intrinsicMode,
                                      SBRC_TRANSPOSE_TYPE           transpose_type,
-                                     bool                          enable_callbacks,
-                                     bool                          enable_scaling,
-                                     BluesteinFuseType             fuseBlue);
+                                     CallbackType                  cbtype,
+                                     BluesteinFuseType             fuseBlue,
+                                     const LoadOps&                loadOps,
+                                     const StoreOps&               storeOps);
 
 // generate source for RTC stockham kernel.  transforms_per_block may
 // be nullptr, but if non-null, stockham_rtc stores the number of
@@ -72,8 +74,9 @@ std::string stockham_rtc(const StockhamGeneratorSpecs& specs,
                          DirectRegType                 dir2regMode,
                          IntrinsicAccessType           intrinsicMode,
                          SBRC_TRANSPOSE_TYPE           transpose_type,
-                         bool                          enable_callbacks,
-                         bool                          enable_scaling,
-                         const BluesteinFuseType&      fuseBlue);
+                         CallbackType                  cbtype,
+                         const BluesteinFuseType&      fuseBlue,
+                         const LoadOps&                loadOps,
+                         const StoreOps&               storeOps);
 
 #endif
