@@ -1,4 +1,4 @@
-// Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2022-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -21,7 +21,8 @@
 #ifndef RTC_TRANSPOSE_GEN
 #define RTC_TRANSPOSE_GEN
 
-#include "rocfft.h"
+#include "load_store_ops.h"
+#include "rocfft/rocfft.h"
 #include "rtc_kernel.h"
 
 #include "../device/kernels/common.h"
@@ -38,8 +39,10 @@ struct TransposeSpecs
     int               largeTwdDirection;
     bool              diagonal;
     bool              tileAligned;
-    bool              enable_callbacks;
-    bool              enable_scaling;
+    CallbackType      cbtype;
+    LoadOps           loadOps;
+    StoreOps          storeOps;
+    bool              grid3D;
 };
 
 // generate name for RTC transpose kernel
