@@ -37,7 +37,7 @@ protected:
         scheme = CS_L1D_TRTRT;
     }
     void AssignParams_internal() override;
-    void BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) override;
+    void BuildTree_internal(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec) override;
 };
 
 /*****************************************************
@@ -54,7 +54,7 @@ protected:
         scheme = CS_L1D_CC;
     }
     void AssignParams_internal() override;
-    void BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) override;
+    void BuildTree_internal(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec) override;
 };
 
 /*****************************************************
@@ -71,7 +71,7 @@ protected:
         scheme = CS_L1D_CRT;
     }
     void AssignParams_internal() override;
-    void BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) override;
+    void BuildTree_internal(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec) override;
 };
 
 /*****************************************************
