@@ -37,7 +37,7 @@ protected:
         scheme = CS_REAL_TRANSFORM_USING_CMPLX;
     }
     void AssignParams_internal() override;
-    void BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) override;
+    void BuildTree_internal(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec) override;
 
 public:
     bool UseOutputLengthForPadding() override
@@ -60,7 +60,7 @@ protected:
         scheme = CS_REAL_TRANSFORM_EVEN;
     }
     void AssignParams_internal() override;
-    void BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) override;
+    void BuildTree_internal(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec) override;
 
 public:
     // 3D Even can possibly set this
@@ -82,7 +82,8 @@ class Real2DEvenNode : public InternalNode
     enum Solution
     {
         INPLACE_SBCC, // SBCC + SBRR with pre/post processing
-        TR_PAIR // TRTR Real2C, or RTRT for C2Real
+        TR_PAIR, // RTRT Real2C, or TRTR for C2Real
+        REAL_2D_SINGLE, // 2D_SINGLE with pre/post processing
     };
 
     bool UseOutputLengthForPadding() override
@@ -97,12 +98,14 @@ protected:
         scheme = CS_REAL_2D_EVEN;
     }
     void     AssignParams_internal() override;
-    void     BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) override;
+    void     BuildTree_internal(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec) override;
     Solution solution = TR_PAIR;
 
-    void BuildTree_internal_SBCC();
-    void BuildTree_internal_TR_pair();
+    void BuildTree_internal_2D_SINGLE();
+    void BuildTree_internal_SBCC(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec);
+    void BuildTree_internal_TR_pair(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec);
 
+    void AssignParams_internal_2D_SINGLE();
     void AssignParams_internal_SBCC();
     void AssignParams_internal_TR_pair();
 };
@@ -116,10 +119,11 @@ class Real3DEvenNode : public InternalNode
 
     enum Solution
     {
-        INPLACE_SBCC, // SBCC + SBCC + SBRR with pre/post processing
+        REAL_2D_SINGLE_SBCC, // FWD: (2D_SINGLE with post-processing) + CC / BWD: CC + (2D_SINGLE with pre-processing)
+        INPLACE_SBCC, // FWD: (RR with post-processing) + CC + CC / BWD: CC + CC + (RR with pre-processing)
         SBCR, // SBCR + SBCR + SBCR with pre-processing for C2Real only
-        SBRC, // SBRC + SBRC + SBRC with post-processing for Real2C only
-        TR_PAIRS // TRTRTR Real2C, or RTRTRT for C2Real
+        SBRC, // SBRC + SBRC + SBRC with post-processing for Real2C only // NOT IMPLEMENTED
+        TR_PAIRS // RTRTRT Real2C, or TRTRTR for C2Real
     };
 
     bool UseOutputLengthForPadding() override
@@ -134,16 +138,18 @@ protected:
         scheme = CS_REAL_3D_EVEN;
     }
     void AssignParams_internal() override;
-    void BuildTree_internal(const SchemeVec& child_schemes = EmptySchemeVec) override;
+    void BuildTree_internal(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec) override;
 
     Solution solution = TR_PAIRS;
 
     void Build_solution();
 
-    void BuildTree_internal_SBCC();
-    void BuildTree_internal_SBCR();
-    void BuildTree_internal_TR_pairs();
+    void BuildTree_internal_2D_SINGLE_CC();
+    void BuildTree_internal_SBCC(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec);
+    void BuildTree_internal_SBCR(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec);
+    void BuildTree_internal_TR_pairs(SchemeTreeVec& child_scheme_trees = EmptySchemeTreeVec);
 
+    void AssignParams_internal_2D_SINGLE_CC();
     void AssignParams_internal_SBCC();
     void AssignParams_internal_SBCR();
     void AssignParams_internal_TR_pairs();
@@ -154,7 +160,6 @@ protected:
  * CS_KERNEL_COPY_HERM_TO_CMPLX
  * CS_KERNEL_COPY_CMPLX_TO_HERM
  * CS_KERNEL_COPY_CMPLX_TO_R
- * CS_KERNEL_APPLY_CALLBACK
  *****************************************************/
 class RealTransDataCopyNode : public LeafNode
 {
@@ -171,11 +176,7 @@ protected:
         /************
         * Placement
         *************/
-        // callback node allows only in-place, others (copy) allow only out-of-place
-        if(scheme == CS_KERNEL_APPLY_CALLBACK)
-            allowOutofplace = false;
-        else
-            allowInplace = false;
+        allowInplace = false;
 
         /********************
         * Buffer and ArrayType
@@ -187,7 +188,7 @@ protected:
             allowedOutArrayTypes = {rocfft_array_type_complex_interleaved};
         }
         // should be real, but could be treated as CI (the alias type)
-        else if(scheme == CS_KERNEL_COPY_CMPLX_TO_R || scheme == CS_KERNEL_APPLY_CALLBACK)
+        else if(scheme == CS_KERNEL_COPY_CMPLX_TO_R)
         {
             allowedOutArrayTypes = {rocfft_array_type_real, rocfft_array_type_complex_interleaved};
         }
