@@ -27,7 +27,10 @@
 #include <unordered_map>
 #include <vector>
 
-#include "rocfft.h"
+#include <hip/hip_runtime.h>
+#include <hip/hip_runtime_api.h>
+
+#include "rocfft/rocfft.h"
 
 #include "logging.h"
 #include "plan.h"
@@ -38,11 +41,13 @@
 #include "kernel_launch.h"
 
 #include "function_pool.h"
-#include "ref_cpu.h"
 
 #include "real2complex.h"
 
+#include "../../shared/array_predicate.h"
 #include "../../shared/environment.h"
+#include "../../shared/fft_hash.h"
+#include "../../shared/hip_object_wrapper.h"
 #include "../../shared/precision_type.h"
 #include "../../shared/printbuffer.h"
 #include "../../shared/ptrdiff.h"
@@ -99,11 +104,13 @@ bool GetTuningKernelInfo(ExecPlan& execPlan)
 
     for(size_t i = 0; i < execPlan.execSeq.size(); ++i)
     {
-        RTCKernel*   localCompiledKernel = execPlan.execSeq[i]->compiledKernel.get().get();
+        TreeNode*    curNode             = execPlan.execSeq[i];
+        RTCKernel*   localCompiledKernel = curNode->compiledKernel.get().get();
         GridParam    gp                  = execPlan.gridParam[i];
-        FMKey        key                 = execPlan.execSeq[i]->GetKernelKey();
-        auto         lengths             = std::get<0>(key);
-        KernelConfig config              = std::get<4>(key);
+        FMKey        key                 = curNode->GetKernelKey();
+        auto         lengths             = key.lengths;
+        auto         scheme              = key.scheme;
+        KernelConfig config              = key.kernel_config;
 
         // get occupancy: 0 means it's compiled (AOT)
         //               -1 means failed on getting occupancy
@@ -145,14 +152,31 @@ bool GetTuningKernelInfo(ExecPlan& execPlan)
         util_rate /= config.factors.size();
         util_ss << std::fixed << util_rate << "]";
 
+        // 2D_SINGLE can get this value easily, others are not that naive
+        tuningPacket->globalRW_per_thread[i]
+            = (scheme == CS_KERNEL_2D_SINGLE) ? (lengths[0] * lengths[1]) / config.workgroup_size
+                                              : -1;
         tuningPacket->num_of_blocks[i] = gp.b_x;
         tuningPacket->lds_bytes[i]     = gp.lds_bytes;
         tuningPacket->occupancy[i]     = occupancy;
         tuningPacket->wgs[i]           = config.workgroup_size;
-        tuningPacket->tpt[i]           = config.threads_per_transform[0];
+        tuningPacket->tpt0[i]          = config.threads_per_transform[0];
+        tuningPacket->tpt1[i]          = config.threads_per_transform[1];
         tuningPacket->tpb[i]           = config.transforms_per_block;
         tuningPacket->util_rate[i]     = util_ss.str();
         tuningPacket->factors_str[i]   = factors_str;
+
+        // save some data back to kernel-config (to the saved references)
+        // since we've done buffer-assignment and collapse-dim now.
+        if(tuningPacket->is_builtin_kernel[i] == false)
+        {
+            KernelConfig* sol_kernel_config = execPlan.sol_kernel_configs[i];
+            sol_kernel_config->static_dim   = curNode->GetStaticDim();
+            sol_kernel_config->iAryType     = curNode->inArrayType;
+            sol_kernel_config->oAryType     = curNode->outArrayType;
+            sol_kernel_config->placement
+                = (curNode->placement == rocfft_placement_inplace) ? PC_IP : PC_OP;
+        }
     }
 
     return true;
@@ -163,8 +187,7 @@ static size_t data_size_bytes(const std::vector<size_t>& lengths,
                               rocfft_array_type          type)
 {
     // first compute the raw number of elements
-    const size_t elems = std::accumulate(
-        lengths.begin(), lengths.end(), static_cast<size_t>(1), std::multiplies<size_t>());
+    const size_t elems = product(lengths.begin(), lengths.end());
     // size of each element
     const size_t elemsize = real_type_size(precision);
     switch(type)
@@ -230,16 +253,15 @@ static float max_memory_bandwidth_GB_per_s()
     return result;
 }
 
-// Print either an input or output buffer, given column-major dimensions
-void DebugPrintBuffer(rocfft_ostream&            stream,
-                      rocfft_array_type          type,
-                      rocfft_precision           precision,
-                      void*                      buffer[],
-                      const std::vector<size_t>& length_cm,
-                      const std::vector<size_t>& stride_cm,
-                      size_t                     dist,
-                      size_t                     offset,
-                      size_t                     batch)
+// Copy device buffer to host buffer in column-major format with given strides
+void CopyDeviceBufferToHost(const rocfft_array_type    type,
+                            const rocfft_precision     precision,
+                            void*                      buffer[],
+                            const std::vector<size_t>& length_cm,
+                            const std::vector<size_t>& stride_cm,
+                            const size_t               dist,
+                            const size_t               batch,
+                            std::vector<hostbuf>&      bufvec)
 {
     const size_t size_elems = compute_ptrdiff(length_cm, stride_cm, batch, dist);
 
@@ -251,13 +273,6 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
     }
 
     size_t size_bytes = size_elems * base_type_size;
-    // convert length, stride to row-major for use with printbuffer
-    auto length_rm = length_cm;
-    auto stride_rm = stride_cm;
-    std::reverse(length_rm.begin(), length_rm.end());
-    std::reverse(stride_rm.begin(), stride_rm.end());
-    std::vector<hostbuf> bufvec;
-    std::vector<size_t>  print_offset(2, offset);
     if(array_type_is_planar(type))
     {
         // separate the real/imag data, so printbuffer will print them separately
@@ -270,12 +285,58 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
         if(hipMemcpy(bufvec.back().data(), buffer[1], size_bytes / 2, hipMemcpyDeviceToHost)
            != hipSuccess)
             throw std::runtime_error("hipMemcpy failure");
+    }
+    else
+    {
+        bufvec.resize(1);
+        bufvec.front().alloc(size_bytes);
+        if(hipMemcpy(bufvec.front().data(), buffer[0], size_bytes, hipMemcpyDeviceToHost)
+           != hipSuccess)
+            throw std::runtime_error("hipMemcpy failure");
+    }
+}
 
+// Print buffer 64-bit hash identifier, given column-major dimensions
+void DebugPrintHash(rocfft_ostream&             stream,
+                    rocfft_array_type           type,
+                    rocfft_precision            precision,
+                    const std::vector<hostbuf>& bufvec,
+                    const std::vector<size_t>&  length_cm,
+                    const std::vector<size_t>&  stride_cm,
+                    size_t                      dist,
+                    size_t                      batch)
+{
+    auto hash_in  = hash_input(precision, length_cm, stride_cm, dist, type, batch);
+    auto hash_out = hash_output<size_t>();
+    compute_hash(bufvec, hash_in, hash_out);
+
+    stream << "(" << hash_out.buffer_real << "," << hash_out.buffer_imag << ")" << std::endl;
+}
+
+// Print either an input or output buffer, given column-major dimensions
+void DebugPrintBuffer(rocfft_ostream&             stream,
+                      rocfft_array_type           type,
+                      rocfft_precision            precision,
+                      const std::vector<hostbuf>& bufvec,
+                      const std::vector<size_t>&  length_cm,
+                      const std::vector<size_t>&  stride_cm,
+                      size_t                      dist,
+                      size_t                      batch)
+{
+    // convert length, stride to row-major for use with printbuffer
+    auto length_rm = length_cm;
+    auto stride_rm = stride_cm;
+    std::reverse(length_rm.begin(), length_rm.end());
+    std::reverse(stride_rm.begin(), stride_rm.end());
+    std::vector<size_t> print_offset{0, 0};
+
+    if(array_type_is_planar(type))
+    {
         switch(precision)
         {
         case rocfft_precision_half:
         {
-            buffer_printer<_Float16> s;
+            buffer_printer<rocfft_fp16> s;
             s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
             break;
         }
@@ -295,12 +356,6 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
     }
     else
     {
-        bufvec.resize(1);
-        bufvec.front().alloc(size_bytes);
-        if(hipMemcpy(bufvec.front().data(), buffer[0], size_bytes, hipMemcpyDeviceToHost)
-           != hipSuccess)
-            throw std::runtime_error("hipMemcpy failure");
-
         switch(precision)
         {
         case rocfft_precision_half:
@@ -310,13 +365,13 @@ void DebugPrintBuffer(rocfft_ostream&            stream,
             case rocfft_array_type_complex_interleaved:
             case rocfft_array_type_hermitian_interleaved:
             {
-                buffer_printer<rocfft_complex<_Float16>> s;
+                buffer_printer<rocfft_complex<rocfft_fp16>> s;
                 s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
             case rocfft_array_type_real:
             {
-                buffer_printer<_Float16> s;
+                buffer_printer<rocfft_fp16> s;
                 s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
@@ -384,12 +439,14 @@ void SetDefaultCallback(const TreeNode* node, const SetCallbackType& type, void*
     auto result = hipSuccess;
 
     auto array_type = (type == SetCallbackType::LOAD) ? node->inArrayType : node->outArrayType;
+    auto node_callback_type = node->GetCallbackType(true);
 
-    // guaranteed to only have interleaved type by the caller (rocfft_execute)
-    auto is_complex = (array_type == rocfft_array_type_complex_interleaved
-                       || array_type == rocfft_array_type_hermitian_interleaved)
-                          ? true
-                          : false;
+    bool is_complex = array_type_is_complex(array_type);
+    // load r2c kernels and store c2r kernels need real-valued callbacks
+    if((type == SetCallbackType::LOAD && node_callback_type == CallbackType::USER_LOAD_STORE_R2C)
+       || (type == SetCallbackType::STORE
+           && node_callback_type == CallbackType::USER_LOAD_STORE_C2R))
+        is_complex = false;
 
     if(is_complex && type == SetCallbackType::LOAD)
     {
@@ -467,7 +524,8 @@ void SetDefaultCallback(const TreeNode* node, const SetCallbackType& type, void*
 void TransformPowX(const ExecPlan&       execPlan,
                    void*                 in_buffer[],
                    void*                 out_buffer[],
-                   rocfft_execution_info info)
+                   rocfft_execution_info info,
+                   size_t                multiPlanIdx)
 {
     assert(execPlan.execSeq.size() == execPlan.devFnCall.size());
     assert(execPlan.execSeq.size() == execPlan.gridParam.size());
@@ -479,13 +537,13 @@ void TransformPowX(const ExecPlan&       execPlan,
     bool emit_profile_log  = (processing_tuning || LOG_PROFILE_ENABLED()) && !info->rocfft_stream;
     bool emit_kernelio_log = LOG_KERNELIO_ENABLED();
 
-    rocfft_ostream* kernelio_stream = nullptr;
-    float           max_memory_bw   = 0.0;
-    hipEvent_t      start, stop;
+    rocfft_ostream*    kernelio_stream = nullptr;
+    float              max_memory_bw   = 0.0;
+    hipEvent_wrapper_t start, stop;
     if(emit_profile_log)
     {
-        if(hipEventCreate(&start) != hipSuccess || hipEventCreate(&stop) != hipSuccess)
-            throw std::runtime_error("hipEventCreate failure");
+        start.alloc();
+        stop.alloc();
         max_memory_bw = max_memory_bandwidth_GB_per_s();
     }
 
@@ -509,10 +567,7 @@ void TransformPowX(const ExecPlan&       execPlan,
         data.node          = execPlan.execSeq[i];
         data.rocfft_stream = (info == nullptr) ? 0 : info->rocfft_stream;
         data.deviceProp    = execPlan.deviceProp;
-        if(LOG_PLAN_ENABLED())
-            data.log_func = log_plan;
-        else
-            data.log_func = nullptr;
+        data.log_func      = LOG_PLAN_ENABLED() ? log_plan : nullptr;
 
         // Size of complex type
         const size_t complexTSize = complex_type_size(data.node->precision);
@@ -560,10 +615,9 @@ void TransformPowX(const ExecPlan&       execPlan,
             // }
             break;
         case OB_TEMP_BLUESTEIN:
-            data.bufIn[0] = (void*)((char*)info->workBuffer
-                                    + (execPlan.tmpWorkBufSize + execPlan.copyWorkBufSize
-                                       + data.node->iOffset)
-                                          * complexTSize);
+            data.bufIn[0]
+                = (void*)((char*)info->workBuffer
+                          + (execPlan.tmpWorkBufSize + execPlan.copyWorkBufSize) * complexTSize);
             // Bluestein mul-kernels (3 types) work well for CI->CI
             // so we only consider CI->CI now
             break;
@@ -619,10 +673,9 @@ void TransformPowX(const ExecPlan&       execPlan,
             // }
             break;
         case OB_TEMP_BLUESTEIN:
-            data.bufOut[0] = (void*)((char*)info->workBuffer
-                                     + (execPlan.tmpWorkBufSize + execPlan.copyWorkBufSize
-                                        + data.node->oOffset)
-                                           * complexTSize);
+            data.bufOut[0]
+                = (void*)((char*)info->workBuffer
+                          + (execPlan.tmpWorkBufSize + execPlan.copyWorkBufSize) * complexTSize);
             // Bluestein mul-kernels (3 types) work well for CI->CI
             // so we only consider CI->CI now
             break;
@@ -630,6 +683,34 @@ void TransformPowX(const ExecPlan&       execPlan,
             assert(false);
         }
 
+        // apply offsets to pointers
+        if(data.node->iOffset)
+        {
+            if(data.bufIn[0])
+                data.bufIn[0] = ptr_offset(data.bufIn[0],
+                                           data.node->iOffset,
+                                           data.node->precision,
+                                           data.node->inArrayType);
+            if(data.bufIn[1])
+                data.bufIn[1] = ptr_offset(data.bufIn[1],
+                                           data.node->iOffset,
+                                           data.node->precision,
+                                           data.node->inArrayType);
+        }
+        if(data.node->oOffset)
+        {
+            if(data.bufOut[0])
+                data.bufOut[0] = ptr_offset(data.bufOut[0],
+                                            data.node->oOffset,
+                                            data.node->precision,
+                                            data.node->outArrayType);
+            if(data.bufOut[1])
+                data.bufOut[1] = ptr_offset(data.bufOut[1],
+                                            data.node->oOffset,
+                                            data.node->precision,
+                                            data.node->outArrayType);
+        }
+
         // single-kernel bluestein requires a bluestein temp buffer separate from input and output
         if(data.node->scheme == CS_KERNEL_BLUESTEIN_SINGLE)
         {
@@ -658,139 +739,149 @@ void TransformPowX(const ExecPlan&       execPlan,
         if(emit_kernelio_log && data.node->scheme != CS_KERNEL_CHIRP)
         {
             kernelio_stream = LogSingleton::GetInstance().GetKernelIOOS();
-            *kernelio_stream << "--- --- kernel " << i << " (" << PrintScheme(data.node->scheme)
-                             << ") input:" << std::endl;
-
+            *kernelio_stream << "--- --- multiPlanIdx " << multiPlanIdx << " kernel " << i << " ("
+                             << PrintScheme(data.node->scheme) << ") input: " << std::endl;
             if(hipDeviceSynchronize() != hipSuccess)
                 throw std::runtime_error("hipDeviceSynchronize failure");
+
+            std::vector<hostbuf> bufInHost;
+            CopyDeviceBufferToHost(data.node->inArrayType,
+                                   data.node->precision,
+                                   data.bufIn,
+                                   data.node->length,
+                                   data.node->inStride,
+                                   data.node->iDist,
+                                   data.node->batch,
+                                   bufInHost);
+
             DebugPrintBuffer(*kernelio_stream,
                              data.node->inArrayType,
                              data.node->precision,
-                             data.bufIn,
+                             bufInHost,
                              data.node->length,
                              data.node->inStride,
                              data.node->iDist,
-                             // offset has already been added to bufIn
-                             0,
                              data.node->batch);
+            *kernelio_stream << "--- --- multiPlanIdx " << multiPlanIdx << " kernel " << i << " ("
+                             << PrintScheme(data.node->scheme) << ") input hash: " << std::endl;
+            DebugPrintHash(*kernelio_stream,
+                           data.node->inArrayType,
+                           data.node->precision,
+                           bufInHost,
+                           data.node->length,
+                           data.node->inStride,
+                           data.node->iDist,
+                           data.node->batch);
+            *kernelio_stream << std::endl;
         }
 
         DevFnCall fn = execPlan.devFnCall[i];
-        if(fn || data.node->compiledKernel.get())
+        if(!fn && !data.node->compiledKernel.get())
+            throw std::runtime_error("rocFFT null ptr function call error");
+
+#ifdef REF_DEBUG
+        rocfft_cout << "\n---------------------------------------------\n";
+        rocfft_cout << "\n\nkernel: " << i << std::endl;
+        rocfft_cout << "\tscheme: " << PrintScheme(execPlan.execSeq[i]->scheme) << std::endl;
+        rocfft_cout << "\titype: " << execPlan.execSeq[i]->inArrayType << std::endl;
+        rocfft_cout << "\totype: " << execPlan.execSeq[i]->outArrayType << std::endl;
+        rocfft_cout << "\tlength: ";
+        for(const auto& i : execPlan.execSeq[i]->length)
         {
-#ifdef REF_DEBUG
-            rocfft_cout << "\n---------------------------------------------\n";
-            rocfft_cout << "\n\nkernel: " << i << std::endl;
-            rocfft_cout << "\tscheme: " << PrintScheme(execPlan.execSeq[i]->scheme) << std::endl;
-            rocfft_cout << "\titype: " << execPlan.execSeq[i]->inArrayType << std::endl;
-            rocfft_cout << "\totype: " << execPlan.execSeq[i]->outArrayType << std::endl;
-            rocfft_cout << "\tlength: ";
-            for(const auto& i : execPlan.execSeq[i]->length)
-            {
-                rocfft_cout << i << " ";
-            }
-            rocfft_cout << std::endl;
-            rocfft_cout << "\tbatch:   " << execPlan.execSeq[i]->batch << std::endl;
-            rocfft_cout << "\tidist:   " << execPlan.execSeq[i]->iDist << std::endl;
-            rocfft_cout << "\todist:   " << execPlan.execSeq[i]->oDist << std::endl;
-            rocfft_cout << "\tistride:";
-            for(const auto& i : execPlan.execSeq[i]->inStride)
-            {
-                rocfft_cout << " " << i;
-            }
-            rocfft_cout << std::endl;
-            rocfft_cout << "\tostride:";
-            for(const auto& i : execPlan.execSeq[i]->outStride)
-            {
-                rocfft_cout << " " << i;
-            }
-            rocfft_cout << std::endl;
-
-            RefLibOp refLibOp(&data);
-#endif
-
-            // execution kernel:
-            if(emit_profile_log)
-                if(hipEventRecord(start) != hipSuccess)
-                    throw std::runtime_error("hipEventRecord failure");
-
-            DeviceCallOut back;
-
-            // give callback parameters to kernel launcher
-            data.callbacks = execPlan.execSeq[i]->callbacks;
-
-            // choose which compiled kernel to run
-            RTCKernel* localCompiledKernel
-                = data.get_callback_type() == CallbackType::NONE
-                      ? data.node->compiledKernel.get().get()
-                      : data.node->compiledKernelWithCallbacks.get().get();
-
-            // skip apply callback kernel if there's no callback
-            if(data.node->scheme != CS_KERNEL_APPLY_CALLBACK
-               || data.get_callback_type() != CallbackType::NONE)
-            {
-                if(localCompiledKernel)
-                    localCompiledKernel->launch(data);
-                else
-                    fn(&data, &back);
-            }
-            if(emit_profile_log)
-                if(hipEventRecord(stop) != hipSuccess)
-                    throw std::runtime_error("hipEventRecord failure");
-
-            // If we were on the null stream, measure elapsed time
-            // and emit profile logging.  If a stream was given, we
-            // can't wait for the transform to finish, so we can't
-            // emit any information.
-            if(emit_profile_log)
-            {
-                if(hipEventSynchronize(stop) != hipSuccess)
-                    throw std::runtime_error("hipEventSynchronize failure");
-                size_t in_size_bytes = data_size_bytes(
-                    data.node->length, data.node->precision, data.node->inArrayType);
-                size_t out_size_bytes = data_size_bytes(
-                    data.node->length, data.node->precision, data.node->outArrayType);
-                size_t total_size_bytes = (in_size_bytes + out_size_bytes) * data.node->batch;
-
-                float duration_ms = 0.0f;
-                if(hipEventElapsedTime(&duration_ms, start, stop) != hipSuccess)
-                    throw std::runtime_error("hipEventElapsedTime failure");
-                auto exec_bw        = execution_bandwidth_GB_per_s(total_size_bytes, duration_ms);
-                auto efficiency_pct = 0.0;
-                if(max_memory_bw != 0.0)
-                    efficiency_pct = 100.0 * exec_bw / max_memory_bw;
-                if(processing_tuning)
-                    tuningPacket->bw_effs[i] = efficiency_pct;
-
-                log_profile(__func__,
-                            "scheme",
-                            PrintScheme(execPlan.execSeq[i]->scheme),
-                            "duration_ms",
-                            duration_ms,
-                            "in_size",
-                            std::make_pair(static_cast<const size_t*>(data.node->length.data()),
-                                           data.node->length.size()),
-                            "total_size_bytes",
-                            total_size_bytes,
-                            "exec_GB_s",
-                            exec_bw,
-                            "max_mem_GB_s",
-                            max_memory_bw,
-                            "bw_efficiency_pct",
-                            efficiency_pct,
-                            "kernel_index",
-                            i);
-            }
-
-#ifdef REF_DEBUG
-            refLibOp.VerifyResult(&data);
-#endif
+            rocfft_cout << i << " ";
         }
+        rocfft_cout << std::endl;
+        rocfft_cout << "\tbatch:   " << execPlan.execSeq[i]->batch << std::endl;
+        rocfft_cout << "\tidist:   " << execPlan.execSeq[i]->iDist << std::endl;
+        rocfft_cout << "\todist:   " << execPlan.execSeq[i]->oDist << std::endl;
+        rocfft_cout << "\tistride:";
+        for(const auto& i : execPlan.execSeq[i]->inStride)
+        {
+            rocfft_cout << " " << i;
+        }
+        rocfft_cout << std::endl;
+        rocfft_cout << "\tostride:";
+        for(const auto& i : execPlan.execSeq[i]->outStride)
+        {
+            rocfft_cout << " " << i;
+        }
+        rocfft_cout << std::endl;
+
+        RefLibOp refLibOp(&data);
+#endif
+
+        // execution kernel:
+        if(emit_profile_log)
+            if(hipEventRecord(start) != hipSuccess)
+                throw std::runtime_error("hipEventRecord failure");
+
+        DeviceCallOut back;
+
+        // give callback parameters to kernel launcher
+        data.callbacks = execPlan.execSeq[i]->callbacks;
+
+        // choose which compiled kernel to run
+        RTCKernel* localCompiledKernel = data.get_callback_type() == CallbackType::NONE
+                                             ? data.node->compiledKernel.get().get()
+                                             : data.node->compiledKernelWithCallbacks.get().get();
+
+        if(localCompiledKernel)
+            localCompiledKernel->launch(data, data.node->deviceProp);
         else
+            fn(&data, &back);
+
+        if(emit_profile_log)
+            if(hipEventRecord(stop) != hipSuccess)
+                throw std::runtime_error("hipEventRecord failure");
+
+        // If we were on the null stream, measure elapsed time
+        // and emit profile logging.  If a stream was given, we
+        // can't wait for the transform to finish, so we can't
+        // emit any information.
+        if(emit_profile_log)
         {
-            rocfft_cout << "null ptr function call error\n";
+            if(hipEventSynchronize(stop) != hipSuccess)
+                throw std::runtime_error("hipEventSynchronize failure");
+            size_t in_size_bytes
+                = data_size_bytes(data.node->length, data.node->precision, data.node->inArrayType);
+            size_t out_size_bytes
+                = data_size_bytes(data.node->length, data.node->precision, data.node->outArrayType);
+            size_t total_size_bytes = (in_size_bytes + out_size_bytes) * data.node->batch;
+
+            float duration_ms = 0.0f;
+            if(hipEventElapsedTime(&duration_ms, start, stop) != hipSuccess)
+                throw std::runtime_error("hipEventElapsedTime failure");
+            auto exec_bw        = execution_bandwidth_GB_per_s(total_size_bytes, duration_ms);
+            auto efficiency_pct = 0.0;
+            if(max_memory_bw != 0.0)
+                efficiency_pct = 100.0 * exec_bw / max_memory_bw;
+            if(processing_tuning)
+                tuningPacket->bw_effs[i] = efficiency_pct;
+
+            log_profile(__func__,
+                        "scheme",
+                        PrintScheme(execPlan.execSeq[i]->scheme),
+                        "duration_ms",
+                        duration_ms,
+                        "in_size",
+                        std::make_pair(static_cast<const size_t*>(data.node->length.data()),
+                                       data.node->length.size()),
+                        "total_size_bytes",
+                        total_size_bytes,
+                        "exec_GB_s",
+                        exec_bw,
+                        "max_mem_GB_s",
+                        max_memory_bw,
+                        "bw_efficiency_pct",
+                        efficiency_pct,
+                        "kernel_index",
+                        i);
         }
 
+#ifdef REF_DEBUG
+        refLibOp.VerifyResult(&data);
+#endif
+
         if(emit_kernelio_log && data.node->scheme != CS_KERNEL_CHIRP)
         {
             hipError_t err = hipPeekAtLastError();
@@ -798,32 +889,67 @@ void TransformPowX(const ExecPlan&       execPlan,
             {
                 *kernelio_stream << "Error: " << hipGetErrorName(err) << ", "
                                  << hipGetErrorString(err) << std::endl;
-                exit(-1);
+                throw std::runtime_error("hipPeekAtLastError failed");
             }
             if(hipDeviceSynchronize() != hipSuccess)
                 throw std::runtime_error("hipDeviceSynchronize failure");
             *kernelio_stream << "executed kernel " << i << " (" << PrintScheme(data.node->scheme)
-                             << ")" << std::endl;
+                             << ")\n\n";
         }
     }
 
     if(emit_kernelio_log)
     {
-        *kernelio_stream << "final output:\n";
+        // offsets have only been applied to pointers given to kernels,
+        // so apply them here for printing too
+        void* out_buffer_offset[2] = {out_buffer[0], nullptr};
+        if(execPlan.rootPlan->oOffset)
+        {
+            out_buffer_offset[0] = ptr_offset(out_buffer_offset[0],
+                                              execPlan.rootPlan->oOffset,
+                                              execPlan.rootPlan->precision,
+                                              execPlan.rootPlan->outArrayType);
+        }
+
+        if(array_type_is_planar(execPlan.rootPlan->outArrayType))
+        {
+            out_buffer_offset[1] = out_buffer[1];
+
+            if(execPlan.rootPlan->oOffset)
+                out_buffer_offset[1] = ptr_offset(out_buffer_offset[1],
+                                                  execPlan.rootPlan->oOffset,
+                                                  execPlan.rootPlan->precision,
+                                                  execPlan.rootPlan->outArrayType);
+        }
+
+        std::vector<hostbuf> bufOutHost;
+        CopyDeviceBufferToHost(execPlan.rootPlan->outArrayType,
+                               execPlan.rootPlan->precision,
+                               out_buffer_offset,
+                               execPlan.rootPlan->GetOutputLength(),
+                               execPlan.rootPlan->outStride,
+                               execPlan.rootPlan->oDist,
+                               execPlan.rootPlan->batch,
+                               bufOutHost);
+
+        *kernelio_stream << "multiPlanIdx " << multiPlanIdx << " final output: " << std::endl;
         DebugPrintBuffer(*kernelio_stream,
                          execPlan.rootPlan->outArrayType,
                          execPlan.rootPlan->precision,
-                         out_buffer,
-                         execPlan.oLength,
+                         bufOutHost,
+                         execPlan.rootPlan->GetOutputLength(),
                          execPlan.rootPlan->outStride,
                          execPlan.rootPlan->oDist,
-                         execPlan.rootPlan->oOffset,
                          execPlan.rootPlan->batch);
+        *kernelio_stream << "multiPlanIdx " << multiPlanIdx << " final output hash: " << std::endl;
+        DebugPrintHash(*kernelio_stream,
+                       execPlan.rootPlan->outArrayType,
+                       execPlan.rootPlan->precision,
+                       bufOutHost,
+                       execPlan.rootPlan->GetOutputLength(),
+                       execPlan.rootPlan->outStride,
+                       execPlan.rootPlan->oDist,
+                       execPlan.rootPlan->batch);
         *kernelio_stream << std::endl;
     }
-    if(emit_profile_log)
-    {
-        (void)hipEventDestroy(start);
-        (void)hipEventDestroy(stop);
-    }
 }
