@@ -27,7 +27,7 @@
 #include "rtc_bluestein_gen.h"
 #include "tree_node.h"
 
-RTCKernel::RTCGenerator RTCKernelBluesteinSingle::generate_from_node(const TreeNode&    node,
+RTCKernel::RTCGenerator RTCKernelBluesteinSingle::generate_from_node(const LeafNode&    node,
                                                                      const std::string& gpu_arch,
                                                                      bool enable_callbacks)
 {
@@ -40,7 +40,7 @@ RTCKernel::RTCGenerator RTCKernelBluesteinSingle::generate_from_node(const TreeN
 
     // find kernel config from function pool
     auto& pool   = function_pool::get_function_pool();
-    auto  config = pool.get_kernel(fpkey(lengthBlue, node.precision));
+    auto  config = pool.get_kernel(FMKey(lengthBlue, node.precision));
 
     // get factors from the leaf node, which might have overridden what's in the pool
     auto&                     leafNode = static_cast<const LeafNode&>(node);
@@ -65,8 +65,9 @@ RTCKernel::RTCGenerator RTCKernelBluesteinSingle::generate_from_node(const TreeN
                                node.placement,
                                node.inArrayType,
                                node.outArrayType,
-                               enable_callbacks,
-                               node.IsScalingEnabled()};
+                               node.GetCallbackType(enable_callbacks),
+                               node.loadOps,
+                               node.storeOps};
 
     generator.generate_name = [=]() { return bluestein_single_rtc_kernel_name(specs); };
 
@@ -107,18 +108,6 @@ RTCKernelArgs RTCKernelBluesteinSingle::get_launch_args(DeviceCallIn& data)
         if(array_type_is_planar(data.node->outArrayType))
             kargs.append_ptr(data.bufOut[1]);
     }
-    switch(data.node->precision)
-    {
-    case rocfft_precision_half:
-        kargs.append_half(data.node->scale_factor);
-        break;
-    case rocfft_precision_single:
-        kargs.append_float(data.node->scale_factor);
-        break;
-    case rocfft_precision_double:
-        kargs.append_double(data.node->scale_factor);
-        break;
-    }
 
     // callback params
     kargs.append_ptr(data.callbacks.load_cb_fn);
@@ -126,10 +115,12 @@ RTCKernelArgs RTCKernelBluesteinSingle::get_launch_args(DeviceCallIn& data)
     kargs.append_unsigned_int(data.callbacks.load_cb_lds_bytes);
     kargs.append_ptr(data.callbacks.store_cb_fn);
     kargs.append_ptr(data.callbacks.store_cb_data);
+
+    append_load_store_args(kargs, *data.node);
     return kargs;
 }
 
-RTCKernel::RTCGenerator RTCKernelBluesteinMulti::generate_from_node(const TreeNode&    node,
+RTCKernel::RTCGenerator RTCKernelBluesteinMulti::generate_from_node(const LeafNode&    node,
                                                                     const std::string& gpu_arch,
                                                                     bool enable_callbacks)
 {
@@ -181,8 +172,9 @@ RTCKernel::RTCGenerator RTCKernelBluesteinMulti::generate_from_node(const TreeNo
                               node.precision,
                               node.inArrayType,
                               node.outArrayType,
-                              enable_callbacks,
-                              node.IsScalingEnabled()};
+                              node.GetCallbackType(enable_callbacks),
+                              node.loadOps,
+                              node.storeOps};
 
     generator.generate_name = [=]() { return bluestein_multi_rtc_kernel_name(specs); };
 
@@ -267,18 +259,7 @@ RTCKernelArgs RTCKernelBluesteinMulti::get_launch_args(DeviceCallIn& data)
         kargs.append_ptr(data.callbacks.store_cb_fn);
         kargs.append_ptr(data.callbacks.store_cb_data);
 
-        switch(data.node->precision)
-        {
-        case rocfft_precision_half:
-            kargs.append_half(data.node->scale_factor);
-            break;
-        case rocfft_precision_single:
-            kargs.append_float(data.node->scale_factor);
-            break;
-        case rocfft_precision_double:
-            kargs.append_double(data.node->scale_factor);
-            break;
-        }
+        append_load_store_args(kargs, *data.node);
     }
     return kargs;
 }
