@@ -20,15 +20,17 @@
 
 #include "rtc_kernel.h"
 #include "../../shared/array_predicate.h"
+#include "../../shared/device_properties.h"
 #include "../../shared/environment.h"
 #include "device/generator/stockham_gen.h"
 
 #include "device/kernel-generator-embed.h"
 #include "kernel_launch.h"
 #include "logging.h"
-#include "plan.h"
 #include "rtc_bluestein_kernel.h"
 #include "rtc_cache.h"
+#include "rtc_partial_pass_sbcc_64_64_64.h"
+#include "rtc_partial_pass_sbrr_64_64_64.h"
 #include "rtc_realcomplex_kernel.h"
 #include "rtc_stockham_kernel.h"
 #include "rtc_transpose_kernel.h"
@@ -40,10 +42,13 @@ RTCKernel::RTCKernel(const std::string&       kernel_name,
                      dim3                     blockDim)
     : gridDim(gridDim)
     , blockDim(blockDim)
+    , kernel_name(kernel_name)
 {
+#ifndef ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS
     // if we're only compiling, no need to actually load the code objects
     if(rocfft_getenv("ROCFFT_INTERNAL_COMPILE_ONLY") == "1")
         return;
+#endif
     if(hipModuleLoadData(&module, code.data()) != hipSuccess)
         throw std::runtime_error("failed to load module for " + kernel_name);
 
@@ -51,41 +56,30 @@ RTCKernel::RTCKernel(const std::string&       kernel_name,
         throw std::runtime_error("failed to get function " + kernel_name);
 }
 
-void RTCKernel::launch(DeviceCallIn& data)
+#ifndef ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS
+void RTCKernel::launch(DeviceCallIn& data, const hipDeviceProp_t& deviceProp)
 {
     RTCKernelArgs kargs = get_launch_args(data);
 
-    // scale factor, if necessary
-    if(data.node->IsScalingEnabled())
-    {
-        // scale factor is always double on the node, but needs to be
-        // the right type for the kernel
-        switch(data.node->precision)
-        {
-        case rocfft_precision_double:
-            kargs.append_double(data.node->scale_factor);
-            break;
-        case rocfft_precision_single:
-            kargs.append_float(data.node->scale_factor);
-            break;
-        case rocfft_precision_half:
-            kargs.append_half(data.node->scale_factor);
-            break;
-        }
-    }
-
     const auto& gp = data.gridParam;
 
     launch(kargs,
            {gp.b_x, gp.b_y, gp.b_z},
            {gp.wgs_x, gp.wgs_y, gp.wgs_z},
            gp.lds_bytes,
+           deviceProp,
            data.rocfft_stream);
 }
+#endif
 
-void RTCKernel::launch(
-    RTCKernelArgs& kargs, dim3 gridDim, dim3 blockDim, unsigned int lds_bytes, hipStream_t stream)
+void RTCKernel::launch(RTCKernelArgs&         kargs,
+                       dim3                   gridDim,
+                       dim3                   blockDim,
+                       unsigned int           lds_bytes,
+                       const hipDeviceProp_t& deviceProp,
+                       hipStream_t            stream)
 {
+    launch_limits_check(kernel_name, gridDim, blockDim, deviceProp);
     auto  size     = kargs.size_bytes();
     void* config[] = {HIP_LAUNCH_PARAM_BUFFER_POINTER,
                       kargs.data(),
@@ -93,6 +87,7 @@ void RTCKernel::launch(
                       &size,
                       HIP_LAUNCH_PARAM_END};
 
+#ifndef ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS
     if(LOG_PLAN_ENABLED())
     {
         int        max_blocks_per_sm;
@@ -104,6 +99,7 @@ void RTCKernel::launch(
         else
             *kernelplan_stream << "Can not retrieve occupancy info." << std::endl;
     }
+#endif
 
     if(hipModuleLaunchKernel(kernel,
                              gridDim.x,
@@ -129,14 +125,12 @@ bool RTCKernel::get_occupancy(dim3 blockDim, unsigned int lds_bytes, int& occupa
 }
 
 std::shared_future<std::unique_ptr<RTCKernel>>
-    RTCKernel::runtime_compile(const TreeNode&    node,
+    RTCKernel::runtime_compile(const LeafNode&    node,
                                const std::string& gpu_arch,
                                std::string&       kernel_name,
                                bool               enable_callbacks)
 {
-
-#ifdef ROCFFT_RUNTIME_COMPILE
-
+#ifndef ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS
     int deviceId = 0;
     if(hipGetDevice(&deviceId) != hipSuccess)
     {
@@ -160,7 +154,12 @@ std::shared_future<std::unique_ptr<RTCKernel>>
     if(!generator.valid())
         generator = RTCKernelBluesteinMulti::generate_from_node(node, gpu_arch, enable_callbacks);
     if(!generator.valid())
-        generator = RTCKernelApplyCallback::generate_from_node(node, gpu_arch, enable_callbacks);
+        generator
+            = RTCKernelPartialPassSBRR64Cubed::generate_from_node(node, gpu_arch, enable_callbacks);
+    if(!generator.valid())
+        generator
+            = RTCKernelPartialPassSBCC64Cubed::generate_from_node(node, gpu_arch, enable_callbacks);
+
     if(generator.valid())
     {
         kernel_name = generator.generate_name();
@@ -172,7 +171,7 @@ std::shared_future<std::unique_ptr<RTCKernel>>
             }
             try
             {
-                std::vector<char> code = cached_compile(
+                std::vector<char> code = RTCCache::cached_compile(
                     kernel_name, gpu_arch, generator.generate_src, generator_sum());
                 return generator.construct_rtckernel(
                     kernel_name, code, generator.gridDim, generator.blockDim);
@@ -194,7 +193,7 @@ std::shared_future<std::unique_ptr<RTCKernel>>
         kernel_name = generator.generate_name();
     }
 #endif
-    // runtime compilation is not enabled or no kernel found, return
+    // kernel harness being generated or no kernel found, return
     // null RTCKernel
     std::promise<std::unique_ptr<RTCKernel>> p;
     p.set_value(nullptr);
