@@ -1,4 +1,4 @@
-// Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2022-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -21,6 +21,7 @@
 #include "rtc_realcomplex_gen.h"
 #include "../../shared/array_predicate.h"
 #include "device/generator/generator.h"
+#include "rtc_test_harness.h"
 
 #include "device/kernel-generator-embed.h"
 
@@ -53,10 +54,8 @@ std::string realcomplex_rtc_kernel_name(const RealComplexSpecs& specs)
     kernel_name += rtc_array_type_name(specs.inArrayType);
     kernel_name += rtc_array_type_name(specs.outArrayType);
 
-    if(specs.enable_callbacks)
-        kernel_name += "_CB";
-    if(specs.enable_scaling)
-        kernel_name += "_scale";
+    kernel_name += load_store_name_suffix(specs.loadOps, specs.storeOps);
+    kernel_name += rtc_cbtype_name(specs.cbtype);
 
     return kernel_name;
 }
@@ -64,6 +63,17 @@ std::string realcomplex_rtc_kernel_name(const RealComplexSpecs& specs)
 std::string r2c_copy_rtc(const std::string& kernel_name, const RealComplexSpecs& specs)
 {
     std::string src;
+    // includes and declarations
+
+    src += rocfft_complex_h;
+    src += common_h;
+    src += callback_h;
+
+    src += rtc_precision_type_decl(specs.precision);
+
+    src += rtc_const_cbtype_decl(specs.cbtype);
+
+    src += "static const unsigned int dim = " + std::to_string(specs.dim) + ";\n";
 
     const char* input_type
         = specs.scheme == CS_KERNEL_COPY_R_TO_CMPLX ? "real_type_t<scalar_type>" : "scalar_type";
@@ -75,6 +85,7 @@ std::string r2c_copy_rtc(const std::string& kernel_name, const RealComplexSpecs&
     Variable lengths0{"lengths0", "unsigned int"};
     Variable lengths1{"lengths1", "unsigned int"};
     Variable lengths2{"lengths2", "unsigned int"};
+    Variable nbatch{"nbatch", "unsigned int"};
     Variable stride_in0{"stride_in0", "unsigned int"};
     Variable stride_in1{"stride_in1", "unsigned int"};
     Variable stride_in2{"stride_in2", "unsigned int"};
@@ -85,7 +96,6 @@ std::string r2c_copy_rtc(const std::string& kernel_name, const RealComplexSpecs&
     Variable stride_out3{"stride_out3", "unsigned int"};
     Variable input{"input", input_type, true, true};
     Variable output{"output", output_type, true, true};
-    Variable scale_factor_var{"scale_factor", "const real_type_t<scalar_type>"};
 
     Function func(kernel_name);
     func.launch_bounds = LAUNCH_BOUNDS_R2C_C2R_KERNEL;
@@ -97,6 +107,7 @@ std::string r2c_copy_rtc(const std::string& kernel_name, const RealComplexSpecs&
     func.arguments.append(lengths0);
     func.arguments.append(lengths1);
     func.arguments.append(lengths2);
+    func.arguments.append(nbatch);
     func.arguments.append(stride_in0);
     func.arguments.append(stride_in1);
     func.arguments.append(stride_in2);
@@ -109,36 +120,66 @@ std::string r2c_copy_rtc(const std::string& kernel_name, const RealComplexSpecs&
     func.arguments.append(output);
     for(const auto& arg : get_callback_args().arguments)
         func.arguments.append(arg);
-    if(specs.scheme == CS_KERNEL_COPY_CMPLX_TO_HERM || specs.scheme == CS_KERNEL_COPY_CMPLX_TO_R)
-        func.arguments.append(scale_factor_var);
 
     Variable dim_var{"dim", "const unsigned int"};
 
+    Variable global_idx{"global_idx", "unsigned int"};
+    func.body += Declaration{global_idx, "blockIdx.x * blockDim.x + threadIdx.x"};
+
     Variable idx_0{"idx_0", "const unsigned int"};
-    func.body += Declaration{idx_0, "blockIdx.x * blockDim.x + threadIdx.x"};
+    Variable idx_1{"idx_1", "const unsigned int"};
+    Variable idx_2{"idx_2", "const unsigned int"};
+    Variable idx_batch{"idx_batch", "const unsigned int"};
+
+    // variable to divide by when counting lengths0 - herm2c
+    // allocates threads along hermitian length, but other kernels
+    // allocate threads along FFT length
+    Variable& lengths0_divide
+        = specs.scheme == CS_KERNEL_COPY_HERM_TO_CMPLX ? hermitian_size : lengths0;
+
+    func.body += CommentLines{"per-dimension indexes"};
+    func.body += Declaration{idx_0, global_idx % lengths0_divide};
+    func.body += Assign{global_idx, global_idx / lengths0_divide};
+    if(specs.dim > 1)
+    {
+        func.body += Declaration{idx_1, global_idx % lengths1};
+        func.body += Assign{global_idx, global_idx / lengths1};
+    }
+    else
+    {
+        func.body += Declaration{idx_1, 0};
+    }
+    if(specs.dim > 2)
+    {
+        func.body += Declaration{idx_2, global_idx % lengths2};
+        func.body += Assign{global_idx, global_idx / lengths2};
+    }
+    else
+    {
+        func.body += Declaration{idx_2, 0};
+    }
+    func.body += Declaration{idx_batch, global_idx};
+
+    func.body += CommentLines{"any excess threads will be past the end of batch"};
+    func.body += If{idx_batch >= nbatch, {Return{}}};
 
     if(specs.scheme == CS_KERNEL_COPY_HERM_TO_CMPLX)
     {
         Variable input_offset{"input_offset", "auto"};
+        func.body += Declaration{input_offset,
+                                 idx_0 * stride_in0 + idx_1 * stride_in1 + idx_2 * stride_in2
+                                     + idx_batch * ("stride_in" + std::to_string(specs.dim))};
+
         Variable outputs_offset{"outputs_offset", "auto"};
         Variable outputc_offset{"outputc_offset", "auto"};
-        func.body += CommentLines{"start with batch offset"};
-        func.body
-            += Declaration{input_offset, "blockIdx.z * stride_in" + std::to_string(specs.dim)};
-        func.body += CommentLines{"straight copy"};
-        func.body
-            += Declaration{outputs_offset, "blockIdx.z * stride_out" + std::to_string(specs.dim)};
-        func.body += CommentLines{"conjugate copy"};
-        func.body
-            += Declaration{outputc_offset, "blockIdx.z * stride_out" + std::to_string(specs.dim)};
 
         func.body += CommentLines{"straight copy indices"};
         Variable is0{"is0", "auto"};
         Variable is1{"is1", "auto"};
         Variable is2{"is2", "auto"};
         func.body += Declaration{is0, idx_0};
-        func.body += Declaration{is1, Literal{"blockIdx.y"} % lengths1};
-        func.body += Declaration{is2, Literal{"blockIdx.y"} / lengths1};
+        func.body += Declaration{is1, idx_1};
+        func.body += Declaration{is2, idx_2};
 
         func.body += CommentLines{"conjugate copy indices"};
         Variable ic0{"ic0", "auto"};
@@ -148,17 +189,15 @@ std::string r2c_copy_rtc(const std::string& kernel_name, const RealComplexSpecs&
         func.body += Declaration{ic1, Ternary{is1 == 0, 0, lengths1 - is1}};
         func.body += Declaration{ic2, Ternary{is2 == 0, 0, lengths2 - is2}};
 
-        func.body
-            += AddAssign(input_offset, is2 * stride_in2 + is1 * stride_in1 + is0 * stride_in0);
+        func.body += Declaration{outputs_offset,
+                                 is0 * stride_out0 + is1 * stride_out1 + is2 * stride_out2
+                                     + idx_batch * ("stride_out" + std::to_string(specs.dim))};
+        func.body += Declaration{outputc_offset,
+                                 ic0 * stride_out0 + ic1 * stride_out1 + ic2 * stride_out2
+                                     + idx_batch * ("stride_out" + std::to_string(specs.dim))};
 
-        func.body += CommentLines{
-            "notice for 1D, blockIdx.y == 0 and thus has no effect for input_offset"};
-        func.body
-            += AddAssign(outputs_offset, is2 * stride_out2 + is1 * stride_out1 + is0 * stride_out0);
-        func.body
-            += AddAssign(outputc_offset, ic2 * stride_out2 + ic1 * stride_out1 + ic0 * stride_out0);
-
-        func.body += CallbackDeclaration("scalar_type", "cbtype");
+        func.body += CallbackLoadDeclaration("scalar_type", "cbtype");
+        func.body += CallbackStoreDeclaration("scalar_type", "cbtype");
 
         func.body += CommentLines{"we would do hermitian2complex at the start of a C2R transform,",
                                   "so it would never be the last kernel to write to global",
@@ -188,52 +227,24 @@ std::string r2c_copy_rtc(const std::string& kernel_name, const RealComplexSpecs&
     }
     else
     {
-        Variable lengths{"lengths", "const unsigned int", false, false, 3};
-        Variable stride_in{"stride_in", "const unsigned int", false, false, 4};
-        Variable stride_out{"stride_out", "const unsigned int", false, false, 4};
-        func.body += Declaration{lengths, ComplexLiteral{lengths0, lengths1, lengths2}};
-        func.body += Declaration{stride_in,
-                                 ComplexLiteral{stride_in0, stride_in1, stride_in2, stride_in3}};
-        func.body += Declaration{
-            stride_out, ComplexLiteral{stride_out0, stride_out1, stride_out2, stride_out3}};
-
-        func.body += CommentLines{"offsets"};
-        Variable offset_in{"offset_in", "size_t"};
-        Variable offset_out{"offset_out", "size_t"};
-        Variable remaining{"remaining", "size_t"};
-        Variable index_along_d{"index_along_d", "size_t"};
-        func.body += Declaration{offset_in, 0};
-        func.body += Declaration{offset_out, 0};
-        func.body += Declaration{remaining, "blockIdx.y"};
-        func.body += Declaration{index_along_d};
-        Variable d{"d", "unsigned int"};
-        For      offset_loop{d, 1, d < dim_var, 1};
-        offset_loop.body += Assign{index_along_d, remaining % lengths[d]};
-        offset_loop.body += Assign{remaining, remaining / lengths[d]};
-        offset_loop.body += Assign{offset_in, offset_in + index_along_d * stride_in[d]};
-        offset_loop.body += Assign{offset_out, offset_out + index_along_d * stride_out[d]};
-        func.body += offset_loop;
-
-        func.body += CommentLines{
-            "remaining should be 1 at this point, since batch goes into blockIdx.z"};
-        Variable batch{"batch", "unsigned int"};
-        func.body += Declaration{batch, "blockIdx.z"};
-        func.body += Assign{offset_in, offset_in + batch * stride_in[dim_var]};
-        func.body += Assign{offset_out, offset_out + batch * stride_out[dim_var]};
-
-        Variable      inputIdx{"inputIdx", "auto"};
-        Variable      outputIdx{"outputIdx", "auto"};
-        StatementList indexes{Declaration{inputIdx, offset_in + idx_0 * stride_in[0]},
-                              Declaration{outputIdx, offset_out + idx_0 * stride_out[0]}};
+        Variable inputIdx{"inputIdx", "auto"};
+        Variable outputIdx{"outputIdx", "auto"};
+        func.body += Declaration{inputIdx,
+                                 idx_0 * stride_in0 + idx_1 * stride_in1 + idx_2 * stride_in2
+                                     + idx_batch * ("stride_in" + std::to_string(specs.dim))};
+        func.body += Declaration{outputIdx,
+                                 idx_0 * stride_out0 + idx_1 * stride_out1 + idx_2 * stride_out2
+                                     + idx_batch * ("stride_out" + std::to_string(specs.dim))};
 
         if(specs.scheme == CS_KERNEL_COPY_R_TO_CMPLX)
         {
-            If guard{idx_0 < lengths[0], indexes};
+            If guard{idx_0 < lengths0, {}};
             guard.body += CommentLines{"we would do real2complex at the beginning of an R2C",
                                        "transform, so it would never be the last kernel to write",
                                        "to global memory.  don't bother going through the store cb",
                                        "to write global memory."};
-            guard.body += CallbackDeclaration("real_type_t<scalar_type>", "cbtype");
+            guard.body += CallbackLoadDeclaration("real_type_t<scalar_type>", "cbtype");
+            guard.body += CallbackStoreDeclaration("real_type_t<scalar_type>", "cbtype");
 
             ComplexLiteral elem{LoadGlobal{input, inputIdx}, "0.0"};
             guard.body += Assign{output[outputIdx], elem};
@@ -243,72 +254,59 @@ std::string r2c_copy_rtc(const std::string& kernel_name, const RealComplexSpecs&
         {
             func.body += CommentLines{"only read and write the first [length0/2+1] elements "
                                       "due to conjugate redundancy"};
-            If guard{idx_0 < Parens{1 + lengths[0] / 2}, indexes};
+
+            If guard{idx_0 < Parens{1 + lengths0 / 2}, {}};
+
             guard.body += CommentLines{"we would do complex2hermitian at the end of an R2C",
                                        "transform, so it would never be the first kernel to read",
                                        "from global memory.  don't bother going through the load",
                                        "callback to read global memory."};
 
-            guard.body += CallbackDeclaration("scalar_type", "cbtype");
+            guard.body += CallbackLoadDeclaration("scalar_type", "cbtype");
+            guard.body += CallbackStoreDeclaration("scalar_type", "cbtype");
 
             Variable elem{"elem", "scalar_type"};
             guard.body += Declaration{elem, input[inputIdx]};
-            if(specs.enable_scaling)
-                guard.body += MultiplyAssign(elem, scale_factor_var);
             guard.body += StoreGlobal{output, outputIdx, elem};
             func.body += guard;
         }
         else if(specs.scheme == CS_KERNEL_COPY_CMPLX_TO_R)
         {
-            If guard{idx_0 < lengths[0], indexes};
-            guard.body
-                += CommentLines{"we would do complex2real at the end of a C2R",
-                                "transform, so it would never be the first kernel to read",
-                                "from global memory.  don't bother going through the load cb",
-                                "to read global memory."};
-            guard.body += CallbackDeclaration("real_type_t<scalar_type>", "cbtype");
+            func.body += CommentLines{"we would do complex2real at the end of a C2R",
+                                      "transform, so it would never be the first kernel to read",
+                                      "from global memory.  don't bother going through the load cb",
+                                      "to read global memory."};
+            func.body += CallbackLoadDeclaration("real_type_t<scalar_type>", "cbtype");
+            func.body += CallbackStoreDeclaration("real_type_t<scalar_type>", "cbtype");
 
             Variable elem{"elem", "auto"};
-            guard.body += Declaration{elem, input[inputIdx].x()};
-            if(specs.enable_scaling)
-                guard.body += MultiplyAssign(elem, scale_factor_var);
-            guard.body += StoreGlobal{output, outputIdx, elem};
-            func.body += guard;
+            func.body += Declaration{elem, input[inputIdx].x()};
+            func.body += StoreGlobal{output, outputIdx, elem};
         }
     }
 
+    make_load_store_ops(func, specs.loadOps, specs.storeOps);
+
     if(array_type_is_planar(specs.inArrayType))
         func = make_planar(func, "input");
     if(array_type_is_planar(specs.outArrayType))
         func = make_planar(func, "output");
 
     src += func.render();
+    write_standalone_test_harness(func, src);
     return src;
 }
 
 // generate source for RTC realcomplex kernel.
 std::string realcomplex_rtc(const std::string& kernel_name, const RealComplexSpecs& specs)
 {
-    std::string src;
-    // includes and declarations
-
-    src += rocfft_complex_h;
-    src += common_h;
-    src += callback_h;
-
-    src += rtc_precision_type_decl(specs.precision);
-
-    src += rtc_const_cbtype_decl(specs.enable_callbacks);
-
-    src += "static const unsigned int dim = " + std::to_string(specs.dim) + ";\n";
-
     switch(specs.scheme)
     {
     case CS_KERNEL_COPY_R_TO_CMPLX:
     case CS_KERNEL_COPY_CMPLX_TO_HERM:
     case CS_KERNEL_COPY_CMPLX_TO_R:
     case CS_KERNEL_COPY_HERM_TO_CMPLX:
-        return src + r2c_copy_rtc(kernel_name, specs);
+        return r2c_copy_rtc(kernel_name, specs);
     default:
         throw std::runtime_error("invalid realcomplex rtc scheme");
     }
@@ -341,10 +339,8 @@ std::string realcomplex_even_rtc_kernel_name(const RealComplexEvenSpecs& specs)
     kernel_name += rtc_array_type_name(specs.inArrayType);
     kernel_name += rtc_array_type_name(specs.outArrayType);
 
-    if(specs.enable_callbacks)
-        kernel_name += "_CB";
-    if(specs.enable_scaling)
-        kernel_name += "_scale";
+    kernel_name += load_store_name_suffix(specs.loadOps, specs.storeOps);
+    kernel_name += rtc_cbtype_name(specs.cbtype);
 
     return kernel_name;
 }
@@ -360,7 +356,7 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
 
     src += rtc_precision_type_decl(specs.precision);
 
-    src += rtc_const_cbtype_decl(specs.enable_callbacks);
+    src += rtc_const_cbtype_decl(specs.cbtype);
 
     src += "static const unsigned int dim = " + std::to_string(specs.dim) + ";\n";
 
@@ -376,12 +372,13 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
     Variable half_N{"half_N", "const unsigned int"};
     Variable idist1D{"idist1D", "const unsigned int"};
     Variable odist1D{"odist1D", "const unsigned int"};
+    Variable higherFFTLengths{"higherFFTLengths", "unsigned int"};
+    Variable nbatch{"nbatch", "unsigned int"};
     Variable input{"input", "scalar_type", true, true};
     Variable idist{"idist", "const unsigned int"};
     Variable output{"output", "scalar_type", true, true};
     Variable odist{"odist", "const unsigned int"};
     Variable twiddles{"twiddles", "const scalar_type", true, true};
-    Variable scale_factor{"scale_factor", "const real_type_t<scalar_type>"};
 
     Function func{kernel_name};
     func.launch_bounds = LAUNCH_BOUNDS_R2C_C2R_KERNEL;
@@ -392,6 +389,8 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
         func.arguments.append(idist1D);
         func.arguments.append(odist1D);
     }
+    func.arguments.append(higherFFTLengths);
+    func.arguments.append(nbatch);
     func.arguments.append(input);
     func.arguments.append(idist);
     func.arguments.append(output);
@@ -399,17 +398,20 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
     func.arguments.append(twiddles);
     for(const auto& arg : get_callback_args().arguments)
         func.arguments.append(arg);
-    if(specs.enable_scaling)
-        func.arguments.append(scale_factor);
 
-    func.body += CommentLines{"blockIdx.y gives the multi-dimensional offset",
-                              "blockIdx.z gives the batch offset"};
+    Variable global_idx{"global_idx", "unsigned int"};
+    func.body += Declaration{global_idx, "blockIdx.x * blockDim.x + threadIdx.x"};
 
     Variable idx_p{"idx_p", "const auto"};
     Variable idx_q{"idx_q", "const auto"};
-    func.body += Declaration{idx_p, "blockIdx.x * blockDim.x + threadIdx.x"};
+    Variable idx_batch{"idx_batch", "const unsigned int"};
+    Variable idx_batch1D{"idx_batch1D", "const unsigned int"};
+
+    func.body += Declaration{idx_p, global_idx % half_N};
     func.body += Declaration{idx_q, half_N - idx_p};
 
+    func.body += Assign{global_idx, global_idx / half_N};
+
     Variable quarter_N{"quarter_N", "const auto"};
     func.body += Declaration{quarter_N, Parens{half_N + 1} / 2};
 
@@ -417,18 +419,24 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
 
     Variable input_offset{"input_offset", "auto"};
     Variable output_offset{"output_offset", "auto"};
-    guard.body += CommentLines{"blockIdx.z gives the batch offset"};
-    guard.body += Declaration(input_offset, Literal{"blockIdx.z"} * idist);
-    guard.body += Declaration{output_offset, Literal{"blockIdx.z"} * odist};
-
+    guard.body += Declaration(input_offset, idx_batch * idist);
+    guard.body += Declaration(output_offset, idx_batch * odist);
     if(specs.dim > 1)
     {
-        guard.body += CommentLines{
-            "blockIdx.y gives the multi-dimensional offset, stride is [i/o]dist1D."};
-        guard.body += AddAssign(input_offset, Literal{"blockIdx.y"} * idist1D);
-        guard.body += AddAssign(output_offset, Literal{"blockIdx.y"} * odist1D);
+        guard.body += AddAssign(input_offset, idx_batch1D * idist1D);
+        guard.body += AddAssign(output_offset, idx_batch1D * odist1D);
     }
 
+    func.body += CommentLines{
+        "this kernel treats all rows as batched 1D (it does not tolerate differing",
+        "strides between higher FFT dimensions), but user-provided batch is tracked",
+        "separately"};
+    func.body += Declaration{idx_batch1D, global_idx % higherFFTLengths};
+    func.body += Declaration{idx_batch, global_idx / higherFFTLengths};
+
+    func.body += CommentLines{"any excess threads will be past the end of batch"};
+    func.body += If{idx_batch >= nbatch, {Return{}}};
+
     if(specs.scheme == CS_KERNEL_R_TO_CMPLX)
     {
         guard.body += CommentLines{"post process can't be the first kernel, so don't bother",
@@ -441,7 +449,8 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
                                    "to global memory.  don't bother going through store",
                                    "callback to write global memory."};
     }
-    guard.body += CallbackDeclaration("scalar_type", "cbtype");
+    guard.body += CallbackLoadDeclaration("scalar_type", "cbtype");
+    guard.body += CallbackStoreDeclaration("scalar_type", "cbtype");
 
     Variable outval{"outval", "scalar_type"};
     guard.body += Declaration{outval};
@@ -461,15 +470,11 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
         if_idx_p_zero.body
             += Assign{outval.x(), input[input_offset + 0].x() - input[input_offset + 0].y()};
         if_idx_p_zero.body += Assign{outval.y(), 0};
-        if(specs.enable_scaling)
-            if_idx_p_zero.body += MultiplyAssign(outval, scale_factor);
         if_idx_p_zero.body += StoreGlobal{output, output_offset + half_N, outval};
 
         if_idx_p_zero.body
             += Assign{outval.x(), input[input_offset + 0].x() + input[input_offset + 0].y()};
         if_idx_p_zero.body += Assign{outval.y(), 0};
-        if(specs.enable_scaling)
-            if_idx_p_zero.body += MultiplyAssign(outval, scale_factor);
         if_idx_p_zero.body += StoreGlobal{output, output_offset + 0, outval};
     }
     else
@@ -487,8 +492,6 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
     {
         if_Ndiv4.body += Assign{outval.x(), input[input_offset + quarter_N].x()};
         if_Ndiv4.body += Assign{outval.y(), -input[input_offset + quarter_N].y()};
-        if(specs.enable_scaling)
-            if_Ndiv4.body += MultiplyAssign(outval, scale_factor);
         if_Ndiv4.body += StoreGlobal{output, output_offset + quarter_N, outval};
     }
     else
@@ -522,16 +525,12 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
             += Assign{outval.x(), u.x() + v.x() * twd_p.y() + u.y() * twd_p.x()};
         else_idx_p_nonzero.body
             += Assign{outval.y(), v.y() + u.y() * twd_p.y() - v.x() * twd_p.x()};
-        if(specs.enable_scaling)
-            else_idx_p_nonzero.body += MultiplyAssign(outval, scale_factor);
         else_idx_p_nonzero.body += StoreGlobal{output, output_offset + idx_p, outval};
 
         else_idx_p_nonzero.body
             += Assign{outval.x(), u.x() - v.x() * twd_p.y() - u.y() * twd_p.x()};
         else_idx_p_nonzero.body
             += Assign{outval.y(), -v.y() + u.y() * twd_p.y() - v.x() * twd_p.x()};
-        if(specs.enable_scaling)
-            else_idx_p_nonzero.body += MultiplyAssign(outval, scale_factor);
         else_idx_p_nonzero.body += StoreGlobal{output, output_offset + idx_q, outval};
     }
     else
@@ -561,12 +560,15 @@ std::string realcomplex_even_rtc(const std::string& kernel_name, const RealCompl
 
     func.body += guard;
 
+    make_load_store_ops(func, specs.loadOps, specs.storeOps);
+
     if(array_type_is_planar(specs.inArrayType))
         func = make_planar(func, "input");
     if(array_type_is_planar(specs.outArrayType))
         func = make_planar(func, "output");
 
     src += func.render();
+    write_standalone_test_harness(func, src);
     return src;
 }
 
@@ -592,10 +594,11 @@ std::string realcomplex_even_transpose_rtc_kernel_name(const RealComplexEvenTran
     kernel_name += rtc_array_type_name(specs.inArrayType);
     kernel_name += rtc_array_type_name(specs.outArrayType);
 
-    if(specs.enable_callbacks)
-        kernel_name += "_CB";
-    if(specs.enable_scaling)
-        kernel_name += "_scale";
+    kernel_name += load_store_name_suffix(specs.loadOps, specs.storeOps);
+    kernel_name += rtc_cbtype_name(specs.cbtype);
+
+    if(!specs.grid3D)
+        kernel_name += "_grid1D";
 
     return kernel_name;
 }
@@ -616,7 +619,7 @@ std::string realcomplex_even_transpose_rtc(const std::string&
 
     src += rtc_precision_type_decl(specs.precision);
 
-    src += rtc_const_cbtype_decl(specs.enable_callbacks);
+    src += rtc_const_cbtype_decl(specs.cbtype);
 
     // function arguments
     Variable dim{"dim", "size_t"};
@@ -628,6 +631,8 @@ std::string realcomplex_even_transpose_rtc(const std::string&
     Variable lengths{"lengths", "size_t", true, true};
     Variable inStride{"inStride", "size_t", true, true};
     Variable outStride{"outStride", "size_t", true, true};
+    Variable gridY{"gridY", "const unsigned int"};
+    Variable gridZ{"gridZ", "const unsigned int"};
 
     // r2c uses a device function helper to work out which dimension
     // we're transposing to
@@ -663,13 +668,45 @@ std::string realcomplex_even_transpose_rtc(const std::string&
     func.arguments.append(lengths);
     func.arguments.append(inStride);
     func.arguments.append(outStride);
+
     for(const auto& arg : get_callback_args().arguments)
         func.arguments.append(arg);
 
+    func.arguments.append(gridY);
+    func.arguments.append(gridZ);
+
+    func.body += CommentLines{"since gridDim is passed as {gridX, 1, 1}, use the",
+                              "following variables to recover block indices in a 3-D fashion:"};
+
+    Variable old_blockIdx_x{"old_blockIdx_x", "unsigned int"};
+    Variable old_blockIdx_y{"old_blockIdx_y", "unsigned int"};
+    Variable old_blockIdx_z{"old_blockIdx_z", "unsigned int"};
+    Variable remaining{"remaining", "unsigned int"};
+
+    // if a 1-D grid was provided because creating a natural 3-D grid exceeded allowed limits, then remap it to a 3-D grid.
+    if(!specs.grid3D)
+    {
+        func.body += Declaration{old_blockIdx_x, Literal{"blockIdx.x"} / (gridY * gridZ)};
+        func.body += Declaration{remaining, Literal{"blockIdx.x"} % (gridY * gridZ)};
+        func.body += Declaration{old_blockIdx_y, (remaining / gridZ)};
+        func.body += Declaration{old_blockIdx_z, Literal{"blockIdx.x"} % gridZ};
+    }
+
+    func.body += LineBreak{};
+
     Variable input_batch_start{"input_batch_start", "size_t"};
     Variable output_batch_start{"output_batch_start", "size_t"};
-    func.body += Declaration{input_batch_start, idist * Literal{"blockIdx.z"}};
-    func.body += Declaration{output_batch_start, odist * Literal{"blockIdx.z"}};
+
+    if(specs.grid3D)
+    {
+        func.body += Declaration{input_batch_start, idist * Literal{"blockIdx.z"}};
+        func.body += Declaration{output_batch_start, odist * Literal{"blockIdx.z"}};
+    }
+    else
+    {
+        func.body += Declaration{input_batch_start, idist * Literal{"old_blockIdx_z"}};
+        func.body += Declaration{output_batch_start, odist * Literal{"old_blockIdx_z"}};
+    }
 
     Variable leftTile{"leftTile", "__shared__ scalar_type", false, false, tileX};
     leftTile.size2D = tileY;
@@ -711,24 +748,44 @@ std::string realcomplex_even_transpose_rtc(const std::string&
     {
         func.body += CommentLines{
             "take fastest dimension and partition it into lengths that will go into each tile"};
-        len_row_init        = lengths[0];
-        tile_size_init      = Ternary{(len_row - 1) / 2 < tileX, (len_row - 1) / 2, tileX};
-        left_col_start_init = Literal{"blockIdx.x"} * tile_size + 1;
-        row_limit_init      = Ternary{dim == 2, lengths[1], lengths[1] * lengths[2]};
-        row_start_init      = Literal{"blockIdx.y"} * tileY;
-        row_end_init        = tileY + row_start;
+        len_row_init   = lengths[0];
+        tile_size_init = Ternary{(len_row - 1) / 2 < tileX, (len_row - 1) / 2, tileX};
+        row_limit_init = Ternary{dim == 2, lengths[1], lengths[1] * lengths[2]};
+
+        if(specs.grid3D)
+        {
+            left_col_start_init = Literal{"blockIdx.x"} * tile_size + 1;
+            row_start_init      = Literal{"blockIdx.y"} * tileY;
+        }
+        else
+        {
+            left_col_start_init = Literal{"old_blockIdx_x"} * tile_size + 1;
+            row_start_init      = Literal{"old_blockIdx_y"} * tileY;
+        }
+
+        row_end_init = tileY + row_start;
     }
     else
     {
         func.body += CommentLines{
             "take middle dimension and partition it into lengths that will go into each tile",
             "note that last row effectively gets thrown away"};
-        len_row_init        = Ternary{dim == 2, lengths[1] - 1, lengths[2] - 1};
-        tile_size_init      = Ternary{(len_row - 1) / 2 < tileY, (len_row - 1) / 2, tileY};
-        left_col_start_init = Literal{"blockIdx.y"} * tile_size + 1;
-        row_limit_init      = Ternary{dim == 2, lengths[0], lengths[0] * lengths[1]};
-        row_start_init      = Literal{"blockIdx.x"} * tileX;
-        row_end_init        = tileX + row_start;
+        len_row_init   = Ternary{dim == 2, lengths[1] - 1, lengths[2] - 1};
+        tile_size_init = Ternary{(len_row - 1) / 2 < tileY, (len_row - 1) / 2, tileY};
+        row_limit_init = Ternary{dim == 2, lengths[0], lengths[0] * lengths[1]};
+
+        if(specs.grid3D)
+        {
+            left_col_start_init = Literal{"blockIdx.y"} * tile_size + 1;
+            row_start_init      = Literal{"blockIdx.x"} * tileX;
+        }
+        else
+        {
+            left_col_start_init = Literal{"old_blockIdx_y"} * tile_size + 1;
+            row_start_init      = Literal{"old_blockIdx_x"} * tileX;
+        }
+
+        row_end_init = tileX + row_start;
     }
 
     func.body += Declaration{len_row, len_row_init};
@@ -803,13 +860,31 @@ std::string realcomplex_even_transpose_rtc(const std::string&
         read_left_idx  = input_batch_start + input_row_base + left_col_start + lds_col;
         read_right_idx = input_batch_start + input_row_base
                          + (len_row - (left_col_start + cols_to_read - 1)) + lds_col;
-        read_first_condition = Literal{"blockIdx.x"} == 0 && Literal{"threadIdx.x"} == 0
-                               && row_start + lds_row < row_end;
+
+        if(specs.grid3D)
+        {
+            read_first_condition = Literal{"blockIdx.x"} == 0 && Literal{"threadIdx.x"} == 0
+                                   && row_start + lds_row < row_end;
+        }
+        else
+        {
+            read_first_condition = Literal{"old_blockIdx_x"} == 0 && Literal{"threadIdx.x"} == 0
+                                   && row_start + lds_row < row_end;
+        }
+
         read_first_idx  = input_batch_start + input_row_base;
         read_middle_idx = input_batch_start + input_row_base + len_row / 2;
 
-        write_condition = Literal{"blockIdx.x"} == 0 && Literal{"threadIdx.x"} == 0
-                          && row_start + lds_row < row_end;
+        if(specs.grid3D)
+        {
+            write_condition = Literal{"blockIdx.x"} == 0 && Literal{"threadIdx.x"} == 0
+                              && row_start + lds_row < row_end;
+        }
+        else
+        {
+            write_condition = Literal{"old_blockIdx_x"} == 0 && Literal{"threadIdx.x"} == 0
+                              && row_start + lds_row < row_end;
+        }
 
         compute_first_val += Assign{val.x(), first_elem.x() - first_elem.y()};
         compute_first_val += Assign{val.y(), Literal{"0.0"}};
@@ -843,14 +918,32 @@ std::string realcomplex_even_transpose_rtc(const std::string&
             = input_batch_start + input_col_base + (left_col_start + lds_row) * input_col_stride;
         read_right_idx = input_batch_start + input_col_base
                          + (len_row - (left_col_start + lds_row)) * input_col_stride;
-        read_first_condition = Literal{"blockIdx.y"} == 0 && Literal{"threadIdx.y"} == 0
-                               && row_start + lds_col < row_end;
+
+        if(specs.grid3D)
+        {
+            read_first_condition = Literal{"blockIdx.y"} == 0 && Literal{"threadIdx.y"} == 0
+                                   && row_start + lds_col < row_end;
+        }
+        else
+        {
+            read_first_condition = Literal{"old_blockIdx_y"} == 0 && Literal{"threadIdx.y"} == 0
+                                   && row_start + lds_col < row_end;
+        }
+
         read_first_idx  = input_batch_start + input_col_base;
         read_middle_idx = input_batch_start + input_col_base + middle * input_col_stride;
         read_last_idx   = input_batch_start + input_col_base + len_row * input_col_stride;
 
-        write_condition = Literal{"blockIdx.y"} == 0 && Literal{"threadIdx.y"} == 0
-                          && row_start + lds_col < row_end;
+        if(specs.grid3D)
+        {
+            write_condition = Literal{"blockIdx.y"} == 0 && Literal{"threadIdx.y"} == 0
+                              && row_start + lds_col < row_end;
+        }
+        else
+        {
+            write_condition = Literal{"old_blockIdx_y"} == 0 && Literal{"threadIdx.y"} == 0
+                              && row_start + lds_col < row_end;
+        }
 
         compute_first_val += Assign{val.x(), first_elem.x() + last_elem.x()};
         compute_first_val += Assign{val.y(), first_elem.x() - last_elem.x()};
@@ -861,7 +954,8 @@ std::string realcomplex_even_transpose_rtc(const std::string&
         write_middle_idx = output_batch_start + output_row_base + middle * output_row_stride;
     }
 
-    func.body += CallbackDeclaration("scalar_type", "cbtype");
+    func.body += CallbackLoadDeclaration("scalar_type", "cbtype");
+    func.body += CallbackStoreDeclaration("scalar_type", "cbtype");
 
     func.body += Declaration{val};
 
@@ -921,8 +1015,17 @@ std::string realcomplex_even_transpose_rtc(const std::string&
         Variable col{"col", "size_t"};
 
         If butterfly{row_start + lds_row < row_end && lds_col < cols_to_read, {}};
-        butterfly.body
-            += Declaration{col, Literal{"blockIdx.x"} * tile_size + 1 + Literal{"threadIdx.x"}};
+
+        if(specs.grid3D)
+        {
+            butterfly.body
+                += Declaration{col, Literal{"blockIdx.x"} * tile_size + 1 + Literal{"threadIdx.x"}};
+        }
+        else
+        {
+            butterfly.body += Declaration{
+                col, Literal{"old_blockIdx_x"} * tile_size + 1 + Literal{"threadIdx.x"}};
+        }
 
         butterfly.body += Declaration{p, leftTile.at(lds_col, lds_row)};
         butterfly.body += Declaration{q, rightTile.at(cols_to_read - lds_col - 1, lds_row)};
@@ -983,106 +1086,14 @@ std::string realcomplex_even_transpose_rtc(const std::string&
         func.body += butterfly;
     }
 
+    make_load_store_ops(func, specs.loadOps, specs.storeOps);
+
     if(array_type_is_planar(specs.inArrayType))
         func = make_planar(func, "input");
     if(array_type_is_planar(specs.outArrayType))
         func = make_planar(func, "output");
 
     src += func.render();
-    return src;
-}
-
-std::string apply_callback_rtc_kernel_name(rocfft_precision precision)
-{
-    std::string kernel_name = "apply_callback";
-    kernel_name += rtc_precision_name(precision);
-    return kernel_name;
-}
-
-std::string apply_callback_rtc(const std::string& kernel_name, rocfft_precision precision)
-{
-    std::string src;
-
-    // includes and declarations
-
-    src += rocfft_complex_h;
-    src += common_h;
-    src += callback_h;
-
-    src += rtc_precision_type_decl(precision);
-
-    // callbacks are always enabled for this kernel
-    src += rtc_const_cbtype_decl(true);
-
-    // function arguments
-    Variable dim{"dim", "unsigned int"};
-    Variable lengths0{"lengths0", "unsigned int"};
-    Variable lengths1{"lengths1", "unsigned int"};
-    Variable lengths2{"lengths2", "unsigned int"};
-    Variable stride_in0{"stride_in0", "unsigned int"};
-    Variable stride_in1{"stride_in1", "unsigned int"};
-    Variable stride_in2{"stride_in2", "unsigned int"};
-    Variable stride_in3{"stride_in3", "unsigned int"};
-    Variable input{"input", "real_type_t<scalar_type>", true, true};
-
-    Function func{kernel_name};
-    func.launch_bounds = APPLY_REAL_CALLBACK_THREADS;
-    func.qualifier     = "extern \"C\" __global__";
-
-    func.arguments.append(dim);
-    func.arguments.append(lengths0);
-    func.arguments.append(lengths1);
-    func.arguments.append(lengths2);
-    func.arguments.append(stride_in0);
-    func.arguments.append(stride_in1);
-    func.arguments.append(stride_in2);
-    func.arguments.append(stride_in3);
-    func.arguments.append(input);
-
-    for(const auto& arg : get_callback_args().arguments)
-        func.arguments.append(arg);
-
-    // local variables
-    Variable idx_0{"idx_0", "const size_t"};
-    Variable lengths{"lengths", "const unsigned int", false, false, 3};
-    Variable stride_in{"stride_in", "const unsigned int", false, false, 4};
-    Variable offset_in{"offset_in", "size_t"};
-    Variable remaining{"remaining", "size_t"};
-    Variable index_along_d{"index_along_d", "size_t"};
-    Variable d{"d", "unsigned int"};
-    Variable batch{"batch", "size_t"};
-    Variable inputIdx{"inputIdx", "auto"};
-    Variable elem{"elem", "auto"};
-
-    func.body += Declaration{idx_0, "blockIdx.x * blockDim.x + threadIdx.x"};
-    func.body += Declaration{lengths, ComplexLiteral{lengths0, lengths1, lengths2}};
-    func.body
-        += Declaration{stride_in, ComplexLiteral{stride_in0, stride_in1, stride_in2, stride_in3}};
-
-    func.body += CommentLines{"offsets"};
-    func.body += Declaration{offset_in, 0};
-    func.body += Declaration{remaining};
-    func.body += Declaration{index_along_d};
-    func.body += Assign{remaining, "blockIdx.y"};
-
-    For offsetLoop{d, 1, d < dim, 1};
-    offsetLoop.body += Assign{index_along_d, remaining % lengths[d]};
-    offsetLoop.body += Assign{remaining, remaining / lengths[d]};
-    offsetLoop.body += Assign{offset_in, offset_in + index_along_d * stride_in[d]};
-    func.body += offsetLoop;
-
-    func.body
-        += CommentLines{"remaining should be 1 at this point, since batch goes into blockIdx.z"};
-    func.body += Declaration{batch, "blockIdx.z"};
-    func.body += Assign{offset_in, offset_in + batch * stride_in[dim]};
-
-    func.body += CallbackDeclaration("real_type_t<scalar_type>", "cbtype");
-    If accessor{idx_0 < lengths[0], {}};
-    accessor.body += Declaration{inputIdx, offset_in + idx_0 * stride_in[0]};
-    accessor.body += Declaration{elem, LoadGlobal{input, inputIdx}};
-    accessor.body += StoreGlobal{input, inputIdx, elem};
-    func.body += accessor;
-
-    src += func.render();
+    write_standalone_test_harness(func, src);
     return src;
 }
