@@ -1,4 +1,4 @@
-// Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2022-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -26,7 +26,7 @@
 #include "rtc_realcomplex_kernel.h"
 #include "tree_node.h"
 
-RTCKernel::RTCGenerator RTCKernelRealComplex::generate_from_node(const TreeNode&    node,
+RTCKernel::RTCGenerator RTCKernelRealComplex::generate_from_node(const LeafNode&    node,
                                                                  const std::string& gpu_arch,
                                                                  bool enable_callbacks)
 {
@@ -43,12 +43,12 @@ RTCKernel::RTCGenerator RTCKernelRealComplex::generate_from_node(const TreeNode&
     // hermitian size is used for hermitian->complex copy
     if(node.scheme == CS_KERNEL_COPY_HERM_TO_CMPLX)
         input_size = node.outputLength[0] / 2 + 1;
-    unsigned int batch          = node.batch;
-    unsigned int high_dimension = std::accumulate(
-        node.length.begin() + 1, node.length.end(), 1, std::multiplies<unsigned int>());
-    unsigned int blocks = (input_size - 1) / LAUNCH_BOUNDS_R2C_C2R_KERNEL + 1;
 
-    generator.gridDim  = {blocks, high_dimension, batch};
+    size_t elems = product(node.length.begin() + 1, node.length.end()) * input_size * node.batch;
+    generator.gridDim
+        = {static_cast<unsigned int>(DivRoundingUp<size_t>(elems, LAUNCH_BOUNDS_R2C_C2R_KERNEL)),
+           1,
+           1};
     generator.blockDim = {LAUNCH_BOUNDS_R2C_C2R_KERNEL, 1, 1};
 
     RealComplexSpecs specs{node.scheme,
@@ -56,8 +56,9 @@ RTCKernel::RTCGenerator RTCKernelRealComplex::generate_from_node(const TreeNode&
                            node.precision,
                            node.inArrayType,
                            node.outArrayType,
-                           enable_callbacks,
-                           node.IsScalingEnabled()};
+                           node.GetCallbackType(enable_callbacks),
+                           node.loadOps,
+                           node.storeOps};
 
     generator.generate_name = [=]() { return realcomplex_rtc_kernel_name(specs); };
 
@@ -99,6 +100,7 @@ RTCKernelArgs RTCKernelRealComplex::get_launch_args(DeviceCallIn& data)
     kargs.append_unsigned_int(kern_lengths[0]);
     kargs.append_unsigned_int(kern_lengths[1]);
     kargs.append_unsigned_int(kern_lengths[2]);
+    kargs.append_unsigned_int(data.node->batch);
     kargs.append_unsigned_int(kern_stride_in[0]);
     kargs.append_unsigned_int(kern_stride_in[1]);
     kargs.append_unsigned_int(kern_stride_in[2]);
@@ -121,27 +123,12 @@ RTCKernelArgs RTCKernelRealComplex::get_launch_args(DeviceCallIn& data)
     kargs.append_unsigned_int(data.callbacks.load_cb_lds_bytes);
     kargs.append_ptr(data.callbacks.store_cb_fn);
     kargs.append_ptr(data.callbacks.store_cb_data);
-    if(data.node->scheme == CS_KERNEL_COPY_CMPLX_TO_HERM
-       || data.node->scheme == CS_KERNEL_COPY_CMPLX_TO_R)
-    {
-        switch(data.node->precision)
-        {
-        case rocfft_precision_half:
-            kargs.append_half(data.node->scale_factor);
-            break;
-        case rocfft_precision_single:
-            kargs.append_float(data.node->scale_factor);
-            break;
-        case rocfft_precision_double:
-            kargs.append_double(data.node->scale_factor);
-            break;
-        }
-    }
+    append_load_store_args(kargs, *data.node);
 
     return kargs;
 }
 
-RTCKernel::RTCGenerator RTCKernelRealComplexEven::generate_from_node(const TreeNode&    node,
+RTCKernel::RTCGenerator RTCKernelRealComplexEven::generate_from_node(const LeafNode&    node,
                                                                      const std::string& gpu_arch,
                                                                      bool enable_callbacks)
 {
@@ -167,24 +154,23 @@ RTCKernel::RTCGenerator RTCKernelRealComplexEven::generate_from_node(const TreeN
     }
     const bool Ndiv4 = half_N % 2 == 0;
 
-    const unsigned int batch = node.batch;
+    size_t elems = std::accumulate(
+        node.length.begin() + 1, node.length.end(), half_N * node.batch, std::multiplies<size_t>());
+    generator.gridDim
+        = {static_cast<unsigned int>(DivRoundingUp<size_t>(elems, LAUNCH_BOUNDS_R2C_C2R_KERNEL)),
+           1,
+           1};
 
-    const unsigned int high_dimension = std::accumulate(
-        node.length.begin() + 1, node.length.end(), 1, std::multiplies<unsigned int>());
-
-    const unsigned int block_size = LAUNCH_BOUNDS_R2C_C2R_KERNEL;
-    const unsigned int blocks     = ((half_N + 1) / 2 + block_size - 1) / block_size;
-
-    generator.gridDim  = {blocks, high_dimension, batch};
-    generator.blockDim = {block_size, 1, 1};
+    generator.blockDim = {LAUNCH_BOUNDS_R2C_C2R_KERNEL, 1, 1};
 
     RealComplexEvenSpecs specs{{node.scheme,
                                 node.length.size(),
                                 node.precision,
                                 node.inArrayType,
                                 node.outArrayType,
-                                enable_callbacks,
-                                node.IsScalingEnabled()},
+                                node.GetCallbackType(enable_callbacks),
+                                node.loadOps,
+                                node.storeOps},
                                Ndiv4};
 
     generator.generate_name = [=]() { return realcomplex_even_rtc_kernel_name(specs); };
@@ -212,6 +198,9 @@ RTCKernelArgs RTCKernelRealComplexEven::get_launch_args(DeviceCallIn& data)
         kargs.append_unsigned_int(data.node->inStride[1]);
         kargs.append_unsigned_int(data.node->outStride[1]);
     }
+    unsigned int higherFFTLengths = product(data.node->length.begin() + 1, data.node->length.end());
+    kargs.append_unsigned_int(higherFFTLengths);
+    kargs.append_unsigned_int(data.node->batch);
     kargs.append_ptr(data.bufIn[0]);
     if(array_type_is_planar(data.node->inArrayType))
         kargs.append_ptr(data.bufIn[1]);
@@ -227,27 +216,13 @@ RTCKernelArgs RTCKernelRealComplexEven::get_launch_args(DeviceCallIn& data)
     kargs.append_unsigned_int(data.callbacks.load_cb_lds_bytes);
     kargs.append_ptr(data.callbacks.store_cb_fn);
     kargs.append_ptr(data.callbacks.store_cb_data);
-    if(data.node->IsScalingEnabled())
-    {
-        switch(data.node->precision)
-        {
-        case rocfft_precision_half:
-            kargs.append_half(data.node->scale_factor);
-            break;
-        case rocfft_precision_single:
-            kargs.append_float(data.node->scale_factor);
-            break;
-        case rocfft_precision_double:
-            kargs.append_double(data.node->scale_factor);
-            break;
-        }
-    }
+    append_load_store_args(kargs, *data.node);
 
     return kargs;
 }
 
 RTCKernel::RTCGenerator RTCKernelRealComplexEvenTranspose::generate_from_node(
-    const TreeNode& node, const std::string& gpu_arch, bool enable_callbacks)
+    const LeafNode& node, const std::string& gpu_arch, bool enable_callbacks)
 {
     RTCGenerator generator;
     if(node.scheme != CS_KERNEL_R_TO_CMPLX_TRANSPOSE
@@ -257,56 +232,69 @@ RTCKernel::RTCGenerator RTCKernelRealComplexEvenTranspose::generate_from_node(
     const unsigned int tileX = RealComplexEvenTransposeSpecs::TileX(node.scheme);
     const unsigned int tileY = RealComplexEvenTransposeSpecs::TileY();
 
-    unsigned int count = node.batch;
-    unsigned int m     = node.length[1];
-    unsigned int n     = node.length[0];
-    unsigned int dim   = node.length.size();
+    unsigned int m   = node.length[1];
+    unsigned int n   = node.length[0];
+    unsigned int dim = node.length.size();
+
+    unsigned int gridX;
+    unsigned int gridX_1d;
+    unsigned int gridY;
+    unsigned int gridZ = node.batch;
 
     if(node.scheme == CS_KERNEL_R_TO_CMPLX_TRANSPOSE)
     {
-        // grid X dimension handles 2 tiles at a time, so allocate enough
-        // blocks to go halfway across 'n'
-        //
-        // grid Y dimension needs enough blocks to handle the second
-        // dimension - multiply by the third dimension to get enough
-        // blocks, if we're doing 3D
-        //
-        // grid Z counts number of batches
-        generator.gridDim
-            = {(n - 1) / tileX / 2 + 1,
-               ((m - 1) / tileY + 1) * (dim > 2 ? static_cast<unsigned int>(node.length[2]) : 1),
-               count};
-        // one thread per element in a tile
-        generator.blockDim = {tileX, tileY, 1};
+        // In a 3-D grid, gridX handles 2 tiles at a time, gridY, the second
+        // dimension - multiplied by the third dimension, and gridZ, the number of batches
+        gridX    = (n - 1) / tileX / 2 + 1;
+        gridX_1d = (tileX * tileY) * gridX;
+        gridY = ((m - 1) / tileY + 1) * (dim > 2 ? static_cast<unsigned int>(node.length[2]) : 1);
     }
     else
     {
-        // grid X dimension needs enough blocks to handle the first
-        // dimension - multiply by the second dimension to get enough
-        // blocks, if we're doing 3D
+        // In a 3-D grid, gridX handles the first dimension - multiplied by the second dimension
         if(dim > 2)
         {
             n *= node.length[1];
             m = node.length[2];
         }
+        gridX    = (n - 1) / tileX + 1;
+        gridX_1d = (tileX * tileY) * (n - 1) / tileX + 1;
 
-        // grid Y dimension handles 2 tiles at a time, so allocate enough
+        // gridY dimension handles 2 tiles at a time, so allocate enough
         // blocks to go halfway across 'm'
-        auto gridY = std::max<unsigned int>((((m - 1) / 2) + (tileY - 1)) / tileY, 1);
-
-        // grid Z counts number of batches
-        generator.gridDim = {(n - 1) / tileX + 1, gridY, count};
-        // one thread per element in a tile
-        generator.blockDim = {tileX, tileY, 1};
+        gridY = std::max<unsigned int>((((m - 1) / 2) + (tileY - 1)) / tileY, 1);
     }
 
+    // check if it is possible to directly use a 3-D grid of GPUs
+    bool grid3D = false;
+    if(gridY < (1U << 16) && gridZ < (1U << 16))
+    {
+        // grid sizes are within limits to use a 3-D grid of GPUs
+        generator.gridDim = {gridX, gridY, gridZ};
+        grid3D            = true;
+    }
+    else
+    {
+        // grid sizes exceed (1U << 16) - 1 limits, then use a 1-D grid,
+        // a natural remap to a 3-D grid is then performed when creating the kernel source code
+        size_t elems      = gridX_1d * gridY * gridZ;
+        generator.gridDim = {
+            static_cast<unsigned int>(DivRoundingUp<size_t>(elems, LAUNCH_BOUNDS_R2C_C2R_KERNEL)),
+            1,
+            1};
+    }
+
+    generator.blockDim = {tileX, tileY, 1};
+
     RealComplexEvenTransposeSpecs specs{{node.scheme,
                                          node.length.size(),
                                          node.precision,
                                          node.inArrayType,
                                          node.outArrayType,
-                                         enable_callbacks,
-                                         node.IsScalingEnabled()}};
+                                         node.GetCallbackType(enable_callbacks),
+                                         node.loadOps,
+                                         node.storeOps,
+                                         grid3D}};
 
     generator.generate_name = [=]() { return realcomplex_even_transpose_rtc_kernel_name(specs); };
 
@@ -321,6 +309,7 @@ RTCKernel::RTCGenerator RTCKernelRealComplexEvenTranspose::generate_from_node(
         return std::unique_ptr<RTCKernel>(
             new RTCKernelRealComplexEvenTranspose(kernel_name, code, gridDim, blockDim));
     };
+
     return generator;
 }
 
@@ -348,82 +337,36 @@ RTCKernelArgs RTCKernelRealComplexEvenTranspose::get_launch_args(DeviceCallIn& d
     kargs.append_ptr(data.callbacks.store_cb_fn);
     kargs.append_ptr(data.callbacks.store_cb_data);
 
-    return kargs;
-}
+    // pass gridY and gridZ to restore a 3-D GPU grid, if needed for large grids
+    unsigned int gridY;
+    unsigned int gridZ = data.node->batch;
 
-RTCKernel::RTCGenerator RTCKernelApplyCallback::generate_from_node(const TreeNode&    node,
-                                                                   const std::string& gpu_arch,
-                                                                   bool enable_callbacks)
-{
-    RTCGenerator generator;
-
-    if(node.scheme != CS_KERNEL_APPLY_CALLBACK)
-        return generator;
-
-    // input_size is the innermost dimension
-    size_t input_size = node.length[0];
-
-    unsigned int batch          = node.batch;
-    unsigned int high_dimension = 1;
-    unsigned int dim            = node.length.size();
-
-    for(unsigned int i = 1; i < dim; i++)
+    if(data.node->scheme == CS_KERNEL_R_TO_CMPLX_TRANSPOSE)
     {
-        high_dimension *= node.length[i];
+        const unsigned int tileY = RealComplexEvenTransposeSpecs::TileY();
+        unsigned int       m     = data.node->length[1];
+        unsigned int       dim   = data.node->length.size();
+
+        gridY = ((m - 1) / tileY + 1)
+                * (dim > 2 ? static_cast<unsigned int>(data.node->length[2]) : 1);
+    }
+    else
+    {
+        unsigned int       m     = data.node->length[1];
+        const unsigned int tileY = RealComplexEvenTransposeSpecs::TileY();
+
+        if(data.node->length.size() > 2)
+        {
+            m = data.node->length[2];
+        }
+
+        gridY = std::max<unsigned int>((((m - 1) / 2) + (tileY - 1)) / tileY, 1);
     }
 
-    rocfft_precision precision = node.precision;
+    kargs.append_unsigned_int(gridY);
+    kargs.append_unsigned_int(gridZ);
 
-    unsigned int blocks = (input_size - 1) / APPLY_REAL_CALLBACK_THREADS + 1;
-
-    generator.gridDim  = {blocks, high_dimension, batch};
-    generator.blockDim = {APPLY_REAL_CALLBACK_THREADS, 1, 1};
-
-    generator.generate_name = [=]() { return apply_callback_rtc_kernel_name(precision); };
-
-    generator.generate_src = [=](const std::string& kernel_name) {
-        return apply_callback_rtc(kernel_name, precision);
-    };
-
-    generator.construct_rtckernel = [=](const std::string&       kernel_name,
-                                        const std::vector<char>& code,
-                                        dim3                     gridDim,
-                                        dim3                     blockDim) {
-        return std::unique_ptr<RTCKernel>(
-            new RTCKernelApplyCallback(kernel_name, code, gridDim, blockDim));
-    };
-    return generator;
-}
-
-RTCKernelArgs RTCKernelApplyCallback::get_launch_args(DeviceCallIn& data)
-{
-    RTCKernelArgs kargs;
-
-    size_t dim = data.node->length.size();
-
-    // explode lengths/strides out to pass to the kernel
-    std::array<unsigned int, 3> kern_lengths{1, 1, 1};
-    std::array<unsigned int, 4> kern_stride_in{1, 1, 1, 1};
-
-    std::copy(data.node->length.begin(), data.node->length.end(), kern_lengths.begin());
-    std::copy(data.node->inStride.begin(), data.node->inStride.end(), kern_stride_in.begin());
-    kern_stride_in[dim] = data.node->iDist;
-
-    kargs.append_unsigned_int(dim);
-    kargs.append_unsigned_int(kern_lengths[0]);
-    kargs.append_unsigned_int(kern_lengths[1]);
-    kargs.append_unsigned_int(kern_lengths[2]);
-    kargs.append_unsigned_int(kern_stride_in[0]);
-    kargs.append_unsigned_int(kern_stride_in[1]);
-    kargs.append_unsigned_int(kern_stride_in[2]);
-    kargs.append_unsigned_int(kern_stride_in[3]);
-    kargs.append_ptr(data.bufIn[0]);
-    // callback params
-    kargs.append_ptr(data.callbacks.load_cb_fn);
-    kargs.append_ptr(data.callbacks.load_cb_data);
-    kargs.append_unsigned_int(data.callbacks.load_cb_lds_bytes);
-    kargs.append_ptr(data.callbacks.store_cb_fn);
-    kargs.append_ptr(data.callbacks.store_cb_data);
+    append_load_store_args(kargs, *data.node);
 
     return kargs;
 }
