@@ -28,7 +28,7 @@
 
 #include "device/kernel-generator-embed.h"
 
-RTCKernel::RTCGenerator RTCKernelStockham::generate_from_node(const TreeNode&    node,
+RTCKernel::RTCGenerator RTCKernelStockham::generate_from_node(const LeafNode&    node,
                                                               const std::string& gpu_arch,
                                                               bool               enable_callbacks)
 {
@@ -38,27 +38,16 @@ RTCKernel::RTCGenerator RTCKernelStockham::generate_from_node(const TreeNode&
     std::optional<StockhamGeneratorSpecs> specs;
     std::optional<StockhamGeneratorSpecs> specs2d;
 
-    // if scale factor is enabled, we force RTC for this kernel
-    bool enable_scaling = node.IsScalingEnabled();
-
-    // for sbrc variant, the sbrcTranstype should be assigned when we are here
-    // since the value is assigned in KernelCheck()
-    SBRC_TRANSPOSE_TYPE transpose_type = node.sbrcTranstype;
-
     // SBRC variants look in the function pool for plain BLOCK_RC to
     // learn the block width, then decide on the transpose type once
     // that's known.
     auto         pool_scheme = node.scheme;
-    unsigned int static_dim  = node.length.size();
+    unsigned int static_dim  = node.GetStaticDim();
     if(pool_scheme == CS_KERNEL_STOCKHAM_TRANSPOSE_XY_Z
        || pool_scheme == CS_KERNEL_STOCKHAM_TRANSPOSE_Z_XY
        || pool_scheme == CS_KERNEL_STOCKHAM_R_TO_CMPLX_TRANSPOSE_Z_XY)
     {
         pool_scheme = CS_KERNEL_STOCKHAM_BLOCK_RC;
-        // These are all 3D kernels, but are sometimes shoehorned
-        // into 2D plans.  Make sure they get at least 3 dims.
-        if(static_dim == 2)
-            static_dim = 3;
     }
 
     std::optional<FFTKernel> kernel;
@@ -76,7 +65,13 @@ RTCKernel::RTCGenerator RTCKernelStockham::generate_from_node(const TreeNode&
     case CS_KERNEL_STOCKHAM_BLOCK_CR:
     case CS_KERNEL_STOCKHAM_BLOCK_RC:
     {
-        if((pool_scheme == CS_KERNEL_STOCKHAM_BLOCK_RC) && (transpose_type == NONE))
+        // Partial-pass nodes have their own generators
+        if(node.applyPartialPass)
+            return generator;
+
+        // for sbrc variant, the sbrcTranstype should be assigned when we are here
+        // since the value is assigned in KernelCheck()
+        if((pool_scheme == CS_KERNEL_STOCKHAM_BLOCK_RC) && (node.sbrcTranstype == NONE))
             throw std::runtime_error("Invalid SBRC_TRANS_TYPE for SBRC kernel");
 
         // these go into the function pool normally and are passed to
@@ -85,7 +80,8 @@ RTCKernel::RTCGenerator RTCKernelStockham::generate_from_node(const TreeNode&
         // if a kernel is already precompiled, just use that.  but
         // changing largeTwdBatch transform count requires RTC, so we
         // can't use a precompiled kernel in that case.
-        if(kernel->device_function && !enable_scaling && !node.largeTwdBatchIsTransformCount)
+        if(kernel->device_function && !node.loadOps.enabled() && !node.storeOps.enabled()
+           && !node.largeTwdBatchIsTransformCount)
         {
             is_pre_compiled = true;
         }
@@ -108,7 +104,7 @@ RTCKernel::RTCGenerator RTCKernelStockham::generate_from_node(const TreeNode&
     {
         kernel = pool.get_kernel(key);
         // already precompiled?
-        if(kernel->device_function && !enable_scaling)
+        if(kernel->device_function && !node.loadOps.enabled() && !node.storeOps.enabled())
         {
             is_pre_compiled = true;
         }
@@ -188,10 +184,11 @@ RTCKernel::RTCGenerator RTCKernelStockham::generate_from_node(const TreeNode&
                                         node.ebtype,
                                         node.dir2regMode,
                                         node.intrinsicMode,
-                                        transpose_type,
-                                        enable_callbacks,
-                                        node.IsScalingEnabled(),
-                                        node.fuseBlue);
+                                        node.sbrcTranstype,
+                                        node.GetCallbackType(enable_callbacks),
+                                        node.fuseBlue,
+                                        node.loadOps,
+                                        node.storeOps);
     };
 
     // if is pre-compiled, we assign the name-function only
@@ -216,10 +213,11 @@ RTCKernel::RTCGenerator RTCKernelStockham::generate_from_node(const TreeNode&
                             node.ebtype,
                             node.dir2regMode,
                             node.intrinsicMode,
-                            transpose_type,
-                            enable_callbacks,
-                            node.IsScalingEnabled(),
-                            node.fuseBlue);
+                            node.sbrcTranstype,
+                            node.GetCallbackType(enable_callbacks),
+                            node.fuseBlue,
+                            node.loadOps,
+                            node.storeOps);
     };
 
     generator.construct_rtckernel
@@ -268,6 +266,9 @@ RTCKernelArgs RTCKernelStockham::get_launch_args(DeviceCallIn& data)
         if(array_type_is_planar(data.node->outArrayType))
             kargs.append_ptr(data.bufOut[1]);
     }
+
+    append_load_store_args(kargs, *data.node);
+
     // fused bluestein data (chirp table and lengths)
     switch(data.node->fuseBlue)
     {
@@ -343,6 +344,5 @@ RTCKernelArgs RTCKernelStockham::get_launch_args(DeviceCallIn& data)
             }
         }
     }
-
     return kargs;
 }
