@@ -1,4 +1,4 @@
-// Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2022-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -21,6 +21,7 @@
 #include "rtc_transpose_gen.h"
 #include "../../shared/array_predicate.h"
 #include "device/generator/generator.h"
+#include "rtc_test_harness.h"
 
 #include "device/kernel-generator-embed.h"
 
@@ -67,10 +68,12 @@ std::string transpose_rtc_kernel_name(const TransposeSpecs& specs)
         kernel_name += "_diag";
     if(specs.tileAligned)
         kernel_name += "_aligned";
-    if(specs.enable_callbacks)
-        kernel_name += "_CB";
-    if(specs.enable_scaling)
-        kernel_name += "_scale";
+    kernel_name += load_store_name_suffix(specs.loadOps, specs.storeOps);
+    kernel_name += rtc_cbtype_name(specs.cbtype);
+
+    if(!specs.grid3D)
+        kernel_name += "_grid1D";
+
     return kernel_name;
 }
 
@@ -84,9 +87,9 @@ std::string transpose_rtc(const std::string& kernel_name, const TransposeSpecs&
     src += common_h;
     src += callback_h;
 
-    src += rtc_precision_type_decl(specs.precision);
+    src += rtc_precision_type_decl(specs.precision, array_type_is_complex(specs.inArrayType));
 
-    src += rtc_const_cbtype_decl(specs.enable_callbacks);
+    src += rtc_const_cbtype_decl(specs.cbtype);
 
     // twiddle code assumes scalar type is named T
     src += "typedef scalar_type T;\n";
@@ -110,7 +113,9 @@ std::string transpose_rtc(const std::string& kernel_name, const TransposeSpecs&
     Variable stride_out2_var{"stride_out2", "unsigned int"};
     Variable stride_out_var{"stride_out", "const size_t", true, true};
     Variable odist_var{"odist", "unsigned int"};
-    Variable scale_factor_var{"scale_factor", "const real_type_t<scalar_type>"};
+    Variable gridX{"gridX", "const unsigned int"};
+    Variable gridY{"gridY", "const unsigned int"};
+    Variable gridZ{"gridZ", "const unsigned int"};
 
     Function func(kernel_name);
     func.launch_bounds = specs.tileX * specs.tileY;
@@ -134,9 +139,12 @@ std::string transpose_rtc(const std::string& kernel_name, const TransposeSpecs&
     func.arguments.append(stride_out2_var);
     func.arguments.append(stride_out_var);
     func.arguments.append(odist_var);
+    func.arguments.append(gridX);
+    func.arguments.append(gridY);
+    func.arguments.append(gridZ);
+
     for(const auto& arg : get_callback_args().arguments)
         func.arguments.append(arg);
-    func.arguments.append(scale_factor_var);
 
     // we use tileX*tileX tiles - tileY must evenly divide into
     // tileX, so that elems_per_thread is integral
@@ -151,19 +159,58 @@ std::string transpose_rtc(const std::string& kernel_name, const TransposeSpecs&
     lds.size2D = Literal{specs.tileX};
     func.body += Declaration{lds};
 
+    // recover a 3D GPU grid layout
+    func.body += CommentLines{"since gridDim is passed as {gridX, 1, 1}, use the",
+                              "following variables to recover block indices in a 3-D fashion:"};
+
+    Variable old_blockIdx_x{"old_blockIdx_x", "unsigned int"};
+    Variable old_blockIdx_y{"old_blockIdx_y", "unsigned int"};
+    Variable old_blockIdx_z{"old_blockIdx_z", "unsigned int"};
+
     Variable tileBlockIdx_y{"tileBlockIdx_y", "unsigned int"};
     Variable tileBlockIdx_x{"tileBlockIdx_x", "unsigned int"};
-    func.body += Declaration{tileBlockIdx_y, "blockIdx.y"};
-    func.body += Declaration{tileBlockIdx_x, "blockIdx.x"};
+
+    Variable remaining{"remaining", "unsigned int"};
+
+    // if a 1-D grid was provided because creating a natural 3-D grid exceeded allowed limits, then remap it to a 3-D grid.
+    if(!specs.grid3D)
+    {
+        func.body += Declaration{old_blockIdx_x, Literal{"blockIdx.x"} / (gridY * gridZ)};
+        func.body += Declaration{remaining, Literal{"blockIdx.x"} % (gridY * gridZ)};
+        func.body += Declaration{old_blockIdx_y, (remaining / gridZ)};
+        func.body += Declaration{old_blockIdx_z, Literal{"blockIdx.x"} % gridZ};
+
+        func.body += Declaration{tileBlockIdx_y, "old_blockIdx_y"};
+        func.body += Declaration{tileBlockIdx_x, "old_blockIdx_x"};
+    }
+    else
+    {
+        func.body += Declaration{tileBlockIdx_y, "blockIdx.y"};
+        func.body += Declaration{tileBlockIdx_x, "blockIdx.x"};
+    }
+
+    func.body += LineBreak{};
 
     if(specs.diagonal)
     {
         Variable bid{"bid", "auto"};
-        func.body += Declaration{bid, "blockIdx.x + gridDim.x * blockIdx.y"};
-        func.body += Assign{tileBlockIdx_y, bid % Variable{"gridDim.y", ""}};
-        func.body += Assign{tileBlockIdx_x,
-                            (bid / Variable{"gridDim.y", ""} + tileBlockIdx_y)
-                                % Variable{"gridDim.x", ""}};
+
+        if(specs.grid3D)
+        {
+            func.body += Declaration{bid, "blockIdx.x + gridDim.x * blockIdx.y"};
+            func.body += Assign{tileBlockIdx_y, bid % Variable{"gridDim.y", ""}};
+            func.body += Assign{tileBlockIdx_x,
+                                (bid / Variable{"gridDim.y", ""} + tileBlockIdx_y)
+                                    % Variable{"gridDim.x", ""}};
+        }
+        else
+        {
+            func.body += Declaration{bid, "old_blockIdx_x + gridX * old_blockIdx_y"};
+            func.body += Assign{tileBlockIdx_y, bid % Variable{"gridY", ""}};
+            func.body
+                += Assign{tileBlockIdx_x,
+                          (bid / Variable{"gridY", ""} + tileBlockIdx_y) % Variable{"gridX", ""}};
+        }
     }
 
     if(specs.dim == 2)
@@ -179,10 +226,16 @@ std::string transpose_rtc(const std::string& kernel_name, const TransposeSpecs&
     func.body += Declaration{tile_y_index, "threadIdx.y"};
 
     func.body += CommentLines{"work out offset for dimensions after the first 3"};
-    Variable remaining{"remaining", "unsigned int"};
     Variable offset_in{"offset_in", "unsigned int"};
     Variable offset_out{"offset_out", "unsigned int"};
-    func.body += Declaration{remaining, "blockIdx.z"};
+    if(specs.grid3D)
+    {
+        func.body += Declaration{remaining, "blockIdx.z"};
+    }
+    else
+    {
+        func.body += Assign{remaining, "old_blockIdx_z"};
+    }
     func.body += Declaration{offset_in, 0};
     func.body += Declaration{offset_out, 0};
 
@@ -208,7 +261,8 @@ std::string transpose_rtc(const std::string& kernel_name, const TransposeSpecs&
     func.body += CommentLines{"remaining is now the batch"};
     func.body += AddAssign(offset_in, remaining * idist_var);
     func.body += AddAssign(offset_out, remaining * odist_var);
-    func.body += CallbackDeclaration("scalar_type", "cbtype");
+    func.body += CallbackLoadDeclaration("scalar_type", "cbtype");
+    func.body += CallbackStoreDeclaration("scalar_type", "cbtype");
 
     // loop variables for reading/writing
     Variable i{"i", "unsigned int"};
@@ -307,19 +361,22 @@ std::string transpose_rtc(const std::string& kernel_name, const TransposeSpecs&
     write_loop.body += Declaration{global_write_idx,
                                    idx0 * stride_out0_var + idx1 * stride_out1_var
                                        + idx2 * stride_out2_var + offset_out};
-    if(specs.enable_scaling)
-        write_loop.body += MultiplyAssign(val[i], scale_factor_var);
-
     write_loop.body += StoreGlobal{output_var, global_write_idx, val[i]};
 
     func.body += write_loop;
 
+    make_load_store_ops(func, specs.loadOps, specs.storeOps);
+
     if(array_type_is_planar(specs.inArrayType))
         func = make_planar(func, "input");
     if(array_type_is_planar(specs.outArrayType))
         func = make_planar(func, "output");
 
+    func = make_callback_realcomplex(func, specs.cbtype);
+
     src += func.render();
 
+    write_standalone_test_harness(func, src);
+
     return src;
 }
