@@ -19,20 +19,37 @@
 // THE SOFTWARE.
 
 #include "tree_node_1D.h"
+#include "../../shared/arithmetic.h"
 #include "../../shared/precision_type.h"
 #include "../device/kernels/bank_shift.h"
 #include "function_pool.h"
 #include "fuse_shim.h"
 #include "node_factory.h"
+#include "repo.h"
 #include "tuning_helper.h"
 #include <numeric>
 
 /*****************************************************
  * L1D_TRTRT  *
  *****************************************************/
-void TRTRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void TRTRT1DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    ComputeScheme determined_scheme_node1 = CS_NONE;
+    if(!noSolution)
+    {
+        if((child_scheme_trees.size() != 5)
+           || (child_scheme_trees[0]->curScheme != CS_KERNEL_TRANSPOSE)
+           || (child_scheme_trees[2]->curScheme != CS_KERNEL_TRANSPOSE)
+           || (child_scheme_trees[3]->curScheme != CS_KERNEL_STOCKHAM)
+           || (child_scheme_trees[4]->curScheme != CS_KERNEL_TRANSPOSE))
+        {
+            throw std::runtime_error("TRTRT1DNode: Unexpected child scheme from solution map");
+        }
+        determined_scheme_node1 = child_scheme_trees[1]->curScheme;
+    }
 
     size_t lenFactor1 = length.back();
     size_t lenFactor0 = length[0] / lenFactor1;
@@ -41,8 +58,6 @@ void TRTRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
     length.pop_back();
 
     // first transpose
-    if(!noSolution)
-        assert(child_schemes[0] == CS_KERNEL_TRANSPOSE);
     auto trans1Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     trans1Plan->length.push_back(lenFactor0);
     trans1Plan->length.push_back(lenFactor1);
@@ -63,15 +78,11 @@ void TRTRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
         row1PlanData.length.push_back(length[index]);
     }
 
-    // skip the decide scheme part in node factory
-    ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[1];
-    auto          row1Plan = NodeFactory::CreateExplicitNode(row1PlanData, this, determined_scheme);
-    row1Plan->large1D      = 0;
-    row1Plan->RecursiveBuildTree();
+    auto row1Plan = NodeFactory::CreateExplicitNode(row1PlanData, this, determined_scheme_node1);
+    row1Plan->large1D = 0;
+    row1Plan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[1].get());
 
     // second transpose
-    if(!noSolution)
-        assert(child_schemes[2] == CS_KERNEL_TRANSPOSE);
     auto trans2Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     trans2Plan->length.push_back(lenFactor1);
     trans2Plan->length.push_back(lenFactor0);
@@ -84,8 +95,6 @@ void TRTRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
     trans2Plan->SetTransposeOutputLength();
 
     // second row fft
-    if(!noSolution)
-        assert(child_schemes[3] == CS_KERNEL_STOCKHAM);
     auto row2Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM, this);
     row2Plan->length.push_back(lenFactor0);
     row2Plan->length.push_back(lenFactor1);
@@ -96,8 +105,6 @@ void TRTRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
     }
 
     // third transpose
-    if(!noSolution)
-        assert(child_schemes[4] == CS_KERNEL_TRANSPOSE);
     auto trans3Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     trans3Plan->length.push_back(lenFactor0);
     trans3Plan->length.push_back(lenFactor1);
@@ -311,9 +318,20 @@ void TRTRT1DNode::AssignParams_internal()
 /*****************************************************
  * L1D_CC  *
  *****************************************************/
-void CC1DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void CC1DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    if(!noSolution)
+    {
+        if((child_scheme_trees.size() != 2)
+           || (child_scheme_trees[0]->curScheme != CS_KERNEL_STOCKHAM_BLOCK_CC)
+           || (child_scheme_trees[1]->curScheme != CS_KERNEL_STOCKHAM_BLOCK_RC))
+        {
+            throw std::runtime_error("CC1DNode: Unexpected child scheme from solution map");
+        }
+    }
 
     size_t lenFactor1 = length.back();
     size_t lenFactor0 = length[0] / lenFactor1;
@@ -322,8 +340,6 @@ void CC1DNode::BuildTree_internal(const SchemeVec& child_schemes)
     length.pop_back();
 
     // first plan, column-to-column
-    if(!noSolution)
-        assert(child_schemes[0] == CS_KERNEL_STOCKHAM_BLOCK_CC);
     auto col2colPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
 
     col2colPlan->typeBlue = typeBlue;
@@ -349,8 +365,6 @@ void CC1DNode::BuildTree_internal(const SchemeVec& child_schemes)
     std::swap(col2colPlan->outputLength[0], col2colPlan->outputLength[1]);
 
     // second plan, row-to-column
-    if(!noSolution)
-        assert(child_schemes[1] == CS_KERNEL_STOCKHAM_BLOCK_RC);
     auto row2colPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_RC, this);
 
     row2colPlan->typeBlue = typeBlue;
@@ -663,9 +677,21 @@ void CC1DNode::AssignParams_internal()
 /*****************************************************
  * L1D_CRT  *
  *****************************************************/
-void CRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void CRT1DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    if(!noSolution)
+    {
+        if((child_scheme_trees.size() != 3)
+           || (child_scheme_trees[0]->curScheme != CS_KERNEL_STOCKHAM_BLOCK_CC)
+           || (child_scheme_trees[1]->curScheme != CS_KERNEL_STOCKHAM)
+           || (child_scheme_trees[2]->curScheme != CS_KERNEL_TRANSPOSE))
+        {
+            throw std::runtime_error("CRT1DNode: Unexpected child scheme from solution map");
+        }
+    }
 
     size_t lenFactor1 = length.back();
     size_t lenFactor0 = length[0] / lenFactor1;
@@ -674,8 +700,6 @@ void CRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
     length.pop_back();
 
     // first plan, column-to-column
-    if(!noSolution)
-        assert(child_schemes[0] == CS_KERNEL_STOCKHAM_BLOCK_CC);
     auto col2colPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
     // large1D flag to confirm we need multiply twiddle factor
     col2colPlan->large1D = length[0];
@@ -690,8 +714,6 @@ void CRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
     std::swap(col2colPlan->outputLength[0], col2colPlan->outputLength[1]);
 
     // second plan, row-to-row
-    if(!noSolution)
-        assert(child_schemes[1] == CS_KERNEL_STOCKHAM);
     auto row2rowPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM, this);
     row2rowPlan->length.push_back(lenFactor0);
     row2rowPlan->length.push_back(lenFactor1);
@@ -704,8 +726,6 @@ void CRT1DNode::BuildTree_internal(const SchemeVec& child_schemes)
     // row2rowPlan->allowOutofplace = false;
 
     // third plan, transpose
-    if(!noSolution)
-        assert(child_schemes[2] == CS_KERNEL_TRANSPOSE);
     auto transPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     transPlan->length.push_back(lenFactor0);
     transPlan->length.push_back(lenFactor1);
@@ -888,6 +908,19 @@ void Stockham1DNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
     if(ebtype != EmbeddedType::NONE)
         lds_padding = 1;
 
+    if(applyPartialPass)
+    {
+        // Special case for partial pass 64 x 64 x 64.
+        // Kernel configuration is hardcoded for now.
+        // TODO: Once the partial-pass kernels are properly
+        // integrated into the Stockham kernel generators,
+        // this configuration will come from the usual location
+        // in kernel-generator.py.
+        kernel.threads_per_transform[0] = 8;
+        kernel.workgroup_size           = 128;
+        kernel.transforms_per_block     = kernel.workgroup_size / kernel.threads_per_transform[0];
+    }
+
     bwd      = kernel.transforms_per_block;
     wgs      = kernel.workgroup_size;
     gp.b_x   = (batch_accum + bwd - 1) / bwd;
@@ -905,19 +938,29 @@ void Stockham1DNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
         //   When lds conflict becomes significant enough, we can apply lds bank shift to reduce it.
         //   One of the costs is extra lds allocation. We enable it for small pow of 2 cases on all
         //   supported archs for now.
-        if(length[0] == 64)
-        {
+        //   Bank-shift not supported in partial pass mode.
+        if(length[0] == 64 && !applyPartialPass)
             lds = (length[0] + lds_padding) * bwd + length[0] * bwd / LDS_BANK_SHIFT;
-        }
         else
-        {
             lds = (length[0] + lds_padding) * bwd;
-        }
     }
 }
 
 bool Stockham1DNode::CreateDeviceResources()
 {
+    if(applyPartialPass)
+    {
+        // handles partial pass 64 x 64 x 64 case.
+        // current dimension y is the dimension to split
+        // into x and z.
+
+        // Create twiddle table for partial pass along y
+        size_t pp_dim = 1;
+
+        std::tie(twiddles_pp, twiddles_pp_size)
+            = Repo::GetTwiddlesPP(length[pp_dim], precision, deviceProp);
+    }
+
     twd_attach_halfN = (ebtype != EmbeddedType::NONE);
     return LeafNode::CreateDeviceResources();
 }
@@ -928,6 +971,10 @@ std::vector<size_t> Stockham1DNode::CollapsibleDims()
     if(typeBlue == BT_MULTI_KERNEL_FUSED)
         return {};
 
+    // do not collapse on partial-pass nodes
+    if(applyPartialPass)
+        return {};
+
     // fastest dim is FFT, the rest is collapsible
     std::vector<size_t> ret(length.size() - 1);
     std::iota(ret.begin(), ret.end(), 1);
@@ -1026,7 +1073,7 @@ void SBCCNode::InitIntrinsicMode()
     // case 1: is runing tuning or a tuned solution, then use the setting in the config
     if(specified_key != nullptr)
     {
-        auto& config  = std::get<4>(*specified_key.get());
+        auto& config  = (*specified_key.get()).kernel_config;
         intrinsicMode = (config.intrinsic_buffer_inst) ? IntrinsicAccessType::ENABLE_BOTH
                                                        : IntrinsicAccessType::DISABLE_BOTH;
         return;
@@ -1087,10 +1134,36 @@ void SBCCNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
     fnPtr       = kernel.device_function;
     bwd         = kernel.transforms_per_block;
     wgs         = kernel.workgroup_size;
-    lds         = length[0] * bwd;
-    gp.b_x      = ((length[1]) - 1) / bwd + 1;
-    gp.b_x *= std::accumulate(length.begin() + 2, length.end(), batch, std::multiplies<size_t>());
+
+    if(applyPartialPass)
+    {
+        // Special case for partial pass 64 x 64 x 64.
+        // Kernel configration is hardcoded for now.
+        // TODO: Once the partial-pass kernels are integrated
+        // into the stockham kernel generators, change this
+        // configuration to use kernel-generator.py data.
+        auto tpt = 8;
+        wgs      = 64;
+        bwd      = wgs / tpt;
+    }
+
+    lds = length[0] * bwd;
+
+    gp.b_x = ((length[1]) - 1) / bwd + 1;
+    gp.b_x *= product(length.begin() + 2, length.end()) * batch;
     gp.wgs_x = wgs;
+
+    if(applyPartialPass)
+    {
+        // grid and thread organization is different
+        // on partial pass sbcc kernels (for improved
+        // global memory access patterns).
+        auto factor = *std::max_element(kernelFactorsPP.begin(), kernelFactorsPP.end());
+
+        gp.b_x /= factor;
+        gp.wgs_x *= factor;
+        lds *= factor;
+    }
 }
 
 std::vector<size_t> SBCCNode::CollapsibleDims()
@@ -1117,7 +1190,7 @@ FMKey SBRCNode::GetKernelKey() const
     if(sbrcTranstype == SBRC_TRANSPOSE_TYPE::NONE)
     {
         // find the base kernel at first
-        FMKey baseKey = fpkey(length[0], precision, scheme, TILE_ALIGNED);
+        FMKey baseKey(length[0], precision, scheme, TILE_ALIGNED);
         // if we have the base kernel, then we set the exact sbrc_trans_type and return the real key
         // if we don't, then we simply return a key with NONE sbrc_trans_type
         // which will make KernelCheck() trigger an exception
@@ -1128,7 +1201,7 @@ FMKey SBRCNode::GetKernelKey() const
         }
     }
 
-    return fpkey(length[0], precision, scheme, sbrcTranstype);
+    return FMKey(length[0], precision, scheme, sbrcTranstype);
 }
 
 bool SBRCNode::KernelCheck(std::vector<FMKey>& kernel_keys)
@@ -1203,7 +1276,7 @@ void SBRCNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
     wgs         = kernel.workgroup_size;
     lds         = length[0] * bwd;
     gp.b_x      = (length[1] - 1) / bwd + 1;
-    gp.b_x *= std::accumulate(length.begin() + 2, length.end(), batch, std::multiplies<size_t>());
+    gp.b_x *= product(length.begin() + 2, length.end()) * batch;
     gp.wgs_x = wgs;
 }
 
@@ -1283,7 +1356,7 @@ void SBCRNode::InitIntrinsicMode()
     // case 1: is runing tuning or a tuned solution, then use the setting in the config
     if(specified_key != nullptr)
     {
-        auto& config  = std::get<4>(*specified_key.get());
+        auto& config  = (*specified_key.get()).kernel_config;
         intrinsicMode = (config.intrinsic_buffer_inst) ? IntrinsicAccessType::ENABLE_BOTH
                                                        : IntrinsicAccessType::DISABLE_BOTH;
         return;
@@ -1320,7 +1393,7 @@ void SBCRNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
     bwd         = kernel.transforms_per_block;
     lds         = length[0] * bwd;
     gp.b_x      = ((length[1]) - 1) / bwd + 1;
-    gp.b_x *= std::accumulate(length.begin() + 2, length.end(), batch, std::multiplies<size_t>());
+    gp.b_x *= product(length.begin() + 2, length.end()) * batch;
     gp.wgs_x = wgs;
 
     if(ebtype != EmbeddedType::NONE)
