@@ -1,4 +1,4 @@
-// Copyright (C) 2021 - 2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2021 - 2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -28,9 +28,24 @@
 /*****************************************************
  * 2D_RTRT  *
  *****************************************************/
-void RTRT2DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void RTRT2DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    ComputeScheme determined_scheme_node0 = CS_NONE;
+    ComputeScheme determined_scheme_node2 = CS_NONE;
+    if(!noSolution)
+    {
+        if((child_scheme_trees.size() != 4)
+           || (child_scheme_trees[1]->curScheme != CS_KERNEL_TRANSPOSE)
+           || (child_scheme_trees[3]->curScheme != CS_KERNEL_TRANSPOSE))
+        {
+            throw std::runtime_error("RTRT2DNode: Unexpected child scheme from solution map");
+        }
+        determined_scheme_node0 = child_scheme_trees[0]->curScheme;
+        determined_scheme_node2 = child_scheme_trees[2]->curScheme;
+    }
 
     // first row fft
     NodeMetaData row1PlanData(this);
@@ -41,14 +56,10 @@ void RTRT2DNode::BuildTree_internal(const SchemeVec& child_schemes)
     {
         row1PlanData.length.push_back(length[index]);
     }
-    // skip the decide scheme part in node factory
-    ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[0];
-    auto          row1Plan = NodeFactory::CreateExplicitNode(row1PlanData, this, determined_scheme);
-    row1Plan->RecursiveBuildTree();
+    auto row1Plan = NodeFactory::CreateExplicitNode(row1PlanData, this, determined_scheme_node0);
+    row1Plan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[0].get());
 
     // first transpose
-    if(!noSolution)
-        assert(child_schemes[1] == CS_KERNEL_TRANSPOSE);
     auto trans1Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     trans1Plan->length.push_back(length[0]);
     trans1Plan->length.push_back(length[1]);
@@ -68,14 +79,10 @@ void RTRT2DNode::BuildTree_internal(const SchemeVec& child_schemes)
     {
         row2PlanData.length.push_back(length[index]);
     }
-    // skip the decide scheme part in node factory
-    determined_scheme = (noSolution) ? CS_NONE : child_schemes[2];
-    auto row2Plan     = NodeFactory::CreateExplicitNode(row2PlanData, this, determined_scheme);
-    row2Plan->RecursiveBuildTree();
+    auto row2Plan = NodeFactory::CreateExplicitNode(row2PlanData, this, determined_scheme_node2);
+    row2Plan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[2].get());
 
     // second transpose
-    if(!noSolution)
-        assert(child_schemes[3] == CS_KERNEL_TRANSPOSE);
     auto trans2Plan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE, this);
     trans2Plan->length.push_back(length[1]);
     trans2Plan->length.push_back(length[0]);
@@ -194,9 +201,21 @@ void RTRT2DNode::AssignParams_internal()
 /*****************************************************
  * 2D_RC  *
  *****************************************************/
-void RC2DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void RC2DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    ComputeScheme determined_scheme_node0 = CS_NONE;
+    if(!noSolution)
+    {
+        if((child_scheme_trees.size() != 2)
+           || (child_scheme_trees[1]->curScheme != CS_KERNEL_STOCKHAM_BLOCK_CC))
+        {
+            throw std::runtime_error("RC2DNode: Unexpected child scheme from solution map");
+        }
+        determined_scheme_node0 = child_scheme_trees[0]->curScheme;
+    }
 
     // row fft
     NodeMetaData rowPlanData(this);
@@ -207,14 +226,10 @@ void RC2DNode::BuildTree_internal(const SchemeVec& child_schemes)
     {
         rowPlanData.length.push_back(length[index]);
     }
-    // skip the decide scheme part in node factory
-    ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[0];
-    auto          rowPlan = NodeFactory::CreateExplicitNode(rowPlanData, this, determined_scheme);
-    rowPlan->RecursiveBuildTree();
+    auto rowPlan = NodeFactory::CreateExplicitNode(rowPlanData, this, determined_scheme_node0);
+    rowPlan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[0].get());
 
     // column fft
-    if(!noSolution)
-        assert(child_schemes[1] == CS_KERNEL_STOCKHAM_BLOCK_CC);
     auto colPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
     colPlan->length.push_back(length[1]);
     colPlan->dimension = 1;
@@ -240,10 +255,18 @@ void RC2DNode::AssignParams_internal()
     rowPlan->inStride = inStride;
     rowPlan->iDist    = iDist;
 
-    // row plan is in-place, so keep same strides in case parent's
-    // in/out strides are incompatible for the same buffer
-    rowPlan->outStride = inStride;
-    rowPlan->oDist     = iDist;
+    // make output of row plan contiguous if the 2D node is at the root of the plan tree
+    if(parent == nullptr)
+    {
+        rowPlan->outStride = {1, length[0]};
+        rowPlan->oDist     = length[0] * length[1];
+    }
+    else
+    {
+        rowPlan->outStride = inStride;
+        rowPlan->oDist     = iDist;
+    }
+
     rowPlan->AssignParams();
 
     colPlan->inStride = rowPlan->outStride;
@@ -262,9 +285,31 @@ void RC2DNode::AssignParams_internal()
  *****************************************************/
 bool Single2DNode::CreateDeviceResources()
 {
+    // we need to put radices1/2 into key for twd_table, since when offline tuning,
+    // we are generating kernels with different radices, so for different radices,
+    // twd_table can't be shared
+    std::vector<size_t> radices1, radices2;
+
+    int    count               = 0;
+    size_t cummulative_product = 1;
+    while(cummulative_product != length[0])
+    {
+        cummulative_product *= kernelFactors[count++];
+    }
+    radices1.insert(radices1.cbegin(), kernelFactors.cbegin(), kernelFactors.cbegin() + count);
+    radices2.insert(radices2.cbegin(), kernelFactors.cbegin() + count, kernelFactors.cend());
+
+    twd_attach_halfN  = (ebtype == EmbeddedType::Real2C_POST);
+    twd_attach_halfN2 = (ebtype == EmbeddedType::C2Real_PRE);
     // create one set of twiddles for each dimension
-    std::tie(twiddles, twiddles_size)
-        = Repo::GetTwiddles2D(length[0], length[1], precision, deviceProp.gcnArchName);
+    std::tie(twiddles, twiddles_size) = Repo::GetTwiddles2D(length[0],
+                                                            length[1],
+                                                            precision,
+                                                            deviceProp,
+                                                            twd_attach_halfN,
+                                                            twd_attach_halfN2,
+                                                            radices1,
+                                                            radices2);
 
     return CreateLargeTwdTable();
 }
@@ -279,9 +324,27 @@ void Single2DNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp)
     gp.b_x   = (batch + bwd - 1) / bwd;
     gp.wgs_x = wgs;
 
+    size_t padded_len0 = length[0];
+    size_t padded_len1 = length[1];
+
+    if(ebtype == EmbeddedType::Real2C_POST)
+        padded_len0 += 1;
+    if(ebtype == EmbeddedType::C2Real_PRE)
+        padded_len1 += 1;
+
     // if fastest length is power of 2, pad it to avoid LDS bank conflicts
-    auto padded_len0 = IsPo2(length[0]) ? length[0] + 1 : length[0];
-    lds              = padded_len0 * length[1] * bwd;
+    if(ebtype != EmbeddedType::C2Real_PRE)
+    {
+        // for EBTYPE NONE/POST, the fastest length is length0
+        padded_len0 = IsPo2(padded_len0) ? padded_len0 + 1 : padded_len0;
+    }
+    else
+    {
+        // for EBTYPE PRE, the fastest length is length1
+        padded_len1 = IsPo2(padded_len1) ? padded_len1 + 1 : padded_len1;
+    }
+
+    lds = padded_len0 * padded_len1 * bwd;
 
     // if we're doing 3D transform, we need to repeat the 2D
     // transform in the 3rd dimension
