@@ -29,22 +29,33 @@
 /*****************************************************
  * 3D_RTRT  *
  *****************************************************/
-void RTRT3DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void RTRT3DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    ComputeScheme determined_scheme_node0 = CS_NONE;
+    ComputeScheme determined_scheme_node2 = CS_NONE;
+    if(!noSolution)
+    {
+        if((child_scheme_trees.size() != 4)
+           || (child_scheme_trees[1]->curScheme != CS_KERNEL_TRANSPOSE_XY_Z)
+           || (child_scheme_trees[3]->curScheme != CS_KERNEL_TRANSPOSE_Z_XY))
+        {
+            throw std::runtime_error("RTRT3DNode: Unexpected child scheme from solution map");
+        }
+        determined_scheme_node0 = child_scheme_trees[0]->curScheme;
+        determined_scheme_node2 = child_scheme_trees[2]->curScheme;
+    }
 
     // 2d fft
     NodeMetaData xyPlanData(this);
     xyPlanData.length    = length;
     xyPlanData.dimension = 2;
-    // skip the decide scheme part in node factory
-    ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[0];
-    auto          xyPlan = NodeFactory::CreateExplicitNode(xyPlanData, this, determined_scheme);
-    xyPlan->RecursiveBuildTree();
+    auto xyPlan = NodeFactory::CreateExplicitNode(xyPlanData, this, determined_scheme_node0);
+    xyPlan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[0].get());
 
     // first transpose
-    if(!noSolution)
-        assert(child_schemes[1] == CS_KERNEL_TRANSPOSE_XY_Z);
     auto trans1Plan    = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE_XY_Z, this);
     trans1Plan->length = length;
     trans1Plan->SetTransposeOutputLength();
@@ -57,14 +68,10 @@ void RTRT3DNode::BuildTree_internal(const SchemeVec& child_schemes)
     zPlanData.length.push_back(length[2]);
     zPlanData.length.push_back(length[0]);
     zPlanData.length.push_back(length[1]);
-    // skip the decide scheme part in node factory
-    determined_scheme = determined_scheme = (noSolution) ? CS_NONE : child_schemes[2];
-    auto zPlan = NodeFactory::CreateExplicitNode(zPlanData, this, determined_scheme);
-    zPlan->RecursiveBuildTree();
+    auto zPlan = NodeFactory::CreateExplicitNode(zPlanData, this, determined_scheme_node2);
+    zPlan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[2].get());
 
     // second transpose
-    if(!noSolution)
-        assert(child_schemes[3] == CS_KERNEL_TRANSPOSE_Z_XY);
     auto trans2Plan    = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE_Z_XY, this);
     trans2Plan->length = zPlan->length;
     trans2Plan->SetTransposeOutputLength();
@@ -154,17 +161,27 @@ void RTRT3DNode::AssignParams_internal()
 /*****************************************************
  * 3D_TRTRTR  *
  *****************************************************/
-void TRTRTR3DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void TRTRTR3DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    if(!noSolution)
+    {
+        if((child_scheme_trees.size() != 6)
+           || (child_scheme_trees[0]->curScheme != CS_KERNEL_TRANSPOSE_Z_XY)
+           || (child_scheme_trees[2]->curScheme != CS_KERNEL_TRANSPOSE_Z_XY)
+           || (child_scheme_trees[4]->curScheme != CS_KERNEL_TRANSPOSE_Z_XY))
+        {
+            throw std::runtime_error("TRTRTR3DNode: Unexpected child scheme from solution map");
+        }
+    }
 
     std::vector<size_t> cur_length = length;
 
     for(int i = 0; i < 6; i += 2)
     {
         // transpose Z_XY
-        if(!noSolution)
-            assert(child_schemes[i] == CS_KERNEL_TRANSPOSE_Z_XY);
         auto trans_plan    = NodeFactory::CreateNodeFromScheme(CS_KERNEL_TRANSPOSE_Z_XY, this);
         trans_plan->length = cur_length;
         trans_plan->SetTransposeOutputLength();
@@ -178,9 +195,10 @@ void TRTRTR3DNode::BuildTree_internal(const SchemeVec& child_schemes)
         row_plan_data.length    = cur_length;
         row_plan_data.dimension = 1;
         // skip the decide scheme part in node factory
-        ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[i + 1];
+        ComputeScheme determined_scheme
+            = (noSolution) ? CS_NONE : child_scheme_trees[i + 1]->curScheme;
         auto row_plan = NodeFactory::CreateExplicitNode(row_plan_data, this, determined_scheme);
-        row_plan->RecursiveBuildTree();
+        row_plan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[i + 1].get());
 
         // TR
         childNodes.emplace_back(std::move(trans_plan));
@@ -281,9 +299,9 @@ void TRTRTR3DNode::AssignParams_internal()
 /*****************************************************
  * CS_3D_BLOCK_RC  *
  *****************************************************/
-void BLOCKRC3DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void BLOCKRC3DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
 
     std::vector<size_t> cur_length = length;
 
@@ -336,7 +354,7 @@ void BLOCKRC3DNode::BuildTree_internal(const SchemeVec& child_schemes)
         if(have_sbrc)
         {
             auto kernel = function_pool::get_kernel(
-                fpkey(cur_length[0], precision, CS_KERNEL_STOCKHAM_BLOCK_RC, TILE_ALIGNED));
+                FMKey(cur_length[0], precision, CS_KERNEL_STOCKHAM_BLOCK_RC, TILE_ALIGNED));
 
             size_t otherDim = use_ZXY_sbrc ? cur_length[1] : cur_length[2];
             if(otherDim % kernel.transforms_per_block != 0)
@@ -365,8 +383,9 @@ void BLOCKRC3DNode::BuildTree_internal(const SchemeVec& child_schemes)
         {
             auto sbrcScheme = (use_ZXY_sbrc) ? CS_KERNEL_STOCKHAM_TRANSPOSE_Z_XY
                                              : CS_KERNEL_STOCKHAM_TRANSPOSE_XY_Z;
-            if(!noSolution)
-                assert(child_schemes[childNodes.size()] == sbrcScheme);
+            if(!noSolution && (child_scheme_trees[childNodes.size()]->curScheme != sbrcScheme))
+                throw std::runtime_error(
+                    "BLOCKRC3DNode: Unexpected child scheme from solution map");
             auto sbrc_node    = NodeFactory::CreateNodeFromScheme(sbrcScheme, this);
             sbrc_node->length = cur_length;
             sbrc_node->SetTransposeOutputLength();
@@ -379,14 +398,16 @@ void BLOCKRC3DNode::BuildTree_internal(const SchemeVec& child_schemes)
             row_plan_data.length    = cur_length;
             row_plan_data.dimension = 1;
             ComputeScheme determined_scheme
-                = (noSolution) ? CS_NONE : child_schemes[childNodes.size()];
+                = (noSolution) ? CS_NONE : child_scheme_trees[childNodes.size()]->curScheme;
             auto row_plan = NodeFactory::CreateExplicitNode(row_plan_data, this, determined_scheme);
-            row_plan->RecursiveBuildTree();
+            row_plan->RecursiveBuildTree(
+                (noSolution) ? nullptr : child_scheme_trees[childNodes.size()].get());
 
             // transpose XY_Z
             auto transScheme = (use_ZXY_sbrc) ? CS_KERNEL_TRANSPOSE_Z_XY : CS_KERNEL_TRANSPOSE_XY_Z;
-            if(!noSolution)
-                assert(child_schemes[childNodes.size() + 1] == transScheme);
+            if(!noSolution && (child_scheme_trees[childNodes.size() + 1]->curScheme != transScheme))
+                throw std::runtime_error(
+                    "BLOCKRC3DNode: Unexpected child scheme from solution map");
             auto trans_plan    = NodeFactory::CreateNodeFromScheme(transScheme, this);
             trans_plan->length = cur_length;
             trans_plan->SetTransposeOutputLength();
@@ -494,9 +515,21 @@ void BLOCKRC3DNode::AssignParams_internal()
 /*****************************************************
  * CS_3D_BLOCK_CR  *
  *****************************************************/
-void BLOCKCR3DNode::BuildTree_internal(const SchemeVec& child_schemes)
+void BLOCKCR3DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
 {
-    bool noSolution = child_schemes.empty();
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    if(!noSolution)
+    {
+        if((child_scheme_trees.size() != 3)
+           || (child_scheme_trees[0]->curScheme != CS_KERNEL_STOCKHAM_BLOCK_CR)
+           || (child_scheme_trees[1]->curScheme != CS_KERNEL_STOCKHAM_BLOCK_CR)
+           || (child_scheme_trees[2]->curScheme != CS_KERNEL_STOCKHAM_BLOCK_CR))
+        {
+            throw std::runtime_error("BLOCKCR3DNode: Unexpected child scheme from solution map");
+        }
+    }
 
     // TODO: It works only for 3 SBCR children nodes for now.
     //       The final logic will be similar to what SBRC has.
@@ -504,8 +537,6 @@ void BLOCKCR3DNode::BuildTree_internal(const SchemeVec& child_schemes)
     std::vector<size_t> cur_length = length;
     for(int i = 0; i < 3; ++i)
     {
-        if(!noSolution)
-            assert(child_schemes[i] == CS_KERNEL_STOCKHAM_BLOCK_CR);
         auto node = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CR, this);
         node->length.push_back(cur_length[2]);
         node->length.push_back(cur_length[0] * cur_length[1]);
@@ -550,69 +581,181 @@ void BLOCKCR3DNode::AssignParams_internal()
 /*****************************************************
  * CS_3D_RC  *
  *****************************************************/
-void RC3DNode::BuildTree_internal(const SchemeVec& child_schemes)
+bool RC3DNode::CheckPartialPassSupport()
 {
-    bool noSolution = child_schemes.empty();
-
-    // 2d fft
-    NodeMetaData xyPlanData(this);
-    xyPlanData.length.push_back(length[0]);
-    xyPlanData.length.push_back(length[1]);
-    xyPlanData.dimension = 2;
-    xyPlanData.length.push_back(length[2]);
-    for(size_t index = 3; index < length.size(); index++)
+    if(parent != nullptr)
     {
-        xyPlanData.length.push_back(length[index]);
+        // 3D_RC node is child of a real tranform node
+        // skip partial pass for now
+        return false;
     }
-    // skip the decide scheme part in node factory
-    ComputeScheme determined_scheme = (noSolution) ? CS_NONE : child_schemes[0];
-    auto          xyPlan = NodeFactory::CreateExplicitNode(xyPlanData, this, determined_scheme);
-    xyPlan->RecursiveBuildTree();
 
-    // z col fft
-    NodeMetaData zPlanData(this);
-    zPlanData.length.push_back(length[2]);
-    zPlanData.dimension = 1;
-    zPlanData.length.push_back(length[0]);
-    zPlanData.length.push_back(length[1]);
-    for(size_t index = 3; index < length.size(); index++)
+    // Partial pass is currently restricted to length 64x64x64,
+    // large enough batch sizes, unite stride, interleaved FFTs
+    // without result scaling.
+
+    // List of supported 3D sizes for partial pass.
+    typedef std::tuple<size_t, size_t, size_t> len3D_item_t;
+    const std::vector<len3D_item_t>            supportedLengthsPP = {len3D_item_t{64, 64, 64}};
+
+    bool lengthCondition = false;
+    for(const auto& lenItem : supportedLengthsPP)
     {
-        zPlanData.length.push_back(length[index]);
+        if(std::get<0>(lenItem) == length[0] && std::get<1>(lenItem) == length[1]
+           && std::get<2>(lenItem) == length[2])
+        {
+            lengthCondition = true;
+            break;
+        }
     }
-    zPlanData.outputLength = length;
 
-    // use explicit SBCC kernel if available
-    std::unique_ptr<TreeNode> zPlan;
+    // TODO: Once partial pass is property integrated into
+    // the Stockham generators, revisit these restrictions.
+    bool batchCondition = (batch >= 5);
 
-    // skip the decide scheme part in node factory
-    determined_scheme = (noSolution) ? CS_NONE : child_schemes[1];
-    if(determined_scheme)
+    size_t checkDist     = product(length.begin(), length.end());
+    bool   distCondition = (iDist == checkDist && oDist == checkDist);
+
+    bool strideCondition = (inStride[0] == 1 && outStride[0] == 1);
+
+    bool arrayTypeCondition = (inArrayType != rocfft_array_type_complex_planar
+                               && outArrayType != rocfft_array_type_complex_planar);
+
+    bool loadStoreOpsCondition = (!loadOps.enabled() && !storeOps.enabled());
+
+    return (lengthCondition && batchCondition && distCondition && strideCondition
+            && arrayTypeCondition && loadStoreOpsCondition);
+}
+
+void RC3DNode::BuildTree_internal(SchemeTreeVec& child_scheme_trees)
+{
+    bool noSolution = child_scheme_trees.empty();
+
+    // check schemes from solution map
+    ComputeScheme determined_scheme_node0 = CS_NONE;
+    ComputeScheme determined_scheme_node1 = CS_NONE;
+    if(!noSolution)
     {
-        zPlan = NodeFactory::CreateExplicitNode(zPlanData, this, determined_scheme);
-        zPlan->RecursiveBuildTree();
+        if((child_scheme_trees.size() != 2))
+            throw std::runtime_error("RC3DNode: Unexpected child scheme from solution map");
+        determined_scheme_node0 = child_scheme_trees[0]->curScheme;
+        determined_scheme_node1 = child_scheme_trees[1]->curScheme;
+    }
+
+    if(CheckPartialPassSupport())
+    {
+        // work along y will be split between x and z
+        applyPartialPass = true;
+
+        // x row fft + partial pass(es) along y
+        NodeMetaData xPartialPassPlanData(this);
+        xPartialPassPlanData.length.push_back(length[0]);
+        xPartialPassPlanData.length.push_back(length[1]);
+        // technically 1 < dimension < 2 for x node.
+        xPartialPassPlanData.dimension        = 1;
+        xPartialPassPlanData.applyPartialPass = true;
+        xPartialPassPlanData.length.push_back(length[2]);
+        for(size_t index = 3; index < length.size(); index++)
+        {
+            xPartialPassPlanData.length.push_back(length[index]);
+        }
+
+        // use explicit (modified) SBRR kernel
+        std::unique_ptr<TreeNode> xPartialPassPlan;
+
+        xPartialPassPlan            = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM, this);
+        xPartialPassPlan->length    = xPartialPassPlanData.length;
+        xPartialPassPlan->dimension = 1;
+        xPartialPassPlan->allowInplace = true;
+        xPartialPassPlan->comments.push_back("partial-pass enabled for second dimension.");
+
+        // partial pass(es) along y + z col fft
+        NodeMetaData zPartialPassPlanData(this);
+        zPartialPassPlanData.length.push_back(length[2]);
+        zPartialPassPlanData.applyPartialPass = true;
+        // technically 1 < dimension < 2 for z node.
+        zPartialPassPlanData.dimension = 1;
+        zPartialPassPlanData.length.push_back(length[0]);
+        zPartialPassPlanData.length.push_back(length[1]);
+        for(size_t index = 3; index < length.size(); index++)
+        {
+            zPartialPassPlanData.length.push_back(length[index]);
+        }
+        zPartialPassPlanData.outputLength = length;
+
+        // use explicit (modified) SBCC kernel
+        std::unique_ptr<TreeNode> zPartialPassPlan;
+
+        zPartialPassPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
+        zPartialPassPlan->length       = zPartialPassPlanData.length;
+        zPartialPassPlan->dimension    = 1;
+        zPartialPassPlan->allowInplace = false;
+        zPartialPassPlan->comments.push_back("partial-pass enabled for second dimension.");
+
+        childNodes.emplace_back(std::move(xPartialPassPlan));
+        childNodes.emplace_back(std::move(zPartialPassPlan));
     }
     else
     {
-        if(function_pool::has_SBCC_kernel(length[2], precision))
+        // 2d fft
+        NodeMetaData xyPlanData(this);
+        xyPlanData.length.push_back(length[0]);
+        xyPlanData.length.push_back(length[1]);
+        xyPlanData.dimension = 2;
+        xyPlanData.length.push_back(length[2]);
+        for(size_t index = 3; index < length.size(); index++)
         {
-            zPlan            = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
-            zPlan->length    = zPlanData.length;
-            zPlan->dimension = 1;
+            xyPlanData.length.push_back(length[index]);
+        }
+        auto xyPlan = NodeFactory::CreateExplicitNode(xyPlanData, this, determined_scheme_node0);
+        xyPlan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[0].get());
+
+        // z col fft
+        NodeMetaData zPlanData(this);
+        zPlanData.length.push_back(length[2]);
+        zPlanData.dimension = 1;
+        zPlanData.length.push_back(length[0]);
+        zPlanData.length.push_back(length[1]);
+        for(size_t index = 3; index < length.size(); index++)
+        {
+            zPlanData.length.push_back(length[index]);
+        }
+        zPlanData.outputLength = length;
+
+        // use explicit SBCC kernel if available
+        std::unique_ptr<TreeNode> zPlan;
+
+        if(determined_scheme_node1 != CS_NONE)
+        {
+            zPlan = NodeFactory::CreateExplicitNode(zPlanData, this, determined_scheme_node1);
+            zPlan->RecursiveBuildTree((noSolution) ? nullptr : child_scheme_trees[1].get());
         }
         else
         {
-            zPlan = NodeFactory::CreateExplicitNode(zPlanData, this);
-            zPlan->RecursiveBuildTree();
+            if(function_pool::has_SBCC_kernel(length[2], precision))
+            {
+                zPlan = NodeFactory::CreateNodeFromScheme(CS_KERNEL_STOCKHAM_BLOCK_CC, this);
+                zPlan->length    = zPlanData.length;
+                zPlan->dimension = 1;
+            }
+            else
+            {
+                zPlan = NodeFactory::CreateExplicitNode(zPlanData, this);
+                zPlan->RecursiveBuildTree(nullptr);
+            }
         }
-    }
 
-    // RC
-    childNodes.emplace_back(std::move(xyPlan));
-    childNodes.emplace_back(std::move(zPlan));
+        // RC
+        childNodes.emplace_back(std::move(xyPlan));
+        childNodes.emplace_back(std::move(zPlan));
+    }
 }
 
 void RC3DNode::AssignParams_internal()
 {
+    // in partial pass case:
+    // xy plan is a x row 1D-FFT + plus partial pass(es) along y
+    // z plan is partial pass(es) along y + z col 1D-FFT.
     auto& xyPlan = childNodes[0];
     auto& zPlan  = childNodes[1];
 
@@ -622,18 +765,23 @@ void RC3DNode::AssignParams_internal()
     xyPlan->outStride = outStride;
     xyPlan->oDist     = oDist;
 
+    xyPlan->applyPartialPass = applyPartialPass;
+
     xyPlan->AssignParams();
 
-    zPlan->inStride.push_back(inStride[2]);
-    zPlan->inStride.push_back(inStride[0]);
-    zPlan->inStride.push_back(inStride[1]);
+    zPlan->inStride.push_back(outStride[2]);
+    zPlan->inStride.push_back(outStride[0]);
+    zPlan->inStride.push_back(outStride[1]);
     for(size_t index = 3; index < length.size(); index++)
-        zPlan->inStride.push_back(inStride[index]);
+        zPlan->inStride.push_back(outStride[index]);
 
     zPlan->iDist = xyPlan->oDist;
 
     zPlan->outStride = zPlan->inStride;
     zPlan->oDist     = zPlan->iDist;
+
+    zPlan->applyPartialPass = applyPartialPass;
+
     zPlan->AssignParams();
 }
 
@@ -650,7 +798,7 @@ FMKey SBRCTranspose3DNode::GetKernelKey() const
     if(sbrcTranstype == SBRC_TRANSPOSE_TYPE::NONE)
     {
         // find the base kernel at first
-        FMKey baseKey = fpkey(length[0], precision, scheme, TILE_ALIGNED);
+        FMKey baseKey(length[0], precision, scheme, TILE_ALIGNED);
         // if we have the base kernel, then we set the exact sbrc_trans_type and return the real key
         // if we don't, then we simply return a key with NONE sbrc_trans_type
         // which will make KernelCheck() trigger an exception
@@ -661,7 +809,7 @@ FMKey SBRCTranspose3DNode::GetKernelKey() const
         }
     }
 
-    return fpkey(length[0], precision, scheme, sbrcTranstype);
+    return FMKey(length[0], precision, scheme, sbrcTranstype);
 }
 
 bool SBRCTranspose3DNode::KernelCheck(std::vector<FMKey>& kernel_keys)
