@@ -22,7 +22,7 @@
 
 #include "twiddles.h"
 #include "../../shared/arithmetic.h"
-#include "../../shared/hipstream_wrapper.h"
+#include "../../shared/hip_object_wrapper.h"
 #include "../../shared/rocfft_hip.h"
 #include "function_pool.h"
 #include "rtc_cache.h"
@@ -63,7 +63,7 @@ protected:
     bool attach_halfN;
 
     const rocfft_precision precision;
-    const std::string      gpu_arch;
+    hipDeviceProp_t        deviceProp;
 
     void GetKernelParams(const std::vector<size_t>& radices,
                          std::vector<size_t>&       radices_prod,
@@ -134,7 +134,7 @@ protected:
 
         if(attach_halfN)
         {
-            launch_half_N_kernel(stream, device_data_ptr + table_sz);
+            launch_half_N_kernel(stream, device_data_ptr + table_sz, half_N, N);
         }
     }
 
@@ -156,17 +156,18 @@ protected:
 
         auto device_data_ptr = static_cast<T*>(output.data());
 
-        auto kernel = RTCKernelTwiddle::generate(gpu_arch, TwiddleTableType::LENGTH_N, precision);
+        auto kernel = RTCKernelTwiddle::generate(
+            deviceProp.gcnArchName, TwiddleTableType::LENGTH_N, precision);
         RTCKernelArgs kargs;
         kargs.append_size_t(length_limit);
         kargs.append_size_t(N);
         kargs.append_ptr(device_data_ptr);
 
-        kernel.launch(kargs, dim3(numBlocks), dim3(blockSize), 0, stream);
+        kernel.launch(kargs, dim3(numBlocks), dim3(blockSize), 0, deviceProp, stream);
 
         if(attach_halfN)
         {
-            launch_half_N_kernel(stream, device_data_ptr + length);
+            launch_half_N_kernel(stream, device_data_ptr + length, half_N, N);
         }
     }
 
@@ -191,7 +192,8 @@ protected:
         std::copy(radices_prod.begin(), radices_prod.end(), radices_prod_device.data);
         std::copy(radices_sum_prod.begin(), radices_sum_prod.end(), radices_sum_prod_device.data);
 
-        auto kernel = RTCKernelTwiddle::generate(gpu_arch, TwiddleTableType::RADICES, precision);
+        auto kernel = RTCKernelTwiddle::generate(
+            deviceProp.gcnArchName, TwiddleTableType::RADICES, precision);
         RTCKernelArgs kargs;
         kargs.append_size_t(length_limit);
         kargs.append_size_t(num_radices);
@@ -199,14 +201,31 @@ protected:
         kargs.append_struct(radices_prod_device);
         kargs.append_struct(radices_sum_prod_device);
         kargs.append_ptr(output);
-        kernel.launch(kargs, dim3(numBlocksX, numBlocksY), dim3(blockSize, blockSize), 0, stream);
+        kernel.launch(
+            kargs, dim3(numBlocksX, numBlocksY), dim3(blockSize, blockSize), 0, deviceProp, stream);
     }
 
-    void launch_half_N_kernel(hipStream_t& stream, T* output)
+    void launch_pp_twiddle_kernel(hipStream_t& stream, T* output, size_t N)
     {
         auto blockSize = TWIDDLES_THREADS;
 
-        auto kernel = RTCKernelTwiddle::generate(gpu_arch, TwiddleTableType::HALF_N, precision);
+        auto kernel = RTCKernelTwiddle::generate(
+            deviceProp.gcnArchName, TwiddleTableType::PARTIAL_PASS_N, precision);
+        RTCKernelArgs kargs;
+        kargs.append_size_t(N);
+        kargs.append_ptr(output);
+
+        auto numBlocks_N = DivRoundingUp<size_t>(N, blockSize);
+
+        kernel.launch(kargs, dim3(numBlocks_N), dim3(blockSize), 0, deviceProp, stream);
+    }
+
+    void launch_half_N_kernel(hipStream_t& stream, T* output, size_t half_N, size_t N)
+    {
+        auto blockSize = TWIDDLES_THREADS;
+
+        auto kernel = RTCKernelTwiddle::generate(
+            deviceProp.gcnArchName, TwiddleTableType::HALF_N, precision);
         RTCKernelArgs kargs;
         kargs.append_size_t(half_N);
         kargs.append_size_t(N);
@@ -214,20 +233,20 @@ protected:
 
         auto numBlocks_halfN = DivRoundingUp<size_t>(half_N, blockSize);
 
-        kernel.launch(kargs, dim3(numBlocks_halfN), dim3(blockSize), 0, stream);
+        kernel.launch(kargs, dim3(numBlocks_halfN), dim3(blockSize), 0, deviceProp, stream);
     }
 
 public:
-    TwiddleTable(rocfft_precision   precision,
-                 const std::string& gpu_arch,
-                 size_t             _N,
-                 size_t             _length_limit,
-                 bool               _attach_halfN)
+    TwiddleTable(rocfft_precision       precision,
+                 const hipDeviceProp_t& deviceProp,
+                 size_t                 _N,
+                 size_t                 _length_limit,
+                 bool                   _attach_halfN)
         : N(_N)
         , length_limit(_length_limit ? _length_limit : _N)
         , attach_halfN(_attach_halfN)
         , precision(precision)
-        , gpu_arch(gpu_arch)
+        , deviceProp(deviceProp)
     {
         half_N = attach_halfN ? (N + 1) / 2 : 0;
     }
@@ -245,12 +264,21 @@ class TwiddleTable2D : public TwiddleTable<T>
 private:
     size_t N1;
     size_t N2;
+    bool   attach_halfN1;
+    bool   attach_halfN2;
 
 public:
-    TwiddleTable2D(rocfft_precision precision, const std::string& gpu_arch, size_t _N1, size_t _N2)
-        : TwiddleTable<T>(precision, gpu_arch, 0, 0, false)
+    TwiddleTable2D(rocfft_precision       precision,
+                   const hipDeviceProp_t& deviceProp,
+                   size_t                 _N1,
+                   size_t                 _N2,
+                   bool                   _attach_halfN1,
+                   bool                   _attach_halfN2)
+        : TwiddleTable<T>(precision, deviceProp, 0, 0, false)
         , N1(_N1)
         , N2(_N2)
+        , attach_halfN1(_attach_halfN1)
+        , attach_halfN2(_attach_halfN2)
     {
     }
 
@@ -259,6 +287,10 @@ public:
                               hipStream_t&               stream,
                               gpubuf&                    output)
     {
+        // _half_N can be either halfN1 or halfN2, we don't enable both at the same time
+        size_t _half_N = (attach_halfN1) ? (N1 + 1) / 2 : ((attach_halfN2) ? (N2 + 1) / 2 : 0);
+        size_t _N1orN2 = (attach_halfN1) ? N1 : ((attach_halfN2) ? N2 : 0);
+
         if(radices1 == radices2)
             N2 = 0;
 
@@ -277,15 +309,16 @@ public:
         else
             table_sz_2 = N2;
 
-        auto table_sz    = (table_sz_1 + table_sz_2);
-        auto table_bytes = table_sz * sizeof(T);
+        auto table_sz       = table_sz_1 + table_sz_2; // basic twd without PRE_POST
+        auto total_table_sz = table_sz + _half_N; // half_N would be zero if not PRE/POST
+        auto table_bytes    = total_table_sz * sizeof(T);
 
         if(table_bytes == 0)
             return;
 
         if(output.alloc(table_bytes) != hipSuccess)
             throw std::runtime_error("unable to allocate twiddle length "
-                                     + std::to_string(table_sz));
+                                     + std::to_string(total_table_sz));
 
         auto device_data_ptr          = static_cast<T*>(output.data());
         TwiddleTable<T>::length_limit = N1;
@@ -307,6 +340,12 @@ public:
                                                    stream,
                                                    device_data_ptr + table_sz_1);
         }
+
+        if(_half_N)
+        {
+            TwiddleTable<T>::launch_half_N_kernel(
+                stream, device_data_ptr + table_sz, _half_N, _N1orN2);
+        }
     }
 };
 
@@ -322,17 +361,17 @@ private:
     size_t tableSize;
 
     const rocfft_precision precision;
-    const std::string      gpu_arch;
+    hipDeviceProp_t        deviceProp;
 
 public:
-    TwiddleTableLarge(rocfft_precision   precision,
-                      const std::string& gpu_arch,
-                      size_t             length,
-                      size_t             base = LTWD_BASE_DEFAULT)
+    TwiddleTableLarge(rocfft_precision       precision,
+                      const hipDeviceProp_t& deviceProp,
+                      size_t                 length,
+                      size_t                 base = LTWD_BASE_DEFAULT)
         : N(length)
         , largeTwdBase(base)
         , precision(precision)
-        , gpu_arch(gpu_arch)
+        , deviceProp(deviceProp)
     {
         X         = static_cast<size_t>(1) << largeTwdBase; // ex: 2^8 = 256
         Y         = DivRoundingUp<size_t>(CeilPo2(N), largeTwdBase);
@@ -357,7 +396,8 @@ public:
         auto numBlocksX = DivRoundingUp<size_t>(X, blockSize);
         auto numBlocksY = DivRoundingUp<size_t>(Y, blockSize);
 
-        auto kernel = RTCKernelTwiddle::generate(gpu_arch, TwiddleTableType::LARGE, precision);
+        auto kernel = RTCKernelTwiddle::generate(
+            deviceProp.gcnArchName, TwiddleTableType::LARGE, precision);
         RTCKernelArgs kargs;
         kargs.append_double(phi);
         kargs.append_size_t(largeTwdBase);
@@ -365,7 +405,60 @@ public:
         kargs.append_size_t(Y);
         kargs.append_ptr(output.data());
 
-        kernel.launch(kargs, dim3(numBlocksX, numBlocksY), dim3(blockSize, blockSize), 0, stream);
+        kernel.launch(
+            kargs, dim3(numBlocksX, numBlocksY), dim3(blockSize, blockSize), 0, deviceProp, stream);
+    }
+};
+
+// Twiddle table for partial pass
+template <typename T>
+class TwiddleTablePP
+{
+protected:
+    size_t N;
+
+    const rocfft_precision precision;
+    hipDeviceProp_t        deviceProp;
+
+    void launch_twiddle_kernel(hipStream_t& stream, T* output, size_t N)
+    {
+        auto blockSize = TWIDDLES_THREADS;
+
+        auto kernel = RTCKernelTwiddle::generate(
+            deviceProp.gcnArchName, TwiddleTableType::PARTIAL_PASS_N, precision);
+        RTCKernelArgs kargs;
+        kargs.append_size_t(N);
+        kargs.append_ptr(output);
+
+        auto numBlocksX = DivRoundingUp<size_t>(N, blockSize);
+        auto numBlocksY = DivRoundingUp<size_t>(N, blockSize);
+
+        kernel.launch(
+            kargs, dim3(numBlocksX, numBlocksY), dim3(blockSize, blockSize), 0, deviceProp, stream);
+    }
+
+public:
+    TwiddleTablePP(rocfft_precision precision, const hipDeviceProp_t& deviceProp, size_t _N)
+        : N(_N)
+        , precision(precision)
+        , deviceProp(deviceProp)
+    {
+    }
+
+    void GenerateTwiddleTable(hipStream_t& stream, gpubuf& twts)
+    {
+        auto table_bytes = N * N * sizeof(T);
+
+        if(table_bytes == 0)
+            return;
+
+        if(twts.alloc(table_bytes) != hipSuccess)
+            throw std::runtime_error("unable to allocate partial-pass twiddle table of length "
+                                     + std::to_string(N * N));
+
+        auto device_data_ptr = static_cast<T*>(twts.data());
+
+        launch_twiddle_kernel(stream, device_data_ptr, N);
     }
 };
 
@@ -373,7 +466,7 @@ template <typename T>
 gpubuf twiddles_create_pr(size_t                     N,
                           size_t                     length_limit,
                           rocfft_precision           precision,
-                          const char*                gpu_arch,
+                          const hipDeviceProp_t&     deviceProp,
                           size_t                     largeTwdBase,
                           bool                       attach_halfN,
                           const std::vector<size_t>& radices,
@@ -387,17 +480,14 @@ gpubuf twiddles_create_pr(size_t                     N,
         twiddle_streams.resize(deviceId + 1);
     if(twiddle_streams[deviceId] == nullptr)
         twiddle_streams[deviceId].alloc();
-    hipStream_t stream = twiddle_streams[deviceId];
+    hipStream_wrapper_t& stream = twiddle_streams[deviceId];
 
-    if(stream == nullptr)
-    {
-        if(hipStreamCreate(&stream) != hipSuccess)
-            throw std::runtime_error("hipStreamCreate failure");
-    }
+    if(!stream)
+        stream.alloc();
 
     if((N <= LARGE_TWIDDLE_THRESHOLD) && largeTwdBase == 0)
     {
-        TwiddleTable<T> twTable(precision, gpu_arch, N, length_limit, attach_halfN);
+        TwiddleTable<T> twTable(precision, deviceProp, N, length_limit, attach_halfN);
         twTable.GenerateTwiddleTable(radices, stream, twts);
     }
     else
@@ -406,13 +496,13 @@ gpubuf twiddles_create_pr(size_t                     N,
 
         if(largeTwdBase == 0)
         {
-            TwiddleTable<T> twTable(precision, gpu_arch, N, length_limit, attach_halfN);
+            TwiddleTable<T> twTable(precision, deviceProp, N, length_limit, attach_halfN);
             twTable.GenerateTwiddleTable(radices, stream, twts);
         }
         else
         {
             TwiddleTableLarge<T> twTable(
-                precision, gpu_arch, N, largeTwdBase); // does not generate radices
+                precision, deviceProp, N, largeTwdBase); // does not generate radices
             twTable.GenerateTwiddleTable(stream, twts);
         }
     }
@@ -426,7 +516,7 @@ gpubuf twiddles_create_pr(size_t                     N,
 gpubuf twiddles_create(size_t                     N,
                        size_t                     length_limit,
                        rocfft_precision           precision,
-                       const char*                gpu_arch,
+                       const hipDeviceProp_t&     deviceProp,
                        size_t                     largeTwdBase,
                        bool                       attach_halfN,
                        const std::vector<size_t>& radices,
@@ -436,46 +526,35 @@ gpubuf twiddles_create(size_t                     N,
     {
     case rocfft_precision_single:
         return twiddles_create_pr<rocfft_complex<float>>(
-            N, length_limit, precision, gpu_arch, largeTwdBase, attach_halfN, radices, deviceId);
+            N, length_limit, precision, deviceProp, largeTwdBase, attach_halfN, radices, deviceId);
     case rocfft_precision_double:
         return twiddles_create_pr<rocfft_complex<double>>(
-            N, length_limit, precision, gpu_arch, largeTwdBase, attach_halfN, radices, deviceId);
+            N, length_limit, precision, deviceProp, largeTwdBase, attach_halfN, radices, deviceId);
     case rocfft_precision_half:
-        return twiddles_create_pr<rocfft_complex<_Float16>>(
-            N, length_limit, precision, gpu_arch, largeTwdBase, attach_halfN, radices, deviceId);
+        return twiddles_create_pr<rocfft_complex<rocfft_fp16>>(
+            N, length_limit, precision, deviceProp, largeTwdBase, attach_halfN, radices, deviceId);
     }
 }
 
 template <typename T>
-gpubuf twiddles_create_2D_pr(
-    size_t N1, size_t N2, rocfft_precision precision, const char* gpu_arch, unsigned int deviceId)
+gpubuf twiddles_create_2D_pr(size_t                     N1,
+                             size_t                     N2,
+                             rocfft_precision           precision,
+                             const hipDeviceProp_t&     deviceProp,
+                             bool                       attach_halfN,
+                             bool                       attach_halfN2,
+                             const std::vector<size_t>& radices1,
+                             const std::vector<size_t>& radices2,
+                             unsigned int               deviceId)
 {
-    auto                kernel = function_pool::get_kernel(fpkey(N1, N2, precision));
-    std::vector<size_t> radices1, radices2;
-
-    int    count               = 0;
-    size_t cummulative_product = 1;
-    while(cummulative_product != N1)
-    {
-        cummulative_product *= kernel.factors[count++];
-    }
-    radices1.insert(radices1.cbegin(), kernel.factors.cbegin(), kernel.factors.cbegin() + count);
-    radices2.insert(radices2.cbegin(), kernel.factors.cbegin() + count, kernel.factors.cend());
-
     gpubuf twts;
     if(deviceId >= twiddle_streams.size())
         twiddle_streams.resize(deviceId + 1);
-    if(twiddle_streams[deviceId] == nullptr)
-        twiddle_streams[deviceId].alloc();
-    hipStream_t stream = twiddle_streams[deviceId];
+    hipStream_wrapper_t& stream = twiddle_streams[deviceId];
+    if(!stream)
+        stream.alloc();
 
-    if(stream == nullptr)
-    {
-        if(hipStreamCreate(&stream) != hipSuccess)
-            throw std::runtime_error("hipStreamCreate failure");
-    }
-
-    TwiddleTable2D<T> twTable(precision, gpu_arch, N1, N2);
+    TwiddleTable2D<T> twTable(precision, deviceProp, N1, N2, attach_halfN, attach_halfN2);
     twTable.GenerateTwiddleTable(radices1, radices2, stream, twts);
 
     if(hipStreamSynchronize(stream) != hipSuccess)
@@ -484,17 +563,89 @@ gpubuf twiddles_create_2D_pr(
     return twts;
 }
 
-gpubuf twiddles_create_2D(
-    size_t N1, size_t N2, rocfft_precision precision, const char* gpu_arch, unsigned int deviceId)
+gpubuf twiddles_create_2D(size_t                     N1,
+                          size_t                     N2,
+                          rocfft_precision           precision,
+                          const hipDeviceProp_t&     deviceProp,
+                          bool                       attach_halfN,
+                          bool                       attach_halfN2,
+                          const std::vector<size_t>& radices1,
+                          const std::vector<size_t>& radices2,
+                          unsigned int               deviceId)
 {
     switch(precision)
     {
     case rocfft_precision_single:
-        return twiddles_create_2D_pr<rocfft_complex<float>>(N1, N2, precision, gpu_arch, deviceId);
+        return twiddles_create_2D_pr<rocfft_complex<float>>(N1,
+                                                            N2,
+                                                            precision,
+                                                            deviceProp,
+                                                            attach_halfN,
+                                                            attach_halfN2,
+                                                            radices1,
+                                                            radices2,
+                                                            deviceId);
     case rocfft_precision_double:
-        return twiddles_create_2D_pr<rocfft_complex<double>>(N1, N2, precision, gpu_arch, deviceId);
+        return twiddles_create_2D_pr<rocfft_complex<double>>(N1,
+                                                             N2,
+                                                             precision,
+                                                             deviceProp,
+                                                             attach_halfN,
+                                                             attach_halfN2,
+                                                             radices1,
+                                                             radices2,
+                                                             deviceId);
     case rocfft_precision_half:
-        return twiddles_create_2D_pr<rocfft_complex<_Float16>>(
-            N1, N2, precision, gpu_arch, deviceId);
+        return twiddles_create_2D_pr<rocfft_complex<rocfft_fp16>>(N1,
+                                                                  N2,
+                                                                  precision,
+                                                                  deviceProp,
+                                                                  attach_halfN,
+                                                                  attach_halfN2,
+                                                                  radices1,
+                                                                  radices2,
+                                                                  deviceId);
+    }
+}
+
+template <typename T>
+gpubuf twiddles_pp_create_pr(size_t                 N,
+                             rocfft_precision       precision,
+                             const hipDeviceProp_t& deviceProp,
+                             unsigned int           deviceId)
+{
+    gpubuf twts;
+    if(deviceId >= twiddle_streams.size())
+        twiddle_streams.resize(deviceId + 1);
+    if(twiddle_streams[deviceId] == nullptr)
+        twiddle_streams[deviceId].alloc();
+    hipStream_wrapper_t& stream = twiddle_streams[deviceId];
+
+    if(!stream)
+        stream.alloc();
+
+    TwiddleTablePP<T> twTable(precision, deviceProp, N);
+    twTable.GenerateTwiddleTable(stream, twts);
+
+    if(hipStreamSynchronize(stream) != hipSuccess)
+        throw std::runtime_error("hipStream failure");
+
+    return twts;
+}
+
+gpubuf twiddles_pp_create(size_t                 N,
+                          rocfft_precision       precision,
+                          const hipDeviceProp_t& deviceProp,
+                          unsigned int           deviceId)
+{
+    switch(precision)
+    {
+    case rocfft_precision_single:
+        return twiddles_pp_create_pr<rocfft_complex<float>>(N, precision, deviceProp, deviceId);
+    case rocfft_precision_double:
+        return twiddles_pp_create_pr<rocfft_complex<double>>(N, precision, deviceProp, deviceId);
+    case rocfft_precision_half:
+        return twiddles_pp_create_pr<rocfft_complex<rocfft_fp16>>(
+            N, precision, deviceProp, deviceId);
     }
 }
