@@ -1,4 +1,4 @@
-# Copyright (C) 2021 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+# Copyright (C) 2021 - 2024 Advanced Micro Devices, Inc. All rights reserved.
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -62,6 +62,11 @@ class Problem:
     istride: List[int] = None
     ostride: List[int] = None
     nbatch: int = 1
+    mp_size: int = 1
+    mp_exec: str = ""
+    ingrid: List[int] = None
+    outgrid: List[int] = None
+    ngpus: int = 1
     idist: int = 0
     odist: int = 0
     direction: int = -1
@@ -71,6 +76,7 @@ class Problem:
     tag: str = None
     min_wgs: int = 64
     max_wgs: int = 512
+    full_token: bool = False
     meta: Dict[str, str] = field(default_factory=dict)
 
     def toJSON(self):
