@@ -1,4 +1,4 @@
-# Copyright (C) 2021 - 2022 Advanced Micro Devices, Inc. All rights reserved.
+# Copyright (C) 2021 - 2024 Advanced Micro Devices, Inc. All rights reserved.
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -33,6 +33,12 @@ all_inplaces = [True, False]
 all_reals = [True, False]
 def_tuning_min_wgs = 64
 def_tuning_max_wgs = 512
+def_export_full_token = False
+default_ngpus = 1
+def_mp_size = 1
+def_mp_exec = '/usr/bin/mpirun'
+def_ingrid = [1, 1, 1]
+def_outgrid = [1, 1, 1]
 
 # yapf: disable
 lengths = {
@@ -81,7 +87,6 @@ lengths = {
 
         (128,128,256),
         (240,224,224),
-        (64,64,64),
         (80,84,14),
         (80,84,144),
 
@@ -324,6 +329,40 @@ lengths = {
          79860, 81920, 83521, 87808, 95832, 98304, 102400, 106496, 110592,
          114688, 117649,
      ],
+
+    'mgpu': [
+        (256),
+        (256, 256),
+        (256, 256, 256),
+        (128, 256, 512),
+        (512, 512, 512),
+        (630, 630, 630),
+    ],
+
+    'mpi': [
+        (128, 256, 512),
+        (256, 256, 256),
+        (378, 189, 63),
+        (512, 512, 512),
+        (630, 310, 630),
+    ],
+
+    'scaling2D': [
+        (128, 128),
+        (512, 512),
+        (1024, 1024),
+        (2048, 2048),
+        (8192, 8192),
+    ],
+
+    'scaling3D': [
+        (64, 64, 64),
+        (128, 128, 128),
+        (512, 512, 512),
+        (1024, 1024, 1024),
+        (2048, 2048, 2048),
+    ],
+
 }
 # yapf: enable
 
@@ -346,36 +385,35 @@ def mktag(tag, dimension, precision, direction, inplace, real):
 
 
 # yield problem sizes with default precision, direction, etc
-def default_length_params(tag, lengths, nbatch, precisions=all_precisions, \
-    directions=all_directions, inplaces=all_inplaces, reals=all_reals, min_wgs=def_tuning_min_wgs, max_wgs=def_tuning_max_wgs):
-
-    # workaround: disable failing token on gfx906
-    if perflib.specs.get_machine_specs(0).gpuid == '0x66a1':
-        gfx906 = True
-    else:
-        gfx906 = False
+def default_length_params(tag, lengths, nbatch, ngpus=default_ngpus, mp_size=def_mp_size, \
+                          mp_exec=def_mp_exec, precisions=all_precisions, \
+                          directions=all_directions, inplaces=all_inplaces, \
+                          reals=all_reals, min_wgs=def_tuning_min_wgs, \
+                          max_wgs=def_tuning_max_wgs,  ingrid=def_ingrid, outgrid=def_outgrid, \
+                          full_token=def_export_full_token, meta = {}):
 
     for precision, direction, inplace, real in product(precisions, directions,
                                                        inplaces, reals):
         for length in lengths:
             length = (length, ) if isinstance(length, int) else length
 
-            # workaround: disable failing token on gfx906
-            if gfx906 and (length == [32768, 32768] and nbatch == 1
-                           and direction == -1 and not inplace and real
-                           and precision == 'single'):
-                continue
-
             yield Problem(length,
                           tag=mktag(tag, len(length), precision, direction,
                                     inplace, real),
                           nbatch=nbatch,
+                          ngpus=ngpus,
+                          mp_size=mp_size,
+                          mp_exec=mp_exec,
+                          ingrid=ingrid,
+                          outgrid=outgrid,
                           direction=direction,
                           inplace=inplace,
                           real=real,
                           precision=precision,
                           min_wgs=min_wgs,
-                          max_wgs=max_wgs)
+                          max_wgs=max_wgs,
+                          full_token=full_token,
+                          meta=meta)
 
 
 def md():
@@ -408,19 +446,26 @@ def qa():
                   real=False,
                   precision='double')
 
-    yield from default_length_params("336x336x56", [(336, 336, 56)],
+    yield from default_length_params("336x336x56_b1", [(336, 336, 56)],
                                      1,
                                      directions=[-1],
                                      precisions=['double'],
                                      inplaces=[True, False],
                                      reals=[False])
 
+    yield from default_length_params("336x336x56_b10", [(336, 336, 56)],
+                                     10,
+                                     directions=[-1],
+                                     precisions=['double'],
+                                     inplaces=[True, False],
+                                     reals=[False])
+
     for length3 in lengths['md']:
         for direction in [-1, 1]:
             yield Problem(length3,
                           tag=mktag('qa3md', 3, 'single', direction, False,
                                     True),
-                          nbatch=1,
+                          nbatch=10,
                           direction=direction,
                           inplace=False,
                           real=True,
@@ -585,6 +630,95 @@ def new_large_1d():
                                      reals=[False])
 
 
+def mpi():
+    """MPI sizes."""
+
+    # Grid sizes can be overwritten from CLI
+    yield from default_length_params("mpi",
+                                     lengths['mpi'],
+                                     1,
+                                     precisions=['single'],
+                                     reals=[False])
+
+
+def mgpu():
+    """Multi-GPU sizes."""
+
+    yield from default_length_params("mgpu",
+                                     lengths['mgpu'],
+                                     1,
+                                     precisions=['single'],
+                                     reals=[False])
+
+
+def strongScaling():
+    """Strong scalability test sizes."""
+
+    for length in lengths['scaling2D']:
+        direction = -1
+        precision = 'double'
+        nbatch = 1
+        yield Problem(length,
+                      tag=mktag(
+                          'strongScaling' + "x".join([str(x) for x in length]),
+                          nbatch, precision, direction, False, False),
+                      nbatch=nbatch,
+                      direction=direction,
+                      inplace=False,
+                      real=False,
+                      precision='double',
+                      meta={'scaling': 'strong'})
+
+    for length in lengths['scaling3D']:
+        direction = -1
+        precision = 'double'
+        nbatch = 1
+        yield Problem(length,
+                      tag=mktag(
+                          'strongScaling' + "x".join([str(x) for x in length]),
+                          nbatch, precision, direction, False, False),
+                      nbatch=nbatch,
+                      direction=direction,
+                      inplace=False,
+                      real=False,
+                      precision='double',
+                      meta={'scaling': 'strong'})
+
+
+def weakScaling():
+    """Weak scalability test sizes."""
+
+    for length in lengths['scaling2D']:
+        direction = -1
+        precision = 'double'
+        nbatch = 1
+        yield Problem(length,
+                      tag=mktag(
+                          'weakScaling' + "x".join([str(x) for x in length]),
+                          nbatch, precision, direction, False, False),
+                      nbatch=nbatch,
+                      direction=direction,
+                      inplace=False,
+                      real=False,
+                      precision='double',
+                      meta={'scaling': 'weak'})
+
+    for length in lengths['scaling3D']:
+        direction = -1
+        precision = 'double'
+        nbatch = 1
+        yield Problem(length,
+                      tag=mktag(
+                          'weakScaling' + "x".join([str(x) for x in length]),
+                          nbatch, precision, direction, False, False),
+                      nbatch=nbatch,
+                      direction=direction,
+                      inplace=False,
+                      real=False,
+                      precision='double',
+                      meta={'scaling': 'weak'})
+
+
 def unbatched_1d():
     """All tested single-batch 1D transforms."""
 
@@ -653,7 +787,7 @@ def batch_const_count():
     # batch * length = 2^25 ... 2^30
     for direction in [-1, 1]:
         for precision in all_precisions:
-            for exp in [25, 26, 27, 28, 29, 30]:
+            for exp in [25, 26, 27, 28, 29, 30, 31, 32]:
                 for place in all_inplaces:
                     for lexp in range(4, exp + 1):
                         length = 2**lexp
@@ -721,6 +855,55 @@ def short_test():
 def tuning_example():
     """tuning 3 examples problems"""
 
+    # real 1d : odd/even/small/large - fwd
+    length1s = [243, 486, 16807, 16384]
+    yield from default_length_params("real_1d_fwd",
+                                     length1s,
+                                     40000,
+                                     directions=[-1],
+                                     precisions=['double'],
+                                     inplaces=[True, False],
+                                     reals=[True],
+                                     min_wgs=128,
+                                     max_wgs=256)
+
+    # real 1d : odd/even/small/large - bwd - FIXME: bugs: batchsize for bwd can't exceed 1024
+    yield from default_length_params("real_1d_fwd",
+                                     length1s,
+                                     1024,
+                                     directions=[1],
+                                     precisions=['double'],
+                                     inplaces=[True, False],
+                                     reals=[True],
+                                     min_wgs=128,
+                                     max_wgs=256)
+
+    # real 2d
+    length2s = [(55, 55), (64, 52), (52, 52)]
+    yield from default_length_params("real_2d",
+                                     length2s,
+                                     10000,
+                                     directions=[-1, 1],
+                                     precisions=['double'],
+                                     inplaces=[False],
+                                     reals=[True],
+                                     min_wgs=128,
+                                     max_wgs=256)
+
+    # real 3d
+    length3s = [(75, 55, 55), (208, 104, 104), (100, 100, 100),
+                (200, 200, 200)]
+    yield from default_length_params("real_3d",
+                                     length3s,
+                                     10,
+                                     directions=[-1, 1],
+                                     precisions=['double'],
+                                     inplaces=[False],
+                                     reals=[True],
+                                     min_wgs=128,
+                                     max_wgs=256)
+
+    # complex
     yield from default_length_params("81_1d", [(81)],
                                      60000,
                                      directions=[-1],
@@ -750,6 +933,21 @@ def tuning_example():
                                      max_wgs=256)
 
 
+def tuning_2D_example():
+    """tuning 3 examples problems"""
+
+    lengths = [(81, 32), (81, 25), (125, 32)]
+
+    yield from default_length_params("2D_SINGLE",
+                                     lengths,
+                                     10000,
+                                     directions=[-1],
+                                     precisions=['single'],
+                                     inplaces=[False],
+                                     reals=[False],
+                                     max_wgs=1024)
+
+
 def tuning_suite():
     """tuning"""
 
@@ -790,14 +988,36 @@ def tuning_suite():
                           max_wgs=256)
 
     # we'd like to search more for this problem, so min_wgs = 64, not 128
-    yield from default_length_params("336x336x56", [(336, 336, 56)],
+    yield from default_length_params("336x336x56_b1", [(336, 336, 56)],
                                      1,
                                      directions=[-1],
                                      precisions=['double'],
                                      inplaces=[True, False],
                                      reals=[False],
+                                     max_wgs=256,
+                                     full_token=True)
+
+    yield from default_length_params("336x336x56_b10", [(336, 336, 56)],
+                                     10,
+                                     directions=[-1],
+                                     precisions=['double'],
+                                     inplaces=[True, False],
+                                     reals=[False],
                                      max_wgs=256)
 
+    for length3 in lengths['md']:
+        for direction in [-1, 1]:
+            yield Problem(length3,
+                          tag=mktag('qa3md', 3, 'single', direction, False,
+                                    True),
+                          nbatch=10,
+                          direction=direction,
+                          inplace=False,
+                          real=True,
+                          precision='single',
+                          min_wgs=128,
+                          max_wgs=256)
+
     for length in lengths['qa1d10b']:
         yield Problem([length],
                       tag=mktag("qa1d10b", 1, 'single', -1, True, False),
@@ -827,3 +1047,27 @@ def tuning_suite():
                       real=False,
                       precision='single',
                       min_wgs=128)
+
+
+def partial_pass():
+    for direction in [-1, 1]:
+        for precision in ['single', 'double']:
+            for place in all_inplaces:
+                for batch in [
+                        1, 2, 5, 10, 15, 20, 25, 50, 75, 100, 150, 200, 250,
+                        500, 750, 1000, 1250, 1500, 1575, 2000, 2500, 3000,
+                        3500, 4000, 4500, 5000, 7500, 10000
+                ]:
+
+                    length = (64, 64, 64)
+                    placestr = "ip" if place else "op"
+                    yield Problem(length,
+                                  tag=mktag("partial_pass_" + placestr, 3,
+                                            precision, direction, False,
+                                            False),
+                                  nbatch=batch,
+                                  direction=direction,
+                                  inplace=place,
+                                  real=False,
+                                  meta={'ivariable': 'batch'},
+                                  precision=precision)
